#!/usr/bin/env pythonw
# coding: latin-1

#-------------------------------------------------------------------------------
#
#  Name:  RCT
#
#  Purpose: Remote configuration of CZM instruments using Zeroconf techniques
#           to discover connected devices. Implemented as a wizard / assistant
#           it guides the user through the configuration process. This version
#           is limited to configuring HFA-NET settings.
#
#  Required Resources: Python, wxPython, pywin32, CZM Zeroconf
#
#  (c) Copyright 2007 Carl Zeiss Meditec, Inc.  All rights reserved.
#  THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF CARL ZEISS MEDITEC, INC.
#  The copyright notice does not evidence any actual or intended publication.
#
#  Author: Dana Wheeler
#
#  Date: 12-03-07 1:41 PM
#
#  Edit History:
#
#  0011 11-20-07 DRW Corrected button label on test for administrator alert.
#                    Only install Bonjour once per session.
#                    Display wait cursor while opening the Windows Firewal.
#  0010 10-19-07 DRW Added white title banner.
#                    Added Apple Bonjour End User License Agreement dialog.
#                    Install the mDNSResponder as a service.
#                    Kill the mDNSResponder when run directly in Report
#                        Mode.
#                    Added start page commands to display the RCT User
#                        Guide and to open CIFS TCP ports in a Windows XP
#                        firewall.
#                    Added password encryption.
#                    Added option of separate EMR subfolders on the Shared
#                        Folder page.
#                    Added support for multi-language RCT User Guide.
#                    Added a popup command menu for each page.
#                    Set focus to password field on entering Custom mode.
#                    Added a warning if a password conflict is likely.
#  0009 08-06-07 DRW Added a saveRctProperties function to save program
#                    properties prior to shutdown, allowing RCT to remember
#                    its current language selection.
#  0008 08-02-07 DRW Support copying FTP configurations, correcting defect
#                    3627.
#                    Changed prompt after the last instrument is tested to
#                    read "Testing is complete, select 'Next' to continue.".
#  0007 06-29-07 DRW Added COPY_MODE to the list of modes from which
#                    test results can be saved. Corrects the re-opened
#                    defect 3615.
#  0006 06-28-07 DRW Conditionally disable the default printer check box.
#  0005 06-27-07 DRW Added dynamic multi-language setup. Only the German
#                    translation is complete in this version.
#                    Corrects defects 3615 and 3618.
#  0004 05-25-07 DRW Initial release to SQA
#
#-------------------------------------------------------------------------------

import wx
import wx.wizard
import wx.html
import codecs
import sys
import gettext
import os
import socket
import pprint
import optparse
import copy
import locale
import time
import threading
import tempfile
import shutil
import subprocess
import ces

# Import the required pywin32 modules.

try:
    import win32api
    import win32pdhutil
    import win32con
    import win32net
    import win32netcon
    import win32security
    import ntsecuritycon
    import win32print
    from win32com.shell import shell
    from win32com.shell import shellcon
    RCT_WIN32 = True
except:
    RCT_WIN32 = False

# Import RCT support modules.

try:
    import ces
except ImportError:
    pass

# Multi-language user interface variables.

try:
    _                         # test for the GNU gettext translation function.
except NameError:             # Ignore error when gettext translation is not available.
    def _(s): return s

DEFAULT_LOCALE = 'en'         # default language, US English.
wxlocale = None               # wxPython locale object.

locales = {'en': (wx.LANGUAGE_ENGLISH, 'en_US.UTF-8'),
           'fr': (wx.LANGUAGE_FRENCH,  'fr_FR.UTF-8'),
           'it': (wx.LANGUAGE_ITALIAN, 'it_IT.UTF-8'),
           'de': (wx.LANGUAGE_GERMAN,  'de_DE.UTF-8'),
           'es': (wx.LANGUAGE_SPANISH, 'es_ES.UTF-8')
          }

# Exception names.

RctTerminated = 'RCT terminated'
RctLanguageChanged = 'Language Changed'

# RCT and Service object instance variables.

rct = None
serviceUser = None

# Global program arguments.

VERBOSE = False
DELETE_FILES = False
DEBUG = False
DEBUG_PAGING = False
CONFIG_ENABLED = True
FRIENDS = []
rct_properties = {}
forceRestart = False

# Global development booleans.

SAVE_TEST_RESULTS = True
INCLUDE_SERVICE_TYPE_QUALIFIER = True
DEFAULT_PRINTER_ENABLED = False
LABEL_UTILITIES = False
CREATE_CMD_POPUP = True

STATUS_OK = 0
STATUS_ERROR = 1

TBD = None
NOT_AVAILABLE = 'N/A'
NULL_VALUE = '__NULL_VALUE__'
ALTERNATE_DELIMITER = '__EQ__'

READY_TO_TEST = 'Ready to Test'
TESTING = 'Testing...'
PASSED = 'Passed'
FAILED = 'Failed'
NO_RESULTS = 'No Results'

# Product name strings.

COMPANY_NAME = 'Zeiss'
COMPANY_SHORT_NAME = 'CZM'
COMPANY_LONG_NAME = 'Carl Zeiss Meditec'
PRODUCT_SHORT_NAME = 'RCT'
PRODUCT_VERSION = '1.0'

# Instrument type strings.

HFA = 'HFA2i'

PHOENIX = 'HFA3'
OCT = 'Stratus'
FRODO = 'Cirrus'
AC_OCT = 'Visante'
GDX = 'GDx'
ATLAS = 'Atlas'

SHARED_DIRECTORY_MODE = 0777

# Data and configuration files

EXCLUDED_PRINTERS_FILE = 'ExcludedPrinters.txt'
LOGO_IMAGE_FILE = 'czm_logo.bmp'
TEST_RESULTS_FILE = 'ConnectionTestResults.txt'
RCT_WORK = ''

# Service constants.

SERVICE_TYPE = '_czm._tcp'

# HFA data export subfolders.

ARCHIVE_FOLDER = '/Archive'
DATA_EXPORT_FOLDER = '/Data_Export'
EMR_EXPORT_FOLDER = '/EMR_Export'
WORK_LIST_FOLDER = '/Work_List'

# Device data names.

DEVICE_TYPE_FIELD = 'Type'
SERIAL_NUMBER_FIELD = 'Serial Number'
DEVICE_NAME_FIELD = 'Name'
DEVICE_NETWORK_SETTINGS_FIELD = 'Network Settings'
CAPABILITIES_FIELD = 'Capabilities'
TEST_RESULTS_FIELD = 'Test'
SELECTED_FIELD = 'Selected'
RAW_DEVICE_FIELD = 'DEVICE_INFO'

FEATURE_FIELD = 'Feature'
SETUP_TYPE_FIELD = 'Setup Type'
HOST_NAME_FIELD = 'File Server'
HOST_IP_ADDRESS_FIELD = 'IP Address'
SHARED_RESOURCE_FIELD = 'Share'
PRINTER_NAME_FIELD = 'Printer Name'
SHARE_NAME_FIELD = 'Share Name'
USER_NAME_FIELD = 'User Name'
PASSWORD_FIELD = 'Password'
NETWORK_PROTOCOL_FIELD = 'Network Protocol'
CONNECTION_TEST_STATUS = 'CONNECTION_TEST_STATUS'
ESTABLISH_CONNECTION = 'ESTABLISH_CONNECTION'
CREATE_FILE = 'CREATE_FILE'
READ_TEST = 'READ_TEST'
DELETE_FILE = 'DELETE_FILE'
COMMUNICATION_TEST_WITH_EMR_APPLICATION = 'COMMUNICATION_TEST_WITH_EMR_APPLICATION'
COMMUNICATION_TEST_WITH_WORK_LIST_APPLICATION = 'COMMUNICATION_TEST_WITH_WORK_LIST_APPLICATION'
DHCP_MODE = 'DHCP'
LINK_LOCAL_MODE = 'LINKLOCAL'
MANUAL_MODE = 'MANUAL'
UNKNOWN_MODE = 'UNEXPECTED IP'

# Instrument Network Settings.

DEVICE_IP_CONFIG_PREFERENCE = 'IP Configuration Preference'
DEVICE_IP_CONFIG = 'IP Configuration'
DEVICE_REMOTE_CONFIGURATION = 'Remote Configuration Service'
DEVICE_IP_ADDRESS = 'IP Address'
DEVICE_SUBNET_MASK = 'Subnet Mask'
DEVICE_DEFAULT_GATEWAY = 'Default Gateway'
DEVICE_WORKGROUP_DOMAIN = 'Workgroup/Domain'

# Capabilities.

HFA_ARCHIVE_SETUP = 'Archive/Retrieve Setup'
HFA_DATA_EXPORT_SETUP = 'Data Export Setup'
HFA_EMR_PMS_SETUP = 'EMR/PMS Export Setup'
HFA_WORK_LIST_SETUP = 'Work List Setup'
HFA_REMOTE_PRINTER_SETUP = 'Shared Printer Setup'

FEATURES_NOT_TO_TEST = [HFA_REMOTE_PRINTER_SETUP]

# Global configuration data containing the currently selected configuration settings.

PATIENT_FOLDER = 'ENABLED'

# Configuration mode names.

EASY_MODE = 'Easy Mode'
COPY_MODE = 'Copy Mode'
CUSTOM_MODE = 'Custom Mode'
REPORT_MODE = 'Report Mode'
TEST_MODE = 'Test Mode'

# Configurable setup names.

SHARED_FOLDER_TYPE = 'Shared Folder'
SHARED_PRINTER_TYPE = 'Shared Printer'
ALL_SETUP_TYPES = 'All Features'

# Initial values for device data.

NULL_NETWORK_SETTINGS = {DEVICE_IP_CONFIG_PREFERENCE: NOT_AVAILABLE, DEVICE_IP_CONFIG: NOT_AVAILABLE,
                         DEVICE_REMOTE_CONFIGURATION: NOT_AVAILABLE,
                         DEVICE_IP_ADDRESS: NOT_AVAILABLE, DEVICE_SUBNET_MASK: NOT_AVAILABLE,
                         DEVICE_DEFAULT_GATEWAY: NOT_AVAILABLE, DEVICE_WORKGROUP_DOMAIN: NOT_AVAILABLE}

NULL_INSTRUMENT = {DEVICE_TYPE_FIELD: NOT_AVAILABLE, SERIAL_NUMBER_FIELD: NOT_AVAILABLE,
                   DEVICE_NAME_FIELD: NOT_AVAILABLE, DEVICE_NETWORK_SETTINGS_FIELD: copy.copy(NULL_NETWORK_SETTINGS),
                   CAPABILITIES_FIELD: None, TEST_RESULTS_FIELD: NOT_AVAILABLE}

NULL_HFA_SETUP = {HOST_NAME_FIELD: '', HOST_IP_ADDRESS_FIELD: '', SHARED_RESOURCE_FIELD: '', SHARE_NAME_FIELD: '',
                  SETUP_TYPE_FIELD: SHARED_FOLDER_TYPE,
                  USER_NAME_FIELD: '', PASSWORD_FIELD: '',
                  TEST_RESULTS_FIELD: READY_TO_TEST,
                  SELECTED_FIELD: False}

NULL_SHARED_PRINTER_SETUP = {HOST_NAME_FIELD: '', HOST_IP_ADDRESS_FIELD: '',
                             SHARED_RESOURCE_FIELD: '', SHARE_NAME_FIELD: '',
                             SETUP_TYPE_FIELD: SHARED_PRINTER_TYPE,
                             USER_NAME_FIELD: '', PASSWORD_FIELD: '',
                             TEST_RESULTS_FIELD: READY_TO_TEST,
                             SELECTED_FIELD: False}

NULL_HFA_CAPABILITY = {HFA_ARCHIVE_SETUP: copy.copy(NULL_HFA_SETUP),
                       HFA_DATA_EXPORT_SETUP: copy.copy(NULL_HFA_SETUP),
                       HFA_EMR_PMS_SETUP: copy.copy(NULL_HFA_SETUP),
                       HFA_WORK_LIST_SETUP: copy.copy(NULL_HFA_SETUP),
                       HFA_REMOTE_PRINTER_SETUP: copy.copy(NULL_SHARED_PRINTER_SETUP)}

# Global references to controls.

# Instrument list control selectors.

instrumentSelector = None
devicesToTest = None

# Other module constants.

PADDING = 8

if os.name == 'nt':
    LIST_WIDTH = 650
    LIST_HEIGHT = 180
    DETAIL_LIST_HEIGHT = 150
else:
    LIST_WIDTH = 600
    LIST_HEIGHT = 160
    DETAIL_LIST_HEIGHT = 130

NO_AVAILABLE_SERVICES = -1
NO_DEVICE_INFORMATION = -2
UNEXPECTED_RESULT = -3
RESTART_REQUESTED = -4
PASSWORD_CHANGE_REQUESTED = -5

WAIT_TO_RETURN = True
DONT_WAIT_TO_RETURN = False

GOING_FORWARD = True
GOING_BACKWARD = False

SUCCESS = True
FAILURE = False

# Background colors.

RCT_DARK_BG = wx.Colour(104, 138, 179)
RCT_LT_BG = wx.Colour(174, 191, 212)
RCT_TITLE_BG = 'white'
RCT_MSG_BG = 'yellow'

NETWORK_ENCRYPT_ID = 1308

bonjourInstalled = False
rctProcs = []


def isWindows():
    
    return 'wxMSW' in wx.PlatformInfo


def isMac():
    
    return 'wxMac' in wx.PlatformInfo


def fieldToString(field):
    
    """
    fieldToString --
    
    Map field names to localizable strings, so that the symbolic constants can be
    used as dictionary keys.
    """
    
    strings = {DEVICE_NAME_FIELD.lower(): _('Name'),
               DEVICE_TYPE_FIELD.lower(): _('Type'),
               FEATURE_FIELD.lower(): _('Feature'),
               HOST_NAME_FIELD.lower(): _('File Server'),
               PASSWORD_FIELD.lower(): _('Password'),
               PRINTER_NAME_FIELD.lower(): _('Printer Name'),
               SERIAL_NUMBER_FIELD.lower(): _('Serial Number'),
               SHARED_FOLDER_TYPE.lower(): _('Shared Folder'),
               SHARED_PRINTER_TYPE.lower(): _('Shared Printer'),
               SHARED_RESOURCE_FIELD.lower(): _('Share'),
               SHARE_NAME_FIELD.lower(): _('Share Name'),
               TEST_RESULTS_FIELD.lower(): _('Test'),
               USER_NAME_FIELD.lower(): _('User Name'),
               CONNECTION_TEST_STATUS.lower(): _('Connection test status'),
               ESTABLISH_CONNECTION.lower(): _('Establish connection'),
               CREATE_FILE.lower().lower(): _('Create file'),
               READ_TEST.lower(): _('Read test'),
               DELETE_FILE.lower(): _('Delete file'),
               COMMUNICATION_TEST_WITH_EMR_APPLICATION.lower(): _('Test for response from the EMR host'),
               COMMUNICATION_TEST_WITH_WORK_LIST_APPLICATION.lower(): _('Test for response from the Work List host'),
               HFA_ARCHIVE_SETUP.lower(): _('Archive/Retrieve Setup'),
               HFA_DATA_EXPORT_SETUP.lower(): _('Data Export Setup'),
               HFA_EMR_PMS_SETUP.lower(): _('EMR/PMS Export Setup'),
               HFA_WORK_LIST_SETUP.lower(): _('Work List Setup'),
               HFA_REMOTE_PRINTER_SETUP.lower(): _('Shared Printer Setup'),
               READY_TO_TEST.lower(): _('Ready to Test'),
               TESTING.lower(): _('Testing...'),
               PASSED.lower(): _('PASSED'),
               FAILED.lower(): _('FAILED'),
               NO_RESULTS.lower(): _('NO RESULTS'),
               EASY_MODE.lower(): _('Easy Mode'),
               COPY_MODE.lower(): _('Copy Mode'),
               CUSTOM_MODE.lower(): _('Custom Mode'),
               REPORT_MODE.lower(): _('Report Mode'),
               TEST_MODE.lower(): _('Test Mode'),
               NOT_AVAILABLE.lower(): _('N/A'),
               DHCP_MODE.lower(): _('DHCP'),
               LINK_LOCAL_MODE.lower(): _('Link Local'),
               MANUAL_MODE.lower(): _('Manual'),
               UNKNOWN_MODE.lower(): _('Unexpected IP')
               }
    
    try:
        return strings[field.lower()]
    except KeyError:
        return field   # fieldToString


def stringToField(localString):
    
    """
    stringToField --
    
    Map locale strings to their corresponding symbolic constants used as dictionary keys.
    """
    
    fields =  {fieldToString(DEVICE_NAME_FIELD): DEVICE_NAME_FIELD,
               fieldToString(DEVICE_TYPE_FIELD): DEVICE_TYPE_FIELD,
               fieldToString(FEATURE_FIELD): FEATURE_FIELD,
               fieldToString(HOST_NAME_FIELD): HOST_NAME_FIELD,
               fieldToString(PASSWORD_FIELD): PASSWORD_FIELD,
               fieldToString(PRINTER_NAME_FIELD): PRINTER_NAME_FIELD,
               fieldToString(SERIAL_NUMBER_FIELD): SERIAL_NUMBER_FIELD,
               fieldToString(SHARED_FOLDER_TYPE): SHARED_FOLDER_TYPE,
               fieldToString(SHARED_PRINTER_TYPE): SHARED_PRINTER_TYPE,
               fieldToString(SHARED_RESOURCE_FIELD): SHARED_RESOURCE_FIELD,
               fieldToString(SHARE_NAME_FIELD): SHARE_NAME_FIELD,
               fieldToString(TEST_RESULTS_FIELD): TEST_RESULTS_FIELD,
               fieldToString(USER_NAME_FIELD): USER_NAME_FIELD,
               fieldToString(CONNECTION_TEST_STATUS): CONNECTION_TEST_STATUS,
               fieldToString(ESTABLISH_CONNECTION): ESTABLISH_CONNECTION,
               fieldToString(CREATE_FILE): CREATE_FILE,
               fieldToString(READ_TEST): READ_TEST,
               fieldToString(DELETE_FILE): DELETE_FILE,
               fieldToString(COMMUNICATION_TEST_WITH_EMR_APPLICATION): COMMUNICATION_TEST_WITH_EMR_APPLICATION,
               fieldToString(COMMUNICATION_TEST_WITH_WORK_LIST_APPLICATION): COMMUNICATION_TEST_WITH_WORK_LIST_APPLICATION,
               fieldToString(HFA_ARCHIVE_SETUP): HFA_ARCHIVE_SETUP,
               fieldToString(HFA_DATA_EXPORT_SETUP): HFA_DATA_EXPORT_SETUP,
               fieldToString(HFA_EMR_PMS_SETUP): HFA_EMR_PMS_SETUP,
               fieldToString(HFA_WORK_LIST_SETUP): HFA_WORK_LIST_SETUP,
               fieldToString(HFA_REMOTE_PRINTER_SETUP): HFA_REMOTE_PRINTER_SETUP,
               fieldToString(READY_TO_TEST): READY_TO_TEST,
               fieldToString(TESTING): TESTING,
               fieldToString(PASSED): PASSED,
               fieldToString(FAILED): FAILED,
               fieldToString(NO_RESULTS): NO_RESULTS,
               fieldToString(EASY_MODE): EASY_MODE,
               fieldToString(COPY_MODE): COPY_MODE,
               fieldToString(CUSTOM_MODE): CUSTOM_MODE,
               fieldToString(REPORT_MODE): REPORT_MODE,
               fieldToString(TEST_MODE): TEST_MODE,
               fieldToString(NOT_AVAILABLE): NOT_AVAILABLE,
               fieldToString(DHCP_MODE): DHCP_MODE,
               fieldToString(LINK_LOCAL_MODE): LINK_LOCAL_MODE,
               fieldToString(MANUAL_MODE): MANUAL_MODE,
               fieldToString(UNKNOWN_MODE): UNKNOWN_MODE
               }
    
    if localString not in fields:
        if DEBUG:
            print 'Could not find field for localized string \'%s\'.' % localString
        return localString
    
    return fields[localString]   # stringToField


class Device(dict):
    
    __deviceList = []
    
    def __init__(self, default = NOT_AVAILABLE, type = HFA, serialNumber = NOT_AVAILABLE,
                       name = NOT_AVAILABLE, capabilities = {},
                       networkSettings = {}):
        
        self.__default = default
        
        dict.__init__(self, {DEVICE_TYPE_FIELD: type,
                             SERIAL_NUMBER_FIELD: serialNumber,
                             DEVICE_NAME_FIELD: name,
                             CAPABILITIES_FIELD: capabilities,
                             DEVICE_NETWORK_SETTINGS_FIELD: networkSettings,
                             TEST_RESULTS_FIELD: NOT_AVAILABLE,
                             RAW_DEVICE_FIELD: NOT_AVAILABLE})
        
        Device.__deviceList.append(self)
    
    
    def devices():
        
        return Device.__deviceList   # devices
    
    devices = staticmethod(devices)
    
    
    def id(self):
        
        """
        id --
        
        Returns the device indentification in the form [type, serial number, name].
        """
        
        return (self[DEVICE_TYPE_FIELD], self[SERIAL_NUMBER_FIELD], self[DEVICE_NAME_FIELD])
        
        return   # Device.id
    
    
    def capabilities(self):
        
        """
        capabilities --
        
        Returns a copy of a device's configurable settings.
        """
        
        capabilities = {}
        
        for feature in self[CAPABILITIES_FIELD]:
            if rct.isSelectedSetupType(self[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
                capabilities[feature] = copy.deepcopy(self[CAPABILITIES_FIELD][feature])
        
        return capabilities   # Device.capabilities


    def emptyCapability(self, feature):
        
        """
        emptyCapability --
        
        Returns True if the specified device feature is not enabled or it is not of the currently
        selected type or any of the feature's required parameters is empty or has the value 'N/A'
        and returns False otherwise.
        """
        # Is the feature enabled?
        
        if feature not in self[CAPABILITIES_FIELD]:
            return True
        
        # Is the feature of the currently selected type?
        
        if not rct.isSelectedSetupType(self[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
            return True
        
        # Are any of the feature's required parameters undefined?
        
        if (not self[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD] or
                self[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD] == NOT_AVAILABLE):
            return True
        
        if (not self[CAPABILITIES_FIELD][feature][USER_NAME_FIELD] or
                self[CAPABILITIES_FIELD][feature][USER_NAME_FIELD] == NOT_AVAILABLE):
            return True
        
        if (not self[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD] or
                self[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD] == NOT_AVAILABLE):
            return True
        
        if rct.hasMode(COPY_MODE) and not self[CAPABILITIES_FIELD][feature].get(NETWORK_PROTOCOL_FIELD):
            return True
        
        """
        TO_DO(07/13/07):
        emptyCapability: Currently the host name field is not manditory, if later it becomes
        manditory then include the following test.
        END_TO_DO
        
        if (not self[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] or
                self[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] == NOT_AVAILABLE):
            return True
        """
        
        if not ((self[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD] == rct.HOST_IP_ADDRESS) and
                (self[CAPABILITIES_FIELD][feature][SHARED_RESOURCE_FIELD] and
                 self[CAPABILITIES_FIELD][feature][SHARED_RESOURCE_FIELD] != NOT_AVAILABLE)
                 or True):
            return True
        
        return False   # Device.emptyCapability
    
    
    def validDomain(self):
        
        """
        validDomain --
        
        Returns True if the device has a Workgroup/Domain Name.
        
        TO_DO(06/18/07):
        validDomain: Test for an actual Workgroup or Domain, instead of just existence.
        END_TO_DO
        """
        
        return self[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_WORKGROUP_DOMAIN] != ''   # Device.validDomain
    
    
    def delete(self):
        
        Device.__deviceList.remove(self)
        
        return   # Device.delete


class ServiceUser(object):
    
    """
    ServiceUser --
    
    Provides a proxy for the Service User client.
    """
    
    def __init__(self, serviceType, configEnabled = False, verbose = True, deleteFiles = True):
        
        self.serviceType = serviceType
        
        self.configEnabled = configEnabled
        self.verbose = verbose
        self.deleteFiles = deleteFiles
        
        self.TEST_DIR = os.path.normcase('testdata/')
        self.SERVICE_DIR = self.makeServiceDirectory()
        
        self.SERVICE_LIST_REQUEST_FILE = os.path.join(self.SERVICE_DIR, 'ServiceListRequest.txt')
        self.SERVICE_LIST_RESPONSE_FILE = os.path.join(self.SERVICE_DIR, 'ServiceListResponse.txt')
        
        self.DEVICE_INFO_REQUEST_FILE = os.path.join(self.SERVICE_DIR, 'DeviceInfoRequest.txt')
        self.DEVICE_INFO_RESPONSE_FILE = os.path.join(self.SERVICE_DIR, 'DeviceInfoResponse.txt')
        
        self.DEVICE_NET_INFO_REQUEST_FILE = os.path.join(self.SERVICE_DIR, 'DeviceNetInfoRequest.txt')
        self.DEVICE_NET_INFO_RESPONSE_FILE = os.path.join(self.SERVICE_DIR, 'DeviceNetInfoResponse.txt')
        
        self.DEVICE_FEATURE_INFO_REQUEST_FILE = os.path.join(self.SERVICE_DIR, 'DeviceFeatureInfoRequest.txt')
        self.DEVICE_FEATURE_INFO_RESPONSE_FILE = os.path.join(self.SERVICE_DIR, 'DeviceFeatureInfoResponse.txt')
        
        self.DEVICE_FEATURE_CONNECTION_TEST_REQUEST_FILE = os.path.join(self.SERVICE_DIR, 'DeviceFeatureTestRequest.txt')
        self.DEVICE_FEATURE_CONNECTION_TEST_RESPONSE_FILE = os.path.join(self.SERVICE_DIR, 'DeviceFeatureTestResponse.txt')
        
        self.CONFIG_SCRIPT_FILE = os.path.join(self.SERVICE_DIR, 'DeviceConfigScript.txt')
        
        self.FIND_SERVICES_SCRIPT = ['CMD_SERVICE_TIMEOUT\n',
                                     'TIME_OUT=5\n',
                                     'CMD_FIND_SVC_INSTANCES\n',
                                     'SERVICE_TYPE=%s\n\n' % self.serviceType,]
    
    
    def request(self, scriptFileName, responseFileName):
        
        """
        request --
        
        Requests that the Service User send the commands in the named script to its target service.
        """
        
        cmd = ''
        result = STATUS_ERROR
        
        if self.configEnabled:
            try:
                if self.verbose:
                    cmd = 'UserExample.exe -v -f "%s"' % scriptFileName
                else:
                    cmd = 'UserExample.exe -f "%s"' % scriptFileName
                result = os.system(cmd)
                return result
            except:
                if DEBUG:
                    print 'Service User (%s) not available.' % os.name
        
        sourceFileName = os.path.join(self.TEST_DIR, os.path.basename(responseFileName))
        
        try:
            shutil.copy(sourceFileName, self.SERVICE_DIR)
            result = STATUS_OK
        except:
            if DEBUG:
                print 'Could not copy %s to %s.' % (os.path.basename(responseFileName), self.SERVICE_DIR)
            result = STATUS_ERROR
        
        return result   # ServiceUser.request
    
    
    def response(self, responseFileName, sublevel, toplevel = None):
        
        """
        response --
        
        Scans a service response file for name=value messages. It expects to find delimiting
        text, toplevel and sublevel.
        """
        
        if not responseFileName or not sublevel:
            return
            
        responses = []
        
        try:
            responseFile = open(responseFileName, 'r')
        except:
            if self.verbose:
                print 'Cannot open the \'%s\' response file.' % (toplevel or sublevel)
            return responses
        
        try:
            name, value = self.readResponse(responseFile)
            while toplevel and name.upper() != toplevel + '_BEGIN':
                if toplevel and name.upper() == toplevel + '_END':
                    self.closeFile(responseFile)
                    return responses
                name, value = self.readResponse(responseFile)
        except:
            if self.verbose:
                print 'Could not find beginning of \'%s\'.' % (toplevel or sublevel)
            self.closeFile(responseFile)
            return responses
        
        while True:
            try:
                while name.upper() != sublevel + '_BEGIN':
                    if toplevel and name.upper() == toplevel + '_END':   # Done processing document
                        self.closeFile(responseFile)
                        return responses
                    name, value = self.readResponse(responseFile)
            except:
                if self.verbose:
                    print 'Could not find beginning of \'%s\'.' % sublevel
                self.closeFile(responseFile)
                return responses
    
            response = {}
    
            try:
                name, value = self.readResponse(responseFile)
                while (name.upper() != sublevel + '_END' and
                       name.upper() != sublevel + '_BEGIN'):
                    if toplevel and name.upper() == toplevel + '_END':   # Done processing document, collect sublevel
                        break
                    if name and value:
                        if value == NULL_VALUE:
                            response[name] = ''
                        else:
                            response[name] = value
                    name, value = self.readResponse(responseFile)
            except:
                if self.verbose:
                    print 'Read failure on \'%s\' response file \'%s\'.' % (toplevel or sublevel), responseFileName
                self.closeFile(responseFile)
                return responses
            
            responses.append(response)
            
            if toplevel and name.upper() == toplevel + '_END':   # Done processing document
                break
        
        self.closeFile(responseFile)
        
        return responses   # ServiceUser.response
    
    
    def readResponse(self, responseFile):
        
        """
        readResponse --
        
        Read a (name, value) tuple from the referenced response file object.
        """
        
        param = responseFile.readline()
        
        if not param:
            closeFile(responseFile)
            return (None, None)
        
        response = self.parseParam(param)
        
        return response   # readResponse
    
    
    def makeServiceDirectory(self):
        
        """
        makeServiceDirectory --
        
        Obtains a path to the directory to be used for scripts and response files. If the files are
        being saved then this is the current working directory, otherwise it is a temporary directory.
        If a temporary directory is created then it must be removed when it is no longer needed.
        """
        
        if not self.deleteFiles:
            self.SERVICE_DIR = getRctHome()
            if DEBUG: print 'Service directory is the current directory \'%s\'.' % self.SERVICE_DIR
        else:
            try:
                self.SERVICE_DIR = tempfile.mkdtemp(prefix = PRODUCT_SHORT_NAME + '_')
                if DEBUG: print 'Created temporary service directory \'%s\'.' % self.SERVICE_DIR
            except:
                self.SERVICE_DIR = getRctHome()
                if DEBUG: print 'Could not create temporary service directory, using \'%s\'.' % self.SERVICE_DIR
        
        return self.SERVICE_DIR   # makeServiceDirectory
    
    
    def removeServiceDirectory(self):
        
        if self.deleteFiles and self.SERVICE_DIR != getRctHome():
            try:
                shutil.rmtree(self.SERVICE_DIR)
                if DEBUG:
                    print 'Removed temporary service directory \'%s\'.' % self.SERVICE_DIR
            except:
                if DEBUG:
                    print 'Error removing temporary service directory \'%s\'.' % self.SERVICE_DIR
    
    
    def closeFile(self, responseFile):
        
        """
        closeFile --
        
        Close and optionally delete a response file.
        """
        
        responseFile.close()
        
        if self.deleteFiles:
            os.remove(responseFile.name)
    
    
    def parseParam(self, param):
        
        """
        parseParam --
        
        Returns a tuple containing a (name, value) configuration item.
        It handles complete, partial, and empty items.
        """
        
        name = ''
        value = ''
        
        paramList = param.replace(ALTERNATE_DELIMITER, '=').split('=')
        
        try:
            name = paramList[0].strip()
            value = paramList[1].strip()
            if name and not value:
                return (name, NULL_VALUE)
            else:
                return (name, value)
        except:
             return (name, '')   # ServiceUser.parseParam


    def runScript(self, script):

        """
        runScript --

        Runs a Service User configuration script, creating and deleting the script file. 
        """

        result = STATUS_ERROR
        
        # Create the configuration script file.

        try:
            scriptFile = open(self.CONFIG_SCRIPT_FILE, 'w')
            scriptFile.writelines(script)
            scriptFile.close()
        except:
            if self.verbose:
                print 'Cannot create the configuration script, \'%s\'.' % self.CONFIG_SCRIPT_FILE
            return result

        # Send the configuration script.

        if not self.configEnabled:
            result = STATUS_OK
        else:
            try:
                if self.verbose:
                    cmd = 'UserExample.exe -v -f "%s"' % self.CONFIG_SCRIPT_FILE
                else:
                    cmd = 'UserExample.exe -f "%s"' % self.CONFIG_SCRIPT_FILE
                result = os.system(cmd)
            except:
                print 'RCT cannot yet configure devices on \'%s\' systems.' % os.name

        # Delete the configuration script file.

        if self.deleteFiles:
            try:
                os.remove(self.CONFIG_SCRIPT_FILE)
            except OSError:
                pass

        return result   # ServiceUser.runScript
    
    
    def getServices(self):
        
        """
        getServices --
        
        Query for available services.
        """
        
        # Create the request services script.
        
        script = ['CMD_FIND_SVC_INSTANCES\n',
                  'SERVICE_TYPE=%s\n' % self.serviceType,
                  'CMD_RESOLVE_SVC_INSTANCES\n',
                  'CMD_FILE_OUTPUT_START\n',
                  'FILE_NAME=%s\n' % self.SERVICE_LIST_RESPONSE_FILE,
                  'DATA=SERVICE_LIST\n',
                  'CMD_FILE_OUTPUT_END\n']
        
        try:
            requestFile = open(self.SERVICE_LIST_REQUEST_FILE, 'w')
            requestFile.writelines(script)
            requestFile.close()
        except:
            if self.verbose:
                print 'Cannot create the request services script, \'%s\'.' % self.SERVICE_LIST_REQUEST_FILE
            return
        
        # Execute the request services script.
        
        status = self.request(self.SERVICE_LIST_REQUEST_FILE, self.SERVICE_LIST_RESPONSE_FILE)
        
        # Delete the request services script file.
        
        if self.deleteFiles:
            try:
                os.remove(self.SERVICE_LIST_REQUEST_FILE)
            except OSError:
                pass
        
        # Check for execution errors.
        
        if status:
            if self.verbose:
                print 'Could not execute the request services script, return code %d.' % status
            return
        
        # Collect the list of available services.
        
        services = self.response(self.SERVICE_LIST_RESPONSE_FILE,
                                 'SERVICE_INSTANCE', 'SERVICE_INSTANCE_LIST')
        
        return services   # ServiceUser.getServices
    
    
    def getDeviceInfo(self, serviceNames):

        """
        getDeviceInfo --
        
        Query for device information from the named service.
        """
        
        # Create the request device information script.
        
        beginScript = ['CMD_FILE_OUTPUT_START\n',
                       'FILE_NAME=%s\n' % self.DEVICE_INFO_RESPONSE_FILE,
                       'TEXT=DEVICE_INFO_BEGIN\n',
                       'CMD_FILE_OUTPUT_END\n',
                       'CMD_ON_ERROR\n',
                       'EXECUTE_NEXT_BLOCK=TRUE\n']
            
        closeConnectionScript = ['CMD_CLOSE_CONN\n',
                                 'CMD_END_BLOCK\n']
        
        endScript = ['CMD_FILE_OUTPUT_APPEND\n',
                     'FILE_NAME=%s\n' % self.DEVICE_INFO_RESPONSE_FILE,
                     'TEXT=DEVICE_INFO_END\n',
                     'CMD_FILE_OUTPUT_END\n']
                     
        script = beginScript
        
        for serviceName in serviceNames:
            openConnectionScript = ['CMD_BEGIN_BLOCK\n',
                                    'CMD_CLOSE_CONN\n',
                                    'CMD_OPEN_CONN\n',
                                    'SERVICE_NAME=%s\n\n' % serviceName]
            
            deviceInfoScript = ['CMD_SEND_MSG\n',
                                'MSG=BEGIN_DEVICE_INFO_CONF\n',
                                'CMD_RECV_MSG\n',
                                'CMD_SEND_MSG\n',
                                'MSG=CONF_GET\n',
                                'CMD_RECV_MSG\n',
                                'CMD_FILE_OUTPUT_APPEND\n',
                                'FILE_NAME=%s\n' % self.DEVICE_INFO_RESPONSE_FILE,
                                'TEXT=DEVICE_INFO_INSTANCE_BEGIN\n',
                                'TEXT=SERVICE_NAME=%s\n' % serviceName,
                                'DATA=LAST_SERVER_RESP\n',
                                'TEXT=DEVICE_INFO_INSTANCE_END\n',
                                'CMD_FILE_OUTPUT_END\n',
                                'CMD_SEND_MSG\n',
                                'MSG=END_DEVICE_INFO_CONF\n',
                                'CMD_RECV_MSG\n\n']
            
            deviceNetInfoScript = ['CMD_SEND_MSG\n',
                                   'MSG=BEGIN_DEVICE_NETWORK_CONF\n',
                                   'CMD_RECV_MSG\n',
                                   'CMD_SEND_MSG\n',
                                   'MSG=CONF_GET\n',
                                   'CMD_RECV_MSG\n',
                                   'CMD_FILE_OUTPUT_APPEND\n',
                                   'FILE_NAME=%s\n' % self.DEVICE_INFO_RESPONSE_FILE,
                                   'TEXT=DEVICE_INFO_INSTANCE_BEGIN\n',
                                   'TEXT=SERVICE_NAME=%s\n' % serviceName,
                                   'DATA=LAST_SERVER_RESP\n',
                                   'TEXT=DEVICE_INFO_INSTANCE_END\n',
                                   'CMD_FILE_OUTPUT_END\n',
                                   'CMD_SEND_MSG\n',
                                   'MSG=END_DEVICE_NETWORK_CONF\n',
                                   'CMD_RECV_MSG\n']
            
            script.extend(openConnectionScript)
            script.extend(deviceInfoScript)
            script.extend(deviceNetInfoScript)
            script.extend(closeConnectionScript)
        
        script.extend(endScript)
        
        try:
            requestFile = open(self.DEVICE_INFO_REQUEST_FILE, 'w')
            requestFile.writelines(self.FIND_SERVICES_SCRIPT)
            requestFile.writelines(script)
            requestFile.close()
        except:
            if self.verbose:
                print 'Cannot create the device information script, \'%s\'.' % requestFileName
            return
        
        # Execute the request device information script.
        
        status = self.request(self.DEVICE_INFO_REQUEST_FILE, self.DEVICE_INFO_RESPONSE_FILE)
        
        if status:
            if self.verbose:
                print 'Could not execute the device information script, return code %d.' % status
            if self.deleteFiles:
                try:
                    os.remove(self.DEVICE_INFO_REQUEST_FILE)
                except OSError:
                    pass
            return
        
        # Delete the request device information script file.
        
        if self.deleteFiles:
            try:
                os.remove(self.DEVICE_INFO_REQUEST_FILE)
            except OSError:
                pass
        
        # Collect the device information.
        
        deviceInfo = self.response(self.DEVICE_INFO_RESPONSE_FILE, 'DEVICE_INFO_INSTANCE', 'DEVICE_INFO')
        
        return deviceInfo   # ServiceUser.getDeviceInfo
    
    
    def getDeviceNetInfo(self, serviceName):
        
        """
        getDeviceNetInfo --
        
        Query for the device network settings from the named service.
        """
        
        # Create the request device network configuration script.
        
        script = ['CMD_CLOSE_CONN\n',
                  'CMD_OPEN_CONN\n',
                  'SERVICE_NAME=%s\n' % serviceName,
                  'CMD_SEND_MSG\n',
                  'MSG=BEGIN_DEVICE_NETWORK_CONF\n',
                  'CMD_RECV_MSG\n',
                  'CMD_SEND_MSG\n',
                  'MSG=CONF_GET\n',
                  'CMD_RECV_MSG\n',
                  'CMD_FILE_OUTPUT_START\n',
                  'FILE_NAME=%s\n' % self.DEVICE_NET_INFO_RESPONSE_FILE,
                  'TEXT=DEVICE_NETWORK_INFO_BEGIN\n',
                  'TEXT=DEVICE_INSTANCE_BEGIN\n',
                  'DATA=LAST_SERVER_RESP\n',
                  'TEXT=DEVICE_INSTANCE_END\n',
                  'TEXT=DEVICE_NETWORK_INFO_END\n',
                  'CMD_FILE_OUTPUT_END\n',
                  'CMD_SEND_MSG\n',
                  'MSG=END_DEVICE_NETWORK_CONF\n',
                  'CMD_RECV_MSG\n',
                  'CMD_CLOSE_CONN\n']
        
        try:
            requestFile = open(self.DEVICE_NET_INFO_REQUEST_FILE, 'a')
            requestFile.writelines(self.FIND_SERVICES_SCRIPT)
            requestFile.writelines(script)
            requestFile.close()
        except:
            if self.verbose:
                print 'Cannot create the device network information script, \'%s\'.' % self.DEVICE_NET_INFO_REQUEST_FILE
            return
        
        # Execute the request device network information script.
        
        status = self.request(self.DEVICE_NET_INFO_REQUEST_FILE, self.DEVICE_NET_INFO_RESPONSE_FILE)
        if status:
            if self.verbose:
                print 'Could not execute the device network information script, return code %d.' % status
            if self.deleteFiles:
                try:
                    os.remove(self.DEVICE_NET_INFO_REQUEST_FILE)
                except OSError:
                    pass
            return
        
        # Delete the request device network information script file.
        
        if self.deleteFiles:
            try:
                os.remove(self.DEVICE_NET_INFO_REQUEST_FILE)
            except OSError:
                pass
        
        # Collect the device information.
        
        deviceNetInfo = self.response(self.DEVICE_NET_INFO_RESPONSE_FILE, 'DEVICE_INSTANCE', 'DEVICE_NETWORK_INFO')
        
        return deviceNetInfo   # ServiceUser.getDeviceNetInfo
    
    
    def getDeviceFeatureInfo(self, serviceFeatures):
        
        """
        getDeviceFeatureInfo --
    
        Request a configurable feature from the named service.
        """
        
        # Create the device feature script.
        
        beginScript = ['CMD_FILE_OUTPUT_START\n',
                       'FILE_NAME=%s\n' % self.DEVICE_FEATURE_INFO_RESPONSE_FILE,
                       'TEXT=DEVICE_CONF_BEGIN\n',
                       'CMD_FILE_OUTPUT_END\n\n']
        
        closeConnectionScript = ['CMD_CLOSE_CONN\n\n']
        
        endScript = ['CMD_FILE_OUTPUT_APPEND\n',
                     'FILE_NAME=%s\n' % self.DEVICE_FEATURE_INFO_RESPONSE_FILE,
                     'TEXT=DEVICE_CONF_END\n',
                     'CMD_FILE_OUTPUT_END\n\n']
        
        script = beginScript
        
        for serviceName in serviceFeatures:
            openConnectionScript = ['CMD_CLOSE_CONN\n',
                                    'CMD_OPEN_CONN\n',
                                    'SERVICE_NAME=%s\n\n' % serviceName]
            script.extend(openConnectionScript)
            for feature in serviceFeatures[serviceName]:
                requestConfigScript = ['CMD_SEND_MSG\n',
                                       'MSG=BEGIN_%s\n' % featureString(feature),
                                       'CMD_RECV_MSG\n',
                                       'CMD_SEND_MSG\n',
                                       'MSG=CONF_GET\n',
                                       'CMD_RECV_MSG\n',
                                       'CMD_FILE_OUTPUT_APPEND\n',
                                       'FILE_NAME=%s\n' % self.DEVICE_FEATURE_INFO_RESPONSE_FILE,
                                       #'TEXT=%s_BEGIN\n' % featureString(feature),
                                       'TEXT=DEVICE_CONF_INSTANCE_BEGIN\n',
                                       'TEXT=SERVICE_NAME=%s\n' % serviceName,
                                       'TEXT=FEATURE=%s\n' % feature,
                                       'DATA=LAST_SERVER_RESP\n',
                                       'TEXT=DEVICE_CONF_INSTANCE_END\n',
                                       #'TEXT=%s_END\n\n' % featureString(feature),
                                       'CMD_FILE_OUTPUT_END\n',
                                       'CMD_SEND_MSG\n',
                                       'MSG=END_%s\n' % featureString(feature),
                                       'CMD_RECV_MSG\n\n']
                script.extend(requestConfigScript)
            script.extend(closeConnectionScript)
        
        script.extend(endScript)
    
        try:
            requestFile = open(self.DEVICE_FEATURE_INFO_REQUEST_FILE, 'w')
            requestFile.writelines(self.FIND_SERVICES_SCRIPT)
            requestFile.writelines(script)
            requestFile.close()
        except:
            if self.verbose:
                print 'Cannot create the device feature script, \'%s\'.' % self.DEVICE_FEATURE_INFO_REQUEST_FILE
            return
    
        # Execute the device feature script.
        
        status = self.request(self.DEVICE_FEATURE_INFO_REQUEST_FILE, self.DEVICE_FEATURE_INFO_RESPONSE_FILE)
        if status:
            if self.verbose:
                print 'Could not execute the device feature script, return code %d.' % status
            if self.deleteFiles:
                try:
                    os.remove(self.DEVICE_FEATURE_INFO_REQUEST_FILE)
                except OSError:
                    pass
            return
        
        # Delete the device feature script file.
        
        if self.deleteFiles:
            try:
                os.remove(self.DEVICE_FEATURE_INFO_REQUEST_FILE)
            except OSError:
                pass
    
        # Collect the device feature settings.
    
        deviceFeatureInfo  = self.response(self.DEVICE_FEATURE_INFO_RESPONSE_FILE,
                                           'DEVICE_CONF_INSTANCE', 'DEVICE_CONF')
        
        return deviceFeatureInfo   # ServiceUser.getDeviceFeatureInfo
    
    
    def testDeviceFeature(self, serviceName, feature):
        
        """
        testDeviceFeature --
    
        Request that a service test the specified device feature.
        """
        
        # Create the device feature test script.
    
        script = ['CMD_SERVICE_TIMEOUT\n',
                  'TIME_OUT=1000\n',         # TEMPORARY, should be 0 for indefinite delay.
                  'CMD_CLOSE_CONN\n',
                  'CMD_OPEN_CONN\n',
                  'SERVICE_NAME=%s\n' % serviceName,
                  'CMD_SEND_MSG\n',
                  'MSG=BEGIN_%s\n' % featureString(feature),
                  'CMD_RECV_MSG\n',
                  'CMD_SEND_MSG\n',
                  'MSG=PERFORM_CONNECTION_TEST\n',
                  'CMD_RECV_MSG\n',
                  'CMD_FILE_OUTPUT_START\n',
                  'FILE_NAME=%s\n' % self.DEVICE_FEATURE_CONNECTION_TEST_RESPONSE_FILE,
                  'TEXT=%s_BEGIN\n' % featureString(feature),
                  'TEXT=DEVICE_INSTANCE_BEGIN\n',
                  'DATA=LAST_SERVER_RESP\n',
                  'TEXT=DEVICE_INSTANCE_END\n',
                  'TEXT=%s_END\n' % featureString(feature),
                  'CMD_FILE_OUTPUT_END\n',
                  'CMD_SEND_MSG\n',
                  'MSG=END_%s\n' % featureString(feature),
                  'CMD_RECV_MSG\n',
                  'CMD_CLOSE_CONN\n']
    
        try:
            requestFile = open(self.DEVICE_FEATURE_CONNECTION_TEST_REQUEST_FILE, 'w')
            requestFile.writelines(self.FIND_SERVICES_SCRIPT)
            requestFile.writelines(script)
            requestFile.close()
        except:
            if self.verbose:
                print 'Cannot create the device test script, \'%s\'.' % self.DEVICE_FEATURE_CONNECTION_TEST_REQUEST_FILE
            return []
        
        # Execute the device feature script.
        
        status = self.request(self.DEVICE_FEATURE_CONNECTION_TEST_REQUEST_FILE,
                              self.DEVICE_FEATURE_CONNECTION_TEST_RESPONSE_FILE)
        
        if status:
            if self.verbose:
                print 'Could not execute the device test script, return code %d.' % status
            if self.deleteFiles:
                try:
                    os.remove(self.DEVICE_FEATURE_INFO_REQUEST_FILE)
                except OSError:
                    pass
            return []
        
        # Delete the device feature script file.
        
        if self.deleteFiles:
            try:
                os.remove(self.DEVICE_FEATURE_CONNECTION_TEST_REQUEST_FILE)
            except OSError:
                pass
        
        # Collect the test results.
        
        testResults = self.response(self.DEVICE_FEATURE_CONNECTION_TEST_RESPONSE_FILE,
                                    'DEVICE_INSTANCE', featureString(feature))
        
        return testResults   # ServiceUser.testDeviceFeature


class RctWizard(wx.wizard.Wizard):
    
    def __init__(self, title):
        self.pages = []
        wx.wizard.Wizard.__init__(self, None, wx.ID_ANY, title)
    
    def addPage(self, page):
        if self.pages:
            previousPage = self.pages[-1]
            page.SetPrev(previousPage)
            previousPage.SetNext(page)
        self.pages.append(page)
    
    def pageNumber(self, page):
        try:
            return self.pages.index(page)
        except:
            return
    
    def run(self):
        return self.RunWizard(self.pages[0])


class TitledPage(wx.wizard.PyWizardPage):
    
    def __init__(self, parent, title, prevPage = None, nextPage = None, useTitleWindow = True):
        
        self.prev = None
        self.next = None
        self.sizer = None
        self.prevPage = prevPage
        self.nextPage = nextPage
        self.banner = None
        self.popupMenu = None
        
        wx.wizard.PyWizardPage.__init__(self, parent)
        
        parent.SetBackgroundColour(RCT_DARK_BG)
        self.SetBackgroundColour(RCT_LT_BG)
        
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        
        try:
            logo = rct_properties['RCT_LOGO']
        except:
            logo = LOGO_IMAGE_FILE
        
        if useTitleWindow:
            self.banner = titleWindow(self, wx.ID_ANY, logo, title, size = (500, 63))
            self.sizer.Add(self.banner, 0, wx.EXPAND | wx.ALL)
        else:
            try:
                logoImage = wx.Image(logo, wx.BITMAP_TYPE_ANY)
                companyLogo = wx.StaticBitmap(self, wx.ID_ANY, bitmap = logoImage.ConvertToBitmap())
            except:
                companyLogo = wx.StaticText(self, wx.ID_ANY, '[%s\nLogo]' % COMPANY_SHORT_NAME)
                companyLogo.SetFont(wx.Font(18, wx.SWISS, wx.NORMAL, wx.BOLD))

            companyLogoSizer = wx.BoxSizer(wx.VERTICAL)
            companyLogoSizer.Add(companyLogo, 0, wx.ALIGN_CENTRE | wx.ALL, 2)

            titleText = wx.StaticText(self, -1, title)
            titleText.SetFont(wx.Font(18, wx.SWISS, wx.ITALIC, wx.BOLD))
            titleTextSizer = wx.BoxSizer(wx.VERTICAL)
            titleTextSizer.Add(titleText, 0, wx.ALIGN_CENTRE | wx.ALL, 10)

            titleSizer = wx.FlexGridSizer(cols = 2, hgap = 0, vgap = 0)
            titleSizer.AddMany([companyLogoSizer, titleTextSizer])

            self.sizer.Add(titleSizer, 0, wx.ALIGN_LEFT | wx.ALL, PADDING)
            self.sizer.Add(wx.StaticLine(self, wx.ID_ANY), 0, wx.EXPAND | wx.ALL, PADDING)
        
        self.SetSizer(self.sizer)
    
    
    def addPopupMenuItems(self, menuItems):
        
        if not self.popupMenu:
            self.popupMenu = wx.Menu()
            self.Bind(wx.EVT_CONTEXT_MENU, self.onShowPopup)
        
        for menuItem in menuItems:
            item = self.popupMenu.Append(-1, menuItem['text'])
            self.Bind(wx.EVT_MENU, menuItem['action'], item)
    
    
    def onShowPopup(self, event):
        
        pos = self.ScreenToClient(event.GetPosition())
        self.PopupMenu(self.popupMenu, pos)

    
    def addItem(self, item, factor = 0):
        self.sizer.Add(item, factor, wx.EXPAND | wx.ALL, PADDING)
        return
    
    
    def SetPrev(self, page):
        self.prev = page
        return
    
    
    def SetNext(self, page):
        self.next = page
        return
    
    
    def GetPrev(self):
        if self.prevPage:
            page = self.prevPage(self)
            if page:
                self.prev = page
        return self.prev
    
    
    def GetNext(self):
        if self.nextPage:
            page = self.nextPage(self)
            if page:
                self.next = page
        return self.next


class InstrumentList(wx.ListCtrl):
    
    """
    InstrumentList --
    
    A ListCtrl specializing in the display of devices and their configurable features.
    """
    
    def __init__(self, parent,
                 heading = [NOT_AVAILABLE, NOT_AVAILABLE, NOT_AVAILABLE],
                 numberOfEntries = 50, size = (LIST_WIDTH, LIST_HEIGHT)):
        
        self.headings = heading
        
        self.MAX_INSTRUMENTS = numberOfEntries
        self.MAX_ATTRIBUTES = len(self.headings)
        
        wx.ListCtrl.__init__(self, parent, wx.ID_ANY, size = size, style = wx.LC_REPORT)
        
        for column in range(self.MAX_ATTRIBUTES):
            self.InsertColumn(column, fieldToString(self.headings[column]),
                      wx.LIST_FORMAT_LEFT)
        
        for row in range(self.MAX_INSTRUMENTS):
            index = self.InsertStringItem(sys.maxint, '')
            self.SetStringItem(index, 1, '')
    
    def clear(self):
        
        """
        InstrumentList.clear --
        
        Removes the current selection and sets all items to the empty string.
        """
        
        self.clearSelection()
        
        for row in range(self.MAX_INSTRUMENTS):
            for column in range(self.GetColumnCount()):
                self.SetStringItem(row, column, '')
        
        return   # InstrumentList.clear
    
    
    def clearSelection(self):
        
        """
        InstrumentList.clearSelection --
        
        De-select the currently selected items.
        """
        
        index = self.GetFirstSelected()
        while index != -1:
            self.Select(index, False)
            index = self.GetNextSelected(index)
        
        return   #   InstrumentList.clearSelection
    
    
    def selectFirst(self):
        
        """
        InstrumentList.selectFirst --
        
        Selects the first non-empty entry.
        """
        
        for index in range(self.GetItemCount()):
            item = self.GetItem(index)
            for column in range(self.GetColumnCount()):
                if self.GetItem(index, column).GetText() != '':
                    self.Select(index, True)
                    return
        
        return   # InstrumentList.selectFirst
    
    
    def selectOne(self, selectedInstrument, n):
        
        """
        InstrumentList.selectOne --
        
        De-selects all but the entry for the specified instrument. List entries are
        compared on their nth column.
        """
        
        if self.GetSelectedItemCount() < 1 or n >= self.GetColumnCount():
            return
        
        index = self.GetFirstSelected()
        while index != -1:
            if self.GetItem(index, n).GetText().lower() == selectedInstrument[n].lower():
                self.Select(index, True)
            else:
                self.Select(index, False)
            index = self.GetNextSelected(index)
        
        return   # InstrumentList.selectOne
    
    
    def selectAll(self):
        
        """
        InstrumentList.selectAll --
        
        Selects all non-empty items.
        """
        
        for index in range(self.GetItemCount()):
            item = self.GetItem(index)
            for column in range(self.GetColumnCount()):
                if self.GetItem(index, column).GetText() != '':
                    self.Select(index, True)
                    continue
        
        return   # InstrumentList.selectAll
    
    
    def getFirstSelection(self):
        
        """
        InstrumentList.getFirstSelection --
        
        Returns the list of all items in the first selected row.
        """
        
        index = self.GetFirstSelected()
        if index < 0:
            return []
        
        selection = []
        for column in range(self.GetColumnCount()):
            selection.append(self.GetItem(index, column).GetText())
        
        return selection   # InstrumentList.getFirstSelection
    
    
    def obscurePassword(self, item, fieldName):
        
        """
        InstrumentList.obscurePassword --
        
        Maps password entries to booleans indicating whether or not the password
        string was non-empty.
        """
        
        if fieldName == PASSWORD_FIELD:
            if item:
                item = True
            else:
                item = False
        
        return item   # InstrumentList.obscurePassword
    
    
    def cleanup(self, data):
        
        """
        InstrumentList.cleanup --
        
        Maps special values to appropriate strings.
        """
        
        if data == None:
            data = ''
        elif data == True:
            data = _('Yes')
        elif data == False:
            data = _('No')
        
        return data   # InstrumentList.cleanup
    
    
    def displayAllInstruments(self):
        
        """
        InstrumentList.displayAllInstruments --
        
        List all responding instruments of the currently selected type.
        """
        
        # Obtain the list of instruments of the currently selected type.
        
        instrumentList = getInstrumentList(getInstrumentTypes())
        self.clear()
        
        # Display each instrument.
        
        for row in range(len(instrumentList)):
            for column in range(self.GetColumnCount()):
                self.SetStringItem(row, column,
                                   instrumentList[row][self.headings[column]])
        
        self.setColumnWidths()
        
        return   # InstrumentList.displayAllInstruments
    
    
    def displaySelectedInstruments(self, includedInstruments):
        
        """
        InstrumentList.displaySelectedInstruments --
        
        Display those instruments specified in a list. Each instrument is represented as a list
        of the form [type, serial number, name]. Comparisons are made on serial number.
        """
        
        # Obtain the list of devices whose serial number is one of those specified.
        
        includedSerialNumbers = [instrument[1] for instrument in includedInstruments]
        instrumentList = [device for device in Device.devices()
                              if device[SERIAL_NUMBER_FIELD] in includedSerialNumbers]
        self.clear()
        
        # Display each instrument.
        
        for row in range(len(instrumentList)):
            for column in range(self.GetColumnCount()):
                self.SetStringItem(row, column,
                                   fieldToString(instrumentList[row][self.headings[column]]))
        
        self.setColumnWidths()
        
        return   # InstrumentList.displaySelectedInstruments
    
    
    def displayOtherInstruments(self, excludedInstruments):
        
        """
        InstrumentList.displayOtherInstruments --
        
        List instruments excluding those specified in a list of excluded instruments.
        Each excluded instrument is represented as a list of the form [type, serial number, name].
        Comparisons are made on serial number.
        """
        
        # Obtain the list of devices whose serial numbers are not any of those specified.
        
        excludedSerialNumbers = [instrument[1] for instrument in excludedInstruments]
        instrumentList = [device for device in Device.devices() if device[SERIAL_NUMBER_FIELD] not in excludedSerialNumbers]
        
        self.clear()
        
        # Display each instrument.
        
        for row in range(len(instrumentList)):
            for column in range(self.GetColumnCount()):
                self.SetStringItem(row, column,
                                   instrumentList[row][self.headings[column]])
        
        self.setColumnWidths()
        
        return   # InstrumentList.displayOtherInstruments
    
    
    def setColumnWidths(self):
        
        """
        InstrumentList.setColumnWidths --
        
        Computes column widths by dividing the available column space equally.
        """
        
        controlWidth = rct.wizard.GetPageSize()[0] - ((isWindows()) and 41 or 16)
        
        columnWidth = controlWidth / self.GetColumnCount()
        for column in range(self.GetColumnCount()):
            self.SetColumnWidth(column, columnWidth)
        
        return columnWidth   # InstrumentList.setColumnWidths
    
    
    def displayCapabilities(self, instrument):
        
        """
        InstrumentList.displayCapabilities --
        
        List the capabilities of the specified instrument.
        """
        
        # Obtain the list of instruments of the currently selected type.
        
        instrumentList = getInstrumentList(getInstrumentTypes())
        self.clear()
        
        # Retrieve this instrument's capabilities, a dictionary of configurable features.
        
        capabilities = {}
        for device in instrumentList:
            if device[SERIAL_NUMBER_FIELD].lower() == instrument[1].lower():
                capabilities = copy.deepcopy(device[CAPABILITIES_FIELD])
                features = capabilities.keys()
                for feature in features:
                    if not rct.isSelectedSetupType(capabilities[feature][SETUP_TYPE_FIELD]):
                        del capabilities[feature]
                
        # Return if there are no capabilities to display.
        
        if not capabilities:
            return capabilities
        
        # Display the setup for each feature
        
        for (row, feature) in zip(range(self.MAX_INSTRUMENTS), capabilities):
            for column in range(self.GetColumnCount()):
                if column == 0:
                    self.SetStringItem(row, 0, fieldToString(feature))
                else:
                    if not capabilities[feature]:
                        entry = None
                    else:
                        entry = self.obscurePassword(capabilities[feature][self.headings[column]],
                                                     self.headings[column])
                    self.SetStringItem(row, column, self.cleanup(entry))
        
        self.setColumnWidths()
        
        return capabilities   # InstrumentList.displayCapabilities


class Selector(object):
    
    """
    Selector --
    
    An adapter class that takes a sequence or ListCtrl object and creates an iterator whose next()
    method dispenses the object's elements. It also includes a hasMore() method to test for more
    selections without calling next() and a current() method for obtaining the last value returned
    by next().
    """
    
    def __init__(self, expr):
        
        self.selectionList = []
        self.currentSelection = None
        
        try:
            self.selectionList = [item for item in expr if not emptySelection(item)]
        except:
            if expr and expr.GetSelectedItemCount() > 0:
                index = expr.GetFirstSelected()
                while (index != -1):
                    entry = []
                    for column in range(expr.GetColumnCount()):
                        entry.append(expr.GetItem(index, column).GetText())
                    if not emptySelection(entry):
                        self.selectionList.append(entry)
                    index = expr.GetNextSelected(index)
    
    def __iter__(self):
        
        return self
    
    def hasMore(self):
        
        """
        hasMore --
        
        Returns True if there are more elements remaining in the Selector, returns False otherwise.
        """
        
        try:
            expr = self.selectionList[0]
            return True
        except:
            return False   # Selector.hasMore
    
    def next(self):
        
        """
        next --
        
        Returns the next item in the Selector or rasises StopIteration exception if there is none
        available.
        """
        
        try:
            self.currentSelection = self.selectionList.pop(0)
            return self.currentSelection
        except:
            raise StopIteration   # Selector.next
    
    def current(self):
        
        """
        current --
        
        Returns the current Selector item without removing it from the Selector.
        """
        
        return self.currentSelection    # Selector.current


class PrinterList(wx.ListCtrl):
    
    """
    PrinterList --
    
    A ListCtrl specializing in the display of printers.
    """
    
    __attributes = ['pPrinterName', 'pShareName']
    
    def __init__(self, parent,
                 heading = [NOT_AVAILABLE, NOT_AVAILABLE],
                 numberOfEntries = 50, size = (LIST_WIDTH, LIST_HEIGHT)):
        
        self.headings = heading
        
        self.MAX_PRINTERS = numberOfEntries
        self.MAX_ATTRIBUTES = len(self.headings)
        
        wx.ListCtrl.__init__(self, parent, wx.ID_ANY, size = size, style = wx.LC_REPORT)
        
        for column in range(self.MAX_ATTRIBUTES):
            self.InsertColumn(column, self.headings[column],
                      wx.LIST_FORMAT_LEFT)
        
        for row in range(self.MAX_PRINTERS):
            index = self.InsertStringItem(sys.maxint, '')
            self.SetStringItem(index, 1, '')
    
    
    def clear(self):
        
        """
        PrinterList.clear --
        
        Removes the current selection and sets all items to the empty string.
        """
        
        self.clearSelection()
        
        for row in range(self.MAX_PRINTERS):
            for column in range(self.GetColumnCount()):
                self.SetStringItem(row, column, '')
        
        return   # PrinterList.clear
    
    
    def clearSelection(self):
        
        """
        PrinterList.clearSelection --
        
        De-select the currently selected items.
        """
        
        index = self.GetFirstSelected()
        while index != -1:
            self.Select(index, False)
            index = self.GetNextSelected(index)
        
        return   #   PrinterList.clearSelection
    
    
    def selectFirst(self):
        
        """
        PrinterList.selectFirst --
        
        Selects the first non-empty entry.
        """
        
        for index in range(self.GetItemCount()):
            item = self.GetItem(index)
            for column in range(self.GetColumnCount()):
                if self.GetItem(index, column).GetText() != '':
                    self.Select(index, True)
                    return
        
        return   # PrinterList.selectFirst
    
    
    def selectOne(self, selectedPrinter, n):
        
        """
        PrinterList.selectOne --
        
        De-selects all but the entry for the specified printer. List entries are
        compared on their nth column.
        """
        
        if self.GetSelectedItemCount() < 1 or n >= self.GetColumnCount():
            return
        
        index = self.GetFirstSelected()
        while index != -1:
            entry = []
            for column in range(self.GetColumnCount()):
                entry.append(self.GetItem(index, column).GetText())
            if entry[n].lower() == selectedPrinter.lower():
                self.Select(index, True)
            else:
                self.Select(index, False)
            index = self.GetNextSelected(index)
        
        return   # PrinterList.selectOne
    
    
    def getFirstSelection(self):
        
        """
        PrinterList.getFirstSelection --
        
        Returns the list of all items in the first selected row.
        """
        
        index = self.GetFirstSelected()
        if index < 0:
            return []
        
        selection = []
        for column in range(self.GetColumnCount()):
            selection.append(self.GetItem(index, column).GetText())
        
        return selection   # PrinterList.getFirstSelection
    
    
    def setColumnWidths(self):
        
        """
        PrinterList.setColumnWidths --
        
        Computes column widths by dividing the available column space equally.
        """

        controlWidth = rct.wizard.GetPageSize()[0] - ((isWindows()) and 41 or 16)
        
        columnWidth = controlWidth / self.GetColumnCount()
        for column in range(self.GetColumnCount()):
            self.SetColumnWidth(column, columnWidth)

        return columnWidth   # PrinterList.setColumnWidths
    
    
    def displayPrinters(self, printers):
        
        """
        PrinterList.displayPrinters --
        
        Display the list of printers including descriptive name and share name.
        """
        
        if not printers or printers == NOT_AVAILABLE:
            return printers
        
        self.clear()
        
        # Display each printer.
        
        for row in range(len(printers)):
            for column in range(self.GetColumnCount()):
                attribute = printers[row][PrinterList.__attributes[column]]
                if not attribute:
                    self.SetStringItem(row, column, '')
                else:
                    self.SetStringItem(row, column, attribute)
        
        self.setColumnWidths()
        
        return printers   # PrinterList.displayPrinters


def getPrinters():
    
    """
    getPrinters --
    
    Obtain the list of printer descriptions, each as a dictionary of printer attributes.
    """
    
    if not isWindows():
        return NOT_AVAILABLE
    
    try:
        return [printer for printer in
                    win32print.EnumPrinters(win32print.PRINTER_ENUM_NAME, None, 2)
                    if not excludePrinter(printer)]
    except:
        return []   # getPrinters


def excludePrinter(printer):
    
    """
    excludePrinter --
    
    Returns True if the specified printer's name is included in a file of excluded printers and
    returns False otherwise. The printer parameter is a settings dictionary returned by the
    win32print.EnumPrinters function.
    """
    
    try:
        excludedPrinterFile = open(EXCLUDED_PRINTERS_FILE, mode = 'r')
        excludedPrinters = [excludedPrinter.strip().lower() for excludedPrinter in excludedPrinterFile]
        excludedPrinterFile.close()
        return printer['pPrinterName'].strip().lower() in excludedPrinters
    except:
        return False   # excludePrinter

NO_PAGE = -2
USE_PAGE = -1
START_PAGE = 0
CUSTOM_PASSWORD_PAGE = 1
SETUP_TYPE_PAGE = 2
DEVICE_SELECT_PAGE = 3
FEATURE_SELECT_PAGE = 4
FOLDER_SELECT_PAGE = 5
USER_NAME_PAGE = 6
COPY_SOURCE_PAGE = 7
COPY_DEST_PAGE = 8
REPORT_DEVICE_SELECT_PAGE = 9
REPORT_NETWORKING_PAGE = 10
REPORT_FEATURE_PAGE = 11
COMMIT_PAGE = 12
PRINTER_SELECT_PAGE = 13
START_TEST_PAGE = 14
TESTING_PAGE = 15
MESSAGE_PAGE = 16
FINISHED_PAGE = 17

def nextPage(page):
    
    nextPages = [

    # Page 0: START_PAGE, Device and mode selection.

    { EASY_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [SETUP_TYPE_PAGE, START_PAGE],           'Next Page': MESSAGE_PAGE},
      COPY_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [COPY_SOURCE_PAGE, START_PAGE],          'Next Page': MESSAGE_PAGE},
      CUSTOM_MODE: {'Action': rct.saveNextPrevPages, 'Args': [CUSTOM_PASSWORD_PAGE, START_PAGE],      'Next Page': MESSAGE_PAGE},
      REPORT_MODE: {'Action': rct.saveNextPrevPages, 'Args': [REPORT_DEVICE_SELECT_PAGE, START_PAGE], 'Next Page': MESSAGE_PAGE},
      TEST_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [DEVICE_SELECT_PAGE, START_PAGE],        'Next Page': MESSAGE_PAGE}
    },

    # Page 1: CUSTOM_PASSWORD_PAGE, RCT Custom Mode start page.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Next Page': SETUP_TYPE_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
    },
    
    # Page 2: SETUP_TYPE_PAGE, Setup type selection.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Next Page': DEVICE_SELECT_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Next Page': DEVICE_SELECT_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
    },
    
    # Page 3: DEVICE_SELECT_PAGE, Instrument selection.
    
    { EASY_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [selectFolderOrPrinter(),
                                                              DEVICE_SELECT_PAGE],     'Next Page': MESSAGE_PAGE},
      COPY_MODE:   {'Action': None,                  'Args': [],                       'Next Page': None},
      CUSTOM_MODE: {'Action': rct.saveNextPrevPages, 'Args': [FEATURE_SELECT_PAGE,
                                                              DEVICE_SELECT_PAGE],     'Next Page': MESSAGE_PAGE},
      REPORT_MODE: {'Action': None,                  'Args': [],                       'Next Page': None},
      TEST_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [START_TEST_PAGE,
                                                              DEVICE_SELECT_PAGE],     'Next Page': MESSAGE_PAGE}
    },
    
    # Page 4: FEATURE_SELECT_PAGE, Selecting features to configure.

    { EASY_MODE:   {'Action': None,                  'Args': [], 'Next Page': None},
      COPY_MODE:   {'Action': None,                  'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': selectFolderOrPrinter, 'Args': [], 'Next Page': USE_PAGE},
      REPORT_MODE: {'Action': None,                  'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None,                  'Args': [], 'Next Page': None},
    },
    
    # Page 5: FOLDER_SELECT_PAGE, Shared Folder Selection.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Next Page': COMMIT_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Next Page': USER_NAME_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
    },
    
    # Page 6: USER_NAME_PAGE, User name and password.

    { EASY_MODE:   {'Action': None,                 'Args': [], 'Next Page': None},
      COPY_MODE:   {'Action': None,                 'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': commitOrMoreFeatures, 'Args': [], 'Next Page': USE_PAGE},
      REPORT_MODE: {'Action': None,                 'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None,                 'Args': [], 'Next Page': None},
    },
    
    # Page 7: COPY_SOURCE_PAGE, Selecting the copy source.
    
    { EASY_MODE:   {'Action': None,                  'Args': [],                 'Next Page': None},
      COPY_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [COPY_DEST_PAGE,
                                                              COPY_SOURCE_PAGE], 'Next Page': MESSAGE_PAGE},
      CUSTOM_MODE: {'Action': None,                  'Args': [],                 'Next Page': None},
      REPORT_MODE: {'Action': None,                  'Args': [],                 'Next Page': None},
      TEST_MODE:   {'Action': None,                  'Args': [],                 'Next Page': None}
    },
    
    # Page 8: COPY_DEST_PAGE, Selecting the copy destination.
    
    { EASY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Next Page': COMMIT_PAGE},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
    },
    
    # Page 9: REPORT_DEVICE_SELECT_PAGE, Selecting the report instrument.
    
    { EASY_MODE:   {'Action': None,                  'Args': [],                          'Next Page': None},
      COPY_MODE:   {'Action': None,                  'Args': [],                          'Next Page': None},
      CUSTOM_MODE: {'Action': None,                  'Args': [],                          'Next Page': None},
      REPORT_MODE: {'Action': rct.saveNextPrevPages, 'Args': [REPORT_NETWORKING_PAGE,
                                                              REPORT_DEVICE_SELECT_PAGE], 'Next Page': MESSAGE_PAGE},
      TEST_MODE:   {'Action': None,                  'Args': [],                          'Next Page': None}
    },
    
    # Page 10: REPORT_NETWORKING_PAGE, Displaying network settings.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Next Page': None},
      REPORT_MODE: {'Action': None, 'Args': [], 'Next Page': REPORT_FEATURE_PAGE},
      TEST_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
    },
    
    # Page 11: REPORT_FEATURE_PAGE, Displaying feature detail.

    { EASY_MODE:   {'Action': None,                   'Args': [], 'Next Page': None},
      COPY_MODE:   {'Action': None,                   'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': None,                   'Args': [], 'Next Page': None},
      REPORT_MODE: {'Action': finishedOrMoreToReport, 'Args': [], 'Next Page': USE_PAGE},
      TEST_MODE:   {'Action': None,                   'Args': [], 'Next Page': None},
    },
    
    # Page 12: COMMIT_PAGE, Commit page.

    { EASY_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [finishedOrTesting(), COMMIT_PAGE], 'Next Page': MESSAGE_PAGE},
      COPY_MODE:   {'Action': rct.saveNextPrevPages, 'Args': [finishedOrTesting(), COMMIT_PAGE], 'Next Page': MESSAGE_PAGE},
      CUSTOM_MODE: {'Action': rct.saveNextPrevPages, 'Args': [finishedOrTesting(), COMMIT_PAGE], 'Next Page': MESSAGE_PAGE},
      REPORT_MODE: {'Action': None,                  'Args': [],                                 'Next Page': None},
      TEST_MODE:   {'Action': None,                  'Args': [],                                 'Next Page': None},
    },
    
    # Page 13: PRINTER_SELECT_PAGE, Shared printer selection.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Next Page': COMMIT_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Next Page': USER_NAME_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Next Page': None},
    },
    
    # Page 14: START_TEST_PAGE, Start testing page.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Next Page': TESTING_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Next Page': TESTING_PAGE},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Next Page': TESTING_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Next Page': TESTING_PAGE},
    },
    
    # Page 15: TESTING_PAGE, Test progress page.

    { EASY_MODE:   {'Action': finishedOrMoreTesting, 'Args': [], 'Next Page': USE_PAGE},
      COPY_MODE:   {'Action': finishedOrMoreTesting, 'Args': [], 'Next Page': USE_PAGE},
      CUSTOM_MODE: {'Action': finishedOrMoreTesting, 'Args': [], 'Next Page': USE_PAGE},
      REPORT_MODE: {'Action': None,                  'Args': [], 'Next Page': None},
      TEST_MODE:   {'Action': finishedOrMoreTesting, 'Args': [], 'Next Page': USE_PAGE},
    },

    # Page 16: MESSAGE_PAGE, Display message.

    { EASY_MODE:   {'Action': rct.messageDone, 'Args': [SUCCESS], 'Next Page': USE_PAGE},
      COPY_MODE:   {'Action': rct.messageDone, 'Args': [SUCCESS], 'Next Page': USE_PAGE},
      CUSTOM_MODE: {'Action': rct.messageDone, 'Args': [SUCCESS], 'Next Page': USE_PAGE},
      REPORT_MODE: {'Action': rct.messageDone, 'Args': [SUCCESS], 'Next Page': USE_PAGE},
      TEST_MODE:   {'Action': rct.messageDone, 'Args': [SUCCESS], 'Next Page': USE_PAGE}
    } ]

    mode = rct.getMode()
    currentPage = rct.wizard.pageNumber(page)
    
    try:
        entry = nextPages[currentPage][mode]
        pg = entry['Next Page']
        if entry['Action']:
            result = apply(entry['Action'], entry['Args'])
            if pg == USE_PAGE:
                if DEBUG_PAGING: print 'nextPage: result pg =', result
                pg = result
        if DEBUG_PAGING: print 'Mode =', mode, 'CurrentPage =', currentPage, 'Next page =', pg
        return rct.wizard.pages[pg]
    except:
        if DEBUG_PAGING: print 'nextPage returning None, mode = %s, currentPage = %d.' % (mode, currentPage)
        return   # nextPage


def selectFolderOrPrinter():
    
    if rct.folderTypeSelected():
        return FOLDER_SELECT_PAGE
    else:
        return PRINTER_SELECT_PAGE   # selectFolderOrPrinter


def commitOrMoreFeatures():
    
    if instrumentSelector.hasMore():
        return FEATURE_SELECT_PAGE
    else:
        return COMMIT_PAGE   # commitOrMoreFeatures


def finishedOrMoreToReport():
    
    if rct.hasMode(REPORT_MODE) and instrumentSelector.hasMore():
        return REPORT_NETWORKING_PAGE
    else:
        return FINISHED_PAGE   # finishedOrMoreToReport
        

def finishedOrTesting():
    
    if rct.performTests():
        return START_TEST_PAGE
    
    return FINISHED_PAGE   # finishedOrTesting


def finishedOrMoreTesting():
    
    try:
        if devicesToTest:
            if devicesToTest.hasMore():
                return TESTING_PAGE
            else:
                return FINISHED_PAGE
    except:
        return FINISHED_PAGE   # finishedOrMoreTesting


def prevPage(page):
    
    prevPages = [

    # Page 0: START_PAGE, Device and mode selection.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },
    
    # Page 1: CUSTOM_PASSWORD_PAGE, RCT Custom Mode start page.
    
    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': START_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
      },
    
    # Page 2: SETUP_TYPE_PAGE, Setup type selection.
    
    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': START_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },
      
    # Page 3: DEVICE_SELECT_PAGE, Instrument selection.
    
    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': SETUP_TYPE_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': SETUP_TYPE_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_PAGE}
    },
      
    # Page 4: FEATURE_SELECT_PAGE, Selecting features to configure.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': DEVICE_SELECT_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },

    # Page 5: FOLDER_SELECT_PAGE, Shared Folder Selection.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': DEVICE_SELECT_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': FEATURE_SELECT_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },

    # Page 6: USER_NAME_PAGE, User name and password.

    { EASY_MODE:   {'Action': None,                  'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None,                  'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': selectFolderOrPrinter, 'Args': [], 'Prev Page': USE_PAGE},
      REPORT_MODE: {'Action': None,                  'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None,                  'Args': [], 'Prev Page': None}
    },

    # Page 7: COPY_SOURCE_PAGE, Selecting the copy source.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_PAGE},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },

    # Page 8: COPY_DEST_PAGE, Selecting the copy destination.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': COPY_SOURCE_PAGE},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },

    # Page 9: REPORT_DEVICE_SELECT_PAGE, Selecting the report instrument.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': START_PAGE},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },

    # Page 10: REPORT_NETWORKING_PAGE, Displaying network settings.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': REPORT_DEVICE_SELECT_PAGE}
    },

    # Page 11: REPORT_FEATURE_PAGE, Displaying feature detail.
    
    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': REPORT_NETWORKING_PAGE},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },

    # Page 12: COMMIT_PAGE, Commit page.

    { EASY_MODE:   {'Action': selectFolderOrPrinter, 'Args': [], 'Prev Page': USE_PAGE},
      COPY_MODE:   {'Action': None,                  'Args': [], 'Prev Page': COPY_DEST_PAGE},
      CUSTOM_MODE: {'Action': None,                  'Args': [], 'Prev Page': DEVICE_SELECT_PAGE},
      REPORT_MODE: {'Action': None,                  'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None,                  'Args': [], 'Prev Page': None}
    },

    # Page 13: PRINTER_SELECT_PAGE, Shared printer selection.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': DEVICE_SELECT_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': None},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': FEATURE_SELECT_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': None}
    },

    # Page 14: START_TEST_PAGE, Start testing page.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_TEST_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_TEST_PAGE},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': START_TEST_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': DEVICE_SELECT_PAGE}
    },

    # Page 15: TESTING_PAGE, Test progress page.

    { EASY_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_TEST_PAGE},
      COPY_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_TEST_PAGE},
      CUSTOM_MODE: {'Action': None, 'Args': [], 'Prev Page': START_TEST_PAGE},
      REPORT_MODE: {'Action': None, 'Args': [], 'Prev Page': None},
      TEST_MODE:   {'Action': None, 'Args': [], 'Prev Page': START_TEST_PAGE}
    },

    # Page 16: MESSAGE_PAGE, Display message.

    { EASY_MODE:   {'Action': rct.messageDone, 'Args': [FAILURE], 'Prev Page': USE_PAGE},
      COPY_MODE:   {'Action': rct.messageDone, 'Args': [FAILURE], 'Prev Page': USE_PAGE},
      CUSTOM_MODE: {'Action': rct.messageDone, 'Args': [FAILURE], 'Prev Page': USE_PAGE},
      REPORT_MODE: {'Action': rct.messageDone, 'Args': [FAILURE], 'Prev Page': USE_PAGE},
      TEST_MODE:   {'Action': rct.messageDone, 'Args': [FAILURE], 'Prev Page': USE_PAGE}
    },

    # Page 17: FINISHED_PAGE, Completion page.

    { EASY_MODE:   {'Action': finishedOrTesting, 'Args': [], 'Prev Page': USE_PAGE},
      COPY_MODE:   {'Action': finishedOrTesting, 'Args': [], 'Prev Page': USE_PAGE},
      CUSTOM_MODE: {'Action': finishedOrTesting, 'Args': [], 'Prev Page': USE_PAGE},
      REPORT_MODE: {'Action': None,              'Args': [], 'Prev Page': REPORT_DEVICE_SELECT_PAGE},
      TEST_MODE:   {'Action': None,              'Args': [], 'Prev Page': START_TEST_PAGE}
    } ]
    
    mode = rct.getMode()
    currentPage = rct.wizard.pageNumber(page)
    
    try:
        entry = prevPages[currentPage][mode]
        pg = entry['Prev Page']
        if entry['Action']:
            result = apply(entry['Action'], entry['Args'])
            if pg == USE_PAGE:
                if DEBUG_PAGING: print 'prevPage: result pg =', result
                pg = result
        if DEBUG_PAGING: print 'Mode =', mode, 'CurrentPage =', currentPage, 'Prev page =', pg
        return rct.wizard.pages[pg]
    except:
        if DEBUG_PAGING: print 'prevPage returning None, mode = %s, currentPage = %d.' % (mode, currentPage)
        return   # prevPage


def onPageChanging(event):
    
    """
    onPageChanging --
    
    Handle the EVT_WIZARD_PAGE_CHANGING events.
    """
    
    global instrumentSelector
    
    # Page 0: Determine mode of operation.
    
    if event.GetPage() == rct.wizard.pages[START_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        rct.setDeviceType()
        
        rct.setMode()
        
        # Require that the operator be an administrator of this computer.
        
        if rct.hasMode([EASY_MODE, COPY_MODE, CUSTOM_MODE, TEST_MODE]) and not validAdmin(userName()):
            alert(_('You must be an administrator of this computer to configure instruments. ') +
                  _('Please select \'%s\' to exit the program and then login as an administrator before trying again.') % rct.cancelButtonLabel)
            event.Veto()
            return
        
        # Require that each user of RCT view the Apple Bonjour End User License Agreement before using the program.
        
        if not rct.licenseAgreement():
            event.Veto()
            rct.stop()
            return
        
        if rct.hasMode([COPY_MODE, REPORT_MODE, TEST_MODE]):
            rct.setAllSetupTypes()
            
        # Start the Bonjour mDNSResponder.
        
        if rct.hasMode(REPORT_MODE):
            startBonjour('server')
        else:
            startBonjour('install')
        
        if not Device.devices():
            browser = Worker(browseServicesTask(rct))
            browser.start()
        
        return
    
    # Page 1 :CUSTOM MODE: When entering the RCT password.
    
    if event.GetPage() == rct.wizard.pages[CUSTOM_PASSWORD_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        if not validRCTPassword(rct.customText.GetValue()):
            alert(_('Password incorrect, please try again.'))
            rct.customText.Clear()
            event.Veto()
        
        return

    # Page 2: Select the setup type.
    
    if event.GetPage() == rct.wizard.pages[SETUP_TYPE_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        rct.setSelectedSetupType()
        
        if rct.printerTypeSelected() and not getPrinters():
            alert(_('There are no printers available on this computer. ') +
                  _('Please select \'%s\' to exit the program and then add a printer before trying again.') % rct.cancelButtonLabel)
            event.Veto()
        
        rct.SHARE_NAME = ''
        
        return
    
    # Page 3: Query selected instruments for configurable services.
    
    if event.GetPage() == rct.wizard.pages[DEVICE_SELECT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        selectedInstruments = list(Selector(rct.instrumentListCtrl))
        
        if not selectedInstruments:
            nothingToConfigure()
            event.Veto()
            return
        
        rct.SELECTED_INSTRUMENTS = copy.deepcopy(selectedInstruments)
                
        if DEBUG:
            print 'SELECTED_INSTRUMENTS =\n', pprint.pformat(rct.SELECTED_INSTRUMENTS, indent = 4)
        
        if not rct.SELECTED_INSTRUMENTS or not anyCapabilities(rct.SELECTED_INSTRUMENTS):
            nothingToConfigure()
            event.Veto()
            return
        
        if not rct.hasMode(TEST_MODE):
            instrumentSelector = Selector(rct.instrumentListCtrl)
        
        requester = Worker(requestCapabilitiesTask(rct, rct.SELECTED_INSTRUMENTS))
        requester.start()
        
        return
    
    # Page 4: CUSTOM MODE: When selecting features to configure.
    
    if event.GetPage() == rct.wizard.pages[FEATURE_SELECT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        selectedInstrument = instrumentSelector.current()
        device = getDevice(selectedInstrument)
        
        capabilities = list(Selector(rct.capabilityListCtrl))
        
        if not capabilities:
            noFeaturesToConfigure()
            event.Veto()
            return
        
        selectedCapabilities = copy.deepcopy(capabilities)
        
        if DEBUG:
            print 'Custom: selected capabilities for', selectedInstrument, '='
               
        for feature in [capability[0] for capability in selectedCapabilities]:
            device[CAPABILITIES_FIELD][feature][SELECTED_FIELD] = True
            if DEBUG:
               print feature, '='
               pprint.pprint(device[CAPABILITIES_FIELD][feature])
        
        if not anySelectedCapabilities(device):
            if DEBUG:
                print 'No capabilities selected.'
            noFeaturesToConfigure()
            event.Veto()
        
        return
    
    # Page 5: When selecting the shared folder.
    
    if event.GetPage() == rct.wizard.pages[FOLDER_SELECT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        enteredSharedResource = rct.sharedResourceEntry.GetValue()
        
        if os.path.exists(enteredSharedResource):
            rct.SHARED_FOLDER = getDirectoryName(enteredSharedResource)
            rct.sharedResourceEntry.SetValue(rct.SHARED_FOLDER)
        elif enteredSharedResource:
            if not createFolderConfirmed(enteredSharedResource):
                event.Veto()
                return
            else:
                os.makedirs(enteredSharedResource, mode = SHARED_DIRECTORY_MODE)
                rct.SHARED_FOLDER = os.path.normpath(enteredSharedResource)
                rct.sharedResourceEntry.SetValue(rct.SHARED_FOLDER)
        else:
            alert(_('You must enter or select a shared folder.'))
            event.Veto()
            return
        
        # Create user to control access to the shared folder.
        
        if rct.hasMode(EASY_MODE):
            userInfo = makeUser(rct.getUserName(), rct.getUserPassword(encrypted = False))
            if userInfo == RESTART_REQUESTED:
                event.Veto()
                rct.restart()
            if not userInfo:
                alert(_('Cannot create user \'%s\' for share.') % rct.getUserName())
                event.Veto()
                return
        
        # Create the network share if it does not already exist.
        
        shareInfo = makeShare(rct.SHARED_FOLDER, defaultShareName())
        if not shareInfo:
            alert(_('Could not share the folder \'%s\'.') % rct.SHARED_FOLDER)
            event.Veto()
            return
        
        if shareInfo == NOT_AVAILABLE:
            print 'Cannot yet create a share on \'%s\' systems.' % os.name
            rct.SHARE_NAME = NOT_AVAILABLE
        else:
            rct.SHARE_NAME = shareInfo['netname']                       # obtain the actual share name.
            if rct.hasMode(EASY_MODE):
                grantAccessToShare(rct.getUserName(), rct.SHARE_NAME)   # set the user permissions.

        if DEBUG:
            print 'SHARE_NAME =', rct.SHARE_NAME
            print 'SHARED_FOLDER =', rct.SHARED_FOLDER
            print 'User Name =', rct.getUserName()
            print 'Password =', rct.getUserPassword(encrypted = True)

        # Create any required subfolders.
        
        if not makeHFAfolders(rct.SHARED_FOLDER):
            alert(_('Could not create the standard HFA subfolders in the shared folder.'))
        
        # Record the configuration data.
        
        for selectedInstrument in rct.SELECTED_INSTRUMENTS:
            device = getDevice(selectedInstrument)
            for feature in device.capabilities():
                if rct.isSharedFolderType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
                    device[CAPABILITIES_FIELD][feature][NETWORK_PROTOCOL_FIELD] = 'CIFS'
                    device[CAPABILITIES_FIELD][feature][SHARED_RESOURCE_FIELD] = rct.SHARED_FOLDER
                    device[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD] = rct.SHARE_NAME
                    device[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] = rct.HOST_NAME
                    device[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD] = rct.HOST_IP_ADDRESS
                    if rct.hasMode(EASY_MODE):
                        device[CAPABILITIES_FIELD][feature][USER_NAME_FIELD] = rct.getUserName()
                        device[CAPABILITIES_FIELD][feature][PASSWORD_FIELD] = rct.getUserPassword(encrypted = True)
        return
    
    # Page 6: When specifying the user name and password.
    
    if event.GetPage() == rct.wizard.pages[USER_NAME_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        if not rct.validUserName() or not rct.validUserPassword():
            alert(_('You have entered an invalid user name or password.'))
            rct.showUserName()
            rct.showUserPassword()
            event.Veto()
            return
        
        userInfo = makeUser(rct.setUserName(), rct.setUserPassword())
        
        if userInfo == PASSWORD_CHANGE_REQUESTED:
            event.Veto()
            return
        
        if not userInfo:
            alert(_('Could not create the user name as entered.'))
            event.Veto()
            return
        
        if rct.printerTypeSelected():
            grantAccessToPrinterShare(rct.getUserName(), rct.SHARE_NAME, rct.SHARED_PRINTER)
        else:
            grantAccessToShare(rct.getUserName(), rct.SHARE_NAME)
        
        if DEBUG:
            print 'User Name =', rct.getUserName()
            print 'Password =', rct.getUserPassword(encrypted = True)
        
        for selectedInstrument in rct.SELECTED_INSTRUMENTS:
            device = getDevice(selectedInstrument)
            for feature in device.capabilities():
                if rct.isSelectedSetupType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
                    device[CAPABILITIES_FIELD][feature][USER_NAME_FIELD] = rct.getUserName()
                    device[CAPABILITIES_FIELD][feature][PASSWORD_FIELD] = rct.getUserPassword(encrypted = True)
        
        return
    
    # Page 7: When specifying the source instrument for a copy.
    
    if event.GetPage() == rct.wizard.pages[COPY_SOURCE_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        if rct.copySrcListCtrl.GetFirstSelected() == -1:
            alert(_('You must select an instrument to copy.'))
            event.Veto()
            return
        
        selectedInstrument = rct.copySrcListCtrl.getFirstSelection()
        
        if emptySelection(selectedInstrument):
            alert(_('You must select an instrument to copy.'))
            event.Veto()
            return
        
        rct.COPY_SRC_INSTRUMENT = copy.copy(selectedInstrument)
        
        if DEBUG:
            print 'Copy: COPY_SRC_INSTRUMENT =', pprint.pformat(rct.COPY_SRC_INSTRUMENT, indent = 4)
        
        requester = Worker(requestCapabilitiesTask(rct, [rct.COPY_SRC_INSTRUMENT]))
        requester.start()
        
        return
    
    # Page 8: When specifying the destination instruments for a copy.
    
    if event.GetPage() == rct.wizard.pages[COPY_DEST_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        if rct.copyDstListCtrl.GetFirstSelected() == -1:
            alert(_('You must select at least one destination instrument.'))
            event.Veto()
            return
        
        selectedInstruments = list(Selector(rct.copyDstListCtrl))
        
        if not selectedInstruments:
            alert(_('You must select at least one destination instrument.'))
            event.Veto()
            return
        
        rct.COPY_DEST_INSTRUMENTS = copy.deepcopy(selectedInstruments)
            
        if DEBUG:
            print 'Copy: COPY_DEST_INSTRUMENTS =\n', pprint.pformat(rct.COPY_DEST_INSTRUMENTS, indent = 4)
        
        # Copy the configuration data.
        
        copyConfig(rct.COPY_DEST_INSTRUMENTS, rct.COPY_SRC_INSTRUMENT)
        
        return
    
    # Page 9: When specifying the instruments for displaying a report.
    
    if event.GetPage() == rct.wizard.pages[REPORT_DEVICE_SELECT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        selectedInstruments = list(Selector(rct.reportInstrListCtrl))
        
        if not selectedInstruments:
            nothingToDisplay()
            event.Veto()
            return
        
        rct.SELECTED_INSTRUMENTS = copy.deepcopy(selectedInstruments)
 
        if DEBUG:
            print 'Report: SELECTED_INSTRUMENTS =\n', pprint.pformat(rct.SELECTED_INSTRUMENTS, indent = 4)
        
        instrumentSelector = Selector(rct.reportInstrListCtrl)
        
        if not anyCapabilities(rct.SELECTED_INSTRUMENTS):
            nothingToDisplay()
            event.Veto()
            return
        
        requester = Worker(requestCapabilitiesTask(rct, rct.SELECTED_INSTRUMENTS))
        requester.start()
        
        return
    
    # Page 10: When displaying the instrument network settings.
    
    if event.GetPage() == rct.wizard.pages[REPORT_FEATURE_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        pass
        return
    
    # Page 11: When displaying the instrument feature configuration details.
    
    if event.GetPage() == rct.wizard.pages[REPORT_FEATURE_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        pass
        return
    
    # Page 12: When prompting to commit the configurations.
    
    if event.GetPage() == rct.wizard.pages[COMMIT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        if rct.hasMode(COPY_MODE):
            devices = [getDevice(instrument) for instrument in rct.COPY_DEST_INSTRUMENTS]
        else:
            devices = [getDevice(instrument) for instrument in rct.SELECTED_INSTRUMENTS]
        
        if devices:
            builder = Worker(configureDevicesTask(rct, devices))
            builder.start()
        
        return
    
    # Page 13: When selecting the shared printer.
    
    if event.GetPage() == rct.wizard.pages[PRINTER_SELECT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        # Retrieve the selected printer description.
        
        printerSelection = rct.printerListCtrl.getFirstSelection()
        if printerSelection:
            rct.SHARED_PRINTER, rct.SHARE_NAME = printerSelection
        
        if not rct.SHARED_PRINTER:
            alert(_('You must select a printer.'))
            rct.printerListCtrl.clearSelection()
            event.Veto()
            return
        
        # Create user to control access to the shared printer.
        
        if rct.hasMode(EASY_MODE):
            userInfo = makeUser(rct.getUserName(), rct.getUserPassword(encrypted = False))
            if userInfo == RESTART_REQUESTED:
                event.Veto()
                rct.restart()
            if not userInfo:
                alert(_('Cannot create user \'%s\' for shared printer.') % rct.getUserName())
                event.Veto()
                return

        # Create the network share for the printer if it does not already exist.

        printerInfo = makePrinterShare(rct.SHARED_PRINTER, rct.SHARE_NAME)
        if not printerInfo:
            alert(_('Could not share the printer \'%s\'.') % rct.SHARED_PRINTER)
            event.Veto()
            return

        if printerInfo == NOT_AVAILABLE:
            print 'Cannot yet create a shared printer on \'%s\' systems.' % os.name
            rct.SHARE_NAME = NOT_AVAILABLE
        else:
            rct.SHARE_NAME = printerInfo['pShareName']   # obtain the actual share name.
            if rct.hasMode(EASY_MODE):
                grantAccessToPrinterShare(rct.getUserName(), rct.SHARE_NAME, rct.SHARED_PRINTER)   # set the user permissions.
        
        if DEBUG:
            print 'SHARE_NAME =', rct.SHARE_NAME
            print 'SHARED_PRINTER =', rct.SHARED_PRINTER
            print 'User Name =', rct.getUserName()
            print 'Password =', rct.getUserPassword(encrypted = True)

        # Record the configuration data.
        
        for selectedInstrument in rct.SELECTED_INSTRUMENTS:
            device = getDevice(selectedInstrument)
            for feature in device.capabilities():
                if rct.isSharedPrinterType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
                    device[CAPABILITIES_FIELD][feature][NETWORK_PROTOCOL_FIELD] = 'CIFS'
                    device[CAPABILITIES_FIELD][feature][SHARED_RESOURCE_FIELD] = rct.SHARED_PRINTER
                    device[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD] = rct.SHARE_NAME
                    device[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] = rct.HOST_NAME
                    device[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD] = rct.HOST_IP_ADDRESS
                    if rct.hasMode(EASY_MODE):
                        device[CAPABILITIES_FIELD][feature][USER_NAME_FIELD] = rct.getUserName()
                        device[CAPABILITIES_FIELD][feature][PASSWORD_FIELD] = rct.getUserPassword(encrypted = True)
    
    # Page 14: When beginning to perform connection tests.
    
    if event.GetPage() == rct.wizard.pages[START_TEST_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        try:
            device = devicesToTest.next()
        except StopIteration:
            return
        
        device[TEST_RESULTS_FIELD] = TESTING
        tester = Worker(testDevicesTask(rct, device))
        tester.start()
        
        return
    
    # Page 15: When continuing to perform connection tests.
    
    if event.GetPage() == rct.wizard.pages[TESTING_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        try:
            device = devicesToTest.next()
        except StopIteration:
            return
        
        device[TEST_RESULTS_FIELD] = TESTING
        tester = Worker(testDevicesTask(rct, device))
        tester.start()
        
        return
    
    return   # onPageChanging


def onPageChanged(event):
    
    """
    onPageChanged --
    
    Handle the EVT_WIZARD_PAGE_CHANGED events.
    """
    
    global devicesToTest
    
    # Page 0: Display the instrument type selection.
    
    if event.GetPage() == rct.wizard.pages[START_PAGE]:
        
        if event.GetDirection() == GOING_BACKWARD:
            for device in tuple(Device.devices()):
                device.delete()
        
        rct.enableNavigationButtons(GOING_FORWARD)
        
        return
    
    # Page 1: CUSTOM MODE: Clear the starting page text entry field.
    
    if event.GetPage() == rct.wizard.pages[CUSTOM_PASSWORD_PAGE]:
        
        rct.customText.Clear()
        rct.customText.SetFocus()
        
        return
    
    # Page 2: Display the setup type selection.
    
    if event.GetPage() == rct.wizard.pages[SETUP_TYPE_PAGE]:
        rct.setupTypePrompt.SetLabel(_('What type of %s setup would you like to perform?') % rct.getDeviceType())
        return
    
    # Page 3: Display the instrument selection list.
    
    if event.GetPage() == rct.wizard.pages[DEVICE_SELECT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        if rct.hasMode(TEST_MODE):
            rct.instrumentSelectionPrompt.SetLabel(_('Please select one or more instruments to test:'))
        else:
            rct.instrumentSelectionPrompt.SetLabel(_('Please select one or more instruments to configure:'))
        
        if rct.instrumentListCtrl:
            rct.instrumentListCtrl.displayAllInstruments()
            rct.instrumentListCtrl.selectAll()
        
        return
    
    # Page 4: CUSTOM MODE: Display the feature detail selection list.
    
    if event.GetPage() == rct.wizard.pages[FEATURE_SELECT_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
    
        if not instrumentSelector:
            nothingToConfigure()
            return
        
        rct.capabilityListCtrl.clearSelection()
        
        try:
            selectedInstrument = instrumentSelector.next()
        except StopIteration:
            rct.instrumentDescriptor.SetLabel(_('No more selected instruments'))
            rct.capabilityListCtrl.clear()
            return
        
        if not selectedInstrument:
            nothingToConfigure()
            return
            
        rct.instrumentDescriptor.SetLabel('%s %s   %s' % tuple(selectedInstrument))
        rct.capabilityListCtrl.displayCapabilities(selectedInstrument)
        rct.capabilityListCtrl.selectAll()
        
        return
    
    # Page 5: Display the shared folder prompt.
        
    if event.GetPage() == rct.wizard.pages[FOLDER_SELECT_PAGE]:
        
        if not rct.sharedResourceEntry or (DEBUG and not rct.shareNameEntry):
            return

        sharedResource = ''
        shareName = ''
        
        rct.sharePrompt.SetLabel('\n'.join([Rct.SHARE_USAGE_MSG, Rct.SHARED_RESOURCE_PROMPT % rct.nextButtonLabel]))
        
        if rct.SHARED_FOLDER:
            sharedResource = rct.SHARED_FOLDER
        else:
            sharedResource = defaultSharedFolder()

        if rct.SHARE_NAME:
            shareName = rct.SHARE_NAME
        else:
            shareName = defaultShareName()
        
        rct.sharedResourceLabel.SetLabel('%s:' % fieldToString(SHARED_FOLDER_TYPE))
        
        rct.sharedResourceEntry.SetValue(sharedResource)
        if shareName == NOT_AVAILABLE:
            rct.shareNameEntry.SetLabel(fieldToString(NOT_AVAILABLE))
        else:
            rct.shareNameEntry.SetLabel(shareName)
        
        return
    
    # Page 6: Display the user name prompt.
    
    if event.GetPage() == rct.wizard.pages[USER_NAME_PAGE]:
        
        if rct.printerTypeSelected():
            rct.userNamePrompt.SetLabel(_('Please enter the user name and password for the shared printer:'))
        elif rct.folderTypeSelected():
            rct.userNamePrompt.SetLabel(_('Please enter the user name and password for the shared folder:'))
        
        return
    
    # Page 7: Display the list of source instruments from which to copy.
    
    if event.GetPage() == rct.wizard.pages[COPY_SOURCE_PAGE]:
        
        if not rct.copySrcListCtrl:
            return
        
        if event.GetDirection() == GOING_BACKWARD:
            rct.copySrcListCtrl.selectOne(rct.COPY_SRC_INSTRUMENT, 1)
        else:
            rct.copySrcListCtrl.displayAllInstruments()
            rct.copySrcListCtrl.selectFirst()
        
        return
    
    # Page 8: Display the list of destination instruments for copy.
    
    if event.GetPage() == rct.wizard.pages[COPY_DEST_PAGE]:
        
        if event.GetDirection() != GOING_FORWARD:
            return
        
        if rct.copyDstListCtrl:
            rct.copyDstListCtrl.displayOtherInstruments([rct.COPY_SRC_INSTRUMENT])
            rct.copyDstListCtrl.selectAll()
        
        return
    
    # Page 9: Display the list of instruments for reporting.
    
    if event.GetPage() == rct.wizard.pages[REPORT_DEVICE_SELECT_PAGE]:
        
        rct.cancelButton.Enable()
        
        if rct.reportInstrListCtrl:
            rct.reportInstrListCtrl.displayAllInstruments()
            rct.reportInstrListCtrl.selectAll()
        return
    
    # Page 10: Display the instrument network settings for Report.
    
    if event.GetPage() == rct.wizard.pages[REPORT_NETWORKING_PAGE]:
        
        if not instrumentSelector:
            nothingToDisplay()
            return
        
        rct.clearNetworkSettings()
        
        if event.GetDirection() == GOING_BACKWARD:
            selectedInstrument = instrumentSelector.current()
        else:
            try:
                selectedInstrument = instrumentSelector.next()
            except StopIteration:
                rct.networkSettingsInstrument.SetLabel(_('No more selected instruments'))
                rct.clearNetworkSettings()
                return
        
        if not selectedInstrument:
            nothingToDisplay()
            return
        
        rct.networkSettingsInstrument.SetLabel(_('Network Settings for %s %s   %s') %
                                                 tuple(selectedInstrument))
        displayNetworkSettings(selectedInstrument)
        
        return
    
    # Page 11: Display the instrument capability detail for Report.
    
    if event.GetPage() == rct.wizard.pages[REPORT_FEATURE_PAGE]:
        
        if not instrumentSelector:
            nothingToDisplay()
            return

        rct.reportDetailListCtrl.clearSelection()
        
        selectedInstrument = instrumentSelector.current()
        rct.reportingInstrument.SetLabel(_('Setup for %s %s   %s') % tuple(selectedInstrument))
        rct.reportDetailListCtrl.displayCapabilities(selectedInstrument)
        
        return
    
    # Page 12: Display the commit page.
    
    if event.GetPage() == rct.wizard.pages[COMMIT_PAGE]:
        
        pass
        
        return
    
    # Page 13: Display the shared printer selection list.
    
    if event.GetPage() == rct.wizard.pages[PRINTER_SELECT_PAGE]:
        
        if not rct.printerListCtrl:
            nothingToDisplay()
            return
            
        if event.GetDirection() == GOING_BACKWARD:
            rct.printerListCtrl.selectOne(rct.SHARED_PRINTER, 0)
            return
        
        if not isWindows():
            print 'Cannot yet display the list of shared printers on \'%s\' systems.' % os.name
            rct.printerListCtrl.setColumnWidths()
        else:
            rct.printerListCtrl.displayPrinters(getPrinters())
            rct.printerListCtrl.selectFirst()
        
        return
    
    # Page 14: Display the test start page.
    
    if event.GetPage() == rct.wizard.pages[START_TEST_PAGE]:
        
        if rct.hasMode(TEST_MODE):
            rct.testStartMessage.SetLabel(' '.join([Rct.TEST_PREP_COMPLETE_MSG,
                                                    Rct.SELECT_BEGIN_TESTING_MSG % rct.nextButtonLabel]))
        else:
            rct.testStartMessage.SetLabel('\n'.join([Rct.CONFIG_COMPLETED_MSG,
                                                     Rct.SELECT_BEGIN_TESTING_MSG % rct.nextButtonLabel]))
            rct.backButton.Disable()
        
        rct.cancelButton.Enable()
        
        if rct.hasMode(COPY_MODE):
            instruments = rct.COPY_DEST_INSTRUMENTS
        else:
            instruments = rct.SELECTED_INSTRUMENTS
        
        # Initialize test results data.
        
        devicesToTest = Selector([getDevice(instrument) for instrument in instruments])
        setTestResults(instruments, READY_TO_TEST)
        
        # Display instrument list at start of test.
        
        rct.testStartListCtrl.displaySelectedInstruments(instruments)
        
        return
    
    # Page 15: Display the test continuation page.
    
    if event.GetPage() == rct.wizard.pages[TESTING_PAGE]:
        
        if rct.hasMode(COPY_MODE):
            rct.testProgressListCtrl.displaySelectedInstruments(rct.COPY_DEST_INSTRUMENTS)
        else:
            rct.testProgressListCtrl.displaySelectedInstruments(rct.SELECTED_INSTRUMENTS)
        
        rct.disableNavigationButtons()
        
        return
    
    # Page 16: Display message page.
    
    if event.GetPage() == rct.wizard.pages[MESSAGE_PAGE]:
        
        if rct.requestingFeatureDetails():
            rct.displayMessage(Rct.FEATURE_REQUEST_IN_PROGRESS_MSG, WAIT_TO_RETURN)
        elif rct.configuringDevices():
            rct.displayMessage(Rct.CONFIGURATION_PROGRESS_MSG, WAIT_TO_RETURN)
        else:
            rct.displayMessage(Rct.BROWSING_IN_PROGRESS_MSG, WAIT_TO_RETURN)
        
        rct.disableNavigationButtons()
        
        return
    
    # Page 17: Display the completion page.
    
    if event.GetPage() == rct.wizard.pages[FINISHED_PAGE]:
    
        rct.testResultsListCtrl.Hide()
        rct.testResultsBox.Hide()
        rct.viewResultsButton.Hide()
        rct.saveResultsButton.Hide()
        
        if rct.hasMode([EASY_MODE, COPY_MODE, CUSTOM_MODE]) and not rct.performTests():
            rct.backButton.Disable()
        
        rct.cancelButton.Disable()
        
        if rct.hasMode(REPORT_MODE):
            rct.completionMessage.SetLabel(_('You have finished displaying the selected instruments.'))
            return
        
        if not rct.hasMode(TEST_MODE) and not rct.performTests():
            rct.completionMessage.SetLabel(rct.configurationMessage() + '\n\n' +
                                           Rct.NO_TESTS_PERFORMED_MSG + '\n' +
                                           Rct.HOW_TO_TEST_MSG)
            return
        
        if rct.hasMode(COPY_MODE):
            selectedInstruments = rct.COPY_DEST_INSTRUMENTS
        else:
            selectedInstruments = rct.SELECTED_INSTRUMENTS
            
        rct.testResultsListCtrl.displaySelectedInstruments(selectedInstruments)
        
        testResult = getTestResults(selectedInstruments)
        
        if not testResult:
            rct.completionMessage.SetLabel(_('One or more instruments failed a connection test.') + '\n' +
                                           _('Select \'%s\' for details.') % Rct.VIEW_TEST_RESULTS_LABEL)
        else:
            if rct.hasMode(TEST_MODE):
                rct.completionMessage.SetLabel(_('Congratulations. All connection tests were successful.'))
            else:
                rct.completionMessage.SetLabel(rct.configurationMessage())
        
        rct.testResultsListCtrl.Show()
        if not testResult:
            rct.testResultsBox.Show()
            rct.viewResultsButton.Show()
            if SAVE_TEST_RESULTS:
                rct.saveResultsButton.Show()
        
        return
    
    return   # onPageChanged


def onCancel(event):
    
    """
    onCancel --
    
    Handle the EVT_WIZARD_CANCEL event.
    """
    
    pass
    
    return   # onCancel


def onFinished(event):
    
    """
    onFinished --
    
    Handle the EVT_WIZARD_FINISHED event.
    """
    
    pass
    
    return   # onFinished


def onLanguageChanged(event):
    
    rct.getLanguageSelection()
    rct.stop()


def sendConfig(device):

    """
    sendConfig --
    
    Create and execute the configuration script.
    """
    
    result = STATUS_OK
    configMessages = []
    
    # Construct the device network configuration messages. 
    
    configMessages = ['WORKGROUP_DOMAIN=%s' % getDomain(device)]
    
    # Build and run the device network configuration script.
    
    script = ['CMD_SERVICE_TIMEOUT\n',
              'TIME_OUT=5\n',
              'CMD_FIND_SVC_INSTANCES\n',
              'SERVICE_TYPE=%s\n' % serviceType(),
              'CMD_CLOSE_CONN\n',
              'CMD_OPEN_CONN\n',
              'SERVICE_NAME=%s\n' % device[RAW_DEVICE_FIELD].get('SERVICE_NAME', ''),

              'CMD_SEND_MSG\n',
              'MSG=BEGIN_DEVICE_NETWORK_CONF\n',
              'CMD_RECV_MSG\n',

              'CMD_SEND_MSG\n',
              'MSG=%s\n' % ';'.join(configMessages),
              'CMD_RECV_MSG\n',

              'CMD_SEND_MSG\n',
              'MSG=CONF_SET\n',
              'CMD_RECV_MSG\n',

              'CMD_SEND_MSG\n',
              'MSG=END_DEVICE_NETWORK_CONF\n',

              'CMD_CLOSE_CONN\n']

    if serviceUser.runScript(script) == STATUS_ERROR:
        result = STATUS_ERROR

    if DEBUG:
        if result == STATUS_ERROR:
            print 'Error configuring network settings for [%s, %s, %s].' % device.id()
        elif rct.hasMode(COPY_MODE):
            print 'Copied network settings to [%s, %s, %s].' % device.id()
        else:
            print 'Configured network settings for [%s, %s, %s].' % device.id()
    
    # Create and run the device feature configuration scripts.
    
    for feature in device[CAPABILITIES_FIELD]:
        
        # Filter for non-empty capabilities. An empty feature could exist if it were copied
        # from an empty feature, but in any case we should not allow RCT to configure an empty feature.
        
        if device.emptyCapability(feature):
            continue
        
        # Filter for the selected configuration type.
        
        if not rct.isSelectedSetupType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
            continue
        
        # Filter for selected capabilities in Custom Mode.
        
        if (rct.hasMode(CUSTOM_MODE) and
            rct.isSharedFolderType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]) and
            not device[CAPABILITIES_FIELD][feature][SELECTED_FIELD]):
            continue
        
        # Construct the list of common configuration messages.
        
        configMessages = ['NETWORK_PROTOCOL=%s' % device[CAPABILITIES_FIELD][feature].get(NETWORK_PROTOCOL_FIELD, 'CIFS'),
                          'USERNAME=%s' % device[CAPABILITIES_FIELD][feature].get(USER_NAME_FIELD, rct.getUserName()),
                          'IP_DOMAIN=%s' % device[CAPABILITIES_FIELD][feature].get(HOST_IP_ADDRESS_FIELD, rct.HOST_IP_ADDRESS),
                          'PASSWORD=%s' % device[CAPABILITIES_FIELD][feature].get(PASSWORD_FIELD, rct.getUserPassword(encrypted = True)),
                          'HOSTNAME=%s' % device[CAPABILITIES_FIELD][feature].get(HOST_NAME_FIELD, rct.HOST_NAME),
                          'SHARED_RESOURCE=%s' % getSharedResource(device, feature)]
        
        # Add any feature specific messages.
        
        if feature == HFA_DATA_EXPORT_SETUP:
            configMessages.append('PATIENT_FOLDER=ENABLED')
            
        """
        TO_DO(06/25/06):
        sendConfig: Add RCS command for selecting the shared printer as the instrument's default.
        END_TO_DO
        """
    
        if feature == HFA_REMOTE_PRINTER_SETUP and rct.defaultPrinterRequested():
            if DEBUG:
                print 'Setting the default printer is not yet implemented.'
            configMessages.append('')   # TEMPORARY, need command to configure shared printer as default.
        
        # Build the configuration script.
        
        script = ['CMD_SERVICE_TIMEOUT\n',
                  'TIME_OUT=5\n',
                  'CMD_FIND_SVC_INSTANCES\n',
                  'SERVICE_TYPE=%s\n' % serviceType(),
                  'CMD_CLOSE_CONN\n',
                  'CMD_OPEN_CONN\n',
                  'SERVICE_NAME=%s\n' % device[RAW_DEVICE_FIELD].get('SERVICE_NAME', ''),
                  
                  'CMD_SEND_MSG\n',
                  'MSG=BEGIN_%s\n' % featureString(feature),
                  'CMD_RECV_MSG\n',
                  
                  'CMD_SEND_MSG\n',
                  'MSG=%s\n' % ';'.join(configMessages),
                  'CMD_RECV_MSG\n',

                  'CMD_SEND_MSG\n',
                  'MSG=CONF_SET\n',
                  'CMD_RECV_MSG\n',

                  'CMD_SEND_MSG\n',
                  'MSG=END_%s\n' % featureString(feature),

                  'CMD_CLOSE_CONN\n']
        
        if serviceUser.runScript(script) == STATUS_ERROR:
            result = STATUS_ERROR
        
        if DEBUG:
            if result == STATUS_ERROR:
                print 'Error configuring \'%s\'' % feature, 'for [%s, %s, %s].' % device.id()
            elif rct.hasMode(COPY_MODE):
                print 'Copied \'%s\'' % feature, 'to [%s, %s, %s].' % device.id()
            else:
                print 'Configured \'%s\'' % feature, 'for [%s, %s, %s].' % device.id()
    
    return result   # sendConfig


def getDomain(device):
    
    if isWindows():
        domain = win32net.NetWkstaGetInfo(None, 100)['langroup']
    else:
        domain = ''
    
    if rct.hasMode(COPY_MODE):
        domain = device[DEVICE_NETWORK_SETTINGS_FIELD].get(DEVICE_WORKGROUP_DOMAIN, '')
    
    return domain   # getDomain


def getSharedResource(device, feature):
    
    if rct.isSharedFolderType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
        return (deviceShareName(device, feature) + subFolder(feature)).replace('\\', '/')
    
    if rct.isSharedPrinterType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
        return printerShareName(device, feature)
    
    return ''   # getSharedResource


def deviceShareName(device, feature):
    
    """
    deviceShareName --
    
    Returns the share name assigned to the specified device feature or the default share name if not found.
    """
    
    if device and feature:
        return device[CAPABILITIES_FIELD][feature].get(SHARE_NAME_FIELD, defaultShareName())
    
    return rct.SHARE_NAME   # deviceShareName


def subFolder(feature):
    
    """
    subFolder --
    
    Returns the subfolder for the named feature.
    """
    
    if feature == HFA_ARCHIVE_SETUP:
        return ARCHIVE_FOLDER
    if feature == HFA_DATA_EXPORT_SETUP:
        return DATA_EXPORT_FOLDER
    if feature == HFA_EMR_PMS_SETUP:
        return EMR_EXPORT_FOLDER
    if feature == HFA_WORK_LIST_SETUP:
        if rct.separateSubfolders():
            return WORK_LIST_FOLDER
        else:
            return EMR_EXPORT_FOLDER
    
    return ''   # subFolder


def printerShareName(device, feature):
    
    if device and feature:
        return device[CAPABILITIES_FIELD][feature].get(SHARE_NAME_FIELD, defaultPrinterShareName())
    
    return rct.SHARE_NAME


def onIdle(event):
    
    pass
    
    return   # onIdle


def openPorts(event):
    
    """
    openPorts --
    
    Open ports in Windows XP Firewall for remote file sharing.
    """
    
    if not isWindows():
        return True
    
    if wx.ID_YES == askQuestion(_('The RCT will open ports 137, 138, 139, and 445 in the Windows Firewall. This is required to export exam files to this computer. Do you want to continue?'), Rct.QUESTION):
        
        wx.SetCursor(wx.StockCursor(wx.CURSOR_WAIT))
        
        try:
            result = os.system('openPorts.bat')
            if result != 0:
                if DEBUG:
                    print 'os.system(\'openPorts.bat\') returned %s.' % result
                wx.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))
                return False
        except:
            if DEBUG:
                print 'Exception calling os.system(\'openPorts.bat\')'
                wx.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))
                return False
    
        wx.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))
        
    return True   # openPorts


def viewDoc(event):
    
    """
    viewDoc --
    
    View the EasyConnect RCT 1.0 User Guide.
    """
    
    try:
        language = rct_properties['RCT_LANG']
        if not language:
            language = 'en'
        viewDocCmd = 'viewDoc_%s.bat' % language
        if not os.path.exists(viewDocCmd):
            if isWindows():
                viewDocCmd = 'viewDoc_en.bat'
            else:
                print 'View RCT User Guide PDF file in Preview or Acrobat Reader.'
                return
    except:
        viewDocCmd = 'viewDoc_en.bat'
    
    if isWindows():
        try:
            os.spawnv(os.P_NOWAIT, viewDocCmd, [''])
        except:
            if DEBUG:
                print 'Exception calling os.spawnv with: os.P_NOWAIT, \'%s\' % viewDocCmd, [\'\']'
    else:
        print 'View RCT User Guide PDF file in Preview or Acrobat Reader.'
    
    return   # viewDoc


def viewTestResults(event):
    
    testSet = {}
    
    testSet[SHARED_FOLDER_TYPE] = [CONNECTION_TEST_STATUS, ESTABLISH_CONNECTION,
                                   READ_TEST, CREATE_FILE, DELETE_FILE,
                                   COMMUNICATION_TEST_WITH_EMR_APPLICATION,
                                   COMMUNICATION_TEST_WITH_WORK_LIST_APPLICATION]
    testSet[SHARED_PRINTER_TYPE] = [CONNECTION_TEST_STATUS, ESTABLISH_CONNECTION]
    
    if rct.hasMode(COPY_MODE):
        instruments = rct.COPY_DEST_INSTRUMENTS
    else:
        instruments = rct.SELECTED_INSTRUMENTS
    
    for instrument in instruments:
        
        # Retrieve the instrument data and select instruments that have failed a connection test.
        
        device = getDevice(instrument)
        if DEBUG:
            print 'Device test results =', device[TEST_RESULTS_FIELD]
        if device[TEST_RESULTS_FIELD] == PASSED:
            continue
        
        # Construct the list of test results for each enabled feature.
        
        msgs = ['[ %s %s   %s ]\n' % tuple(instrument)]
        for feature in device[CAPABILITIES_FIELD]:
            msgs.append(fieldToString(feature))
            details = device[CAPABILITIES_FIELD][feature].get('TEST_DATA', {CONNECTION_TEST_STATUS: NO_RESULTS})
            setupType = device[CAPABILITIES_FIELD][feature].get(SETUP_TYPE_FIELD, SHARED_FOLDER_TYPE)
            for item in testSet[setupType]:
                if item in details:
                    msgs.append('   %s: %s' % formatTestDetail(item, details[item]))
        
        # Display the test results for each failed device.
        
        logText = '\n'.join(msgs)
        if DEBUG:
            print '----------------------------'
            print '         Test Log           '
            print '----------------------------\n'
            oldStream = sys.stdout
            try:
                sys.stdout = codecs.lookup('latin-1')[-1](sys.stdout)
                print(logText)
            except UnicodeEncodeError:
                print('Could not print log trace, text contains non-ascii characters.')
            sys.stdout = oldStream
            print '\n----------------------------'
            print '       End Test Log         '
            print '----------------------------'
        testReport(logText)
    
    return   # viewTestResults


def saveResults(event):
    
    global TEST_RESULTS_FILE
    global RCT_WORK
    
    # Select or create the default working directory.
    
    if not os.path.exists(RCT_WORK):
        if os.environ.get('USERPROFILE'):
            RCT_WORK = os.path.join(os.environ['USERPROFILE'], 'My Documents', PRODUCT_SHORT_NAME)
        elif os.environ.get('HOME'):
            RCT_WORK = os.path.join(os.environ['HOME'], 'Documents', PRODUCT_SHORT_NAME)
        else:
            if DEBUG:
                print 'Could not find environment variable for RCT_WORK, using current working directory.'
            RCT_WORK = os.getcwd()
    
    if not os.path.exists(RCT_WORK):
        os.makedirs(RCT_WORK)
    
    # Save Report Mode data, not implemented in RCT 1.0.
    
    if rct.hasMode(REPORT_MODE):
        alert('Save Results is not implemented for Report Mode in RCT 1.0.')
        return
    
    # Save test results.
    
    if rct.hasMode([EASY_MODE, COPY_MODE, CUSTOM_MODE, TEST_MODE]):
        dialog = wx.FileDialog(None, Rct.SAVE_TEST_RESULTS_LABEL, RCT_WORK, TEST_RESULTS_FILE, '*.txt', wx.SAVE)
        if dialog.ShowModal() == wx.ID_OK:
            TEST_RESULTS_FILE = dialog.GetPath()
            try:
                testResultsFile = codecs.open(TEST_RESULTS_FILE, 'w', 'latin-1', 'replace')
            except:
                if VERBOSE:
                    alert('Cannot create the test results file, \'%s\'.' % TEST_RESULTS_FILE)
                return
            saveTestResults(testResultsFile)
            testResultsFile.close()
    
    return   # saveResults


def saveTestResults(testResultsFile):
    
    """
    saveTestResults --
    
    Save all test results to a text file.
    """
    
    testSet = {}
    
    testSet[SHARED_FOLDER_TYPE] = [CONNECTION_TEST_STATUS, ESTABLISH_CONNECTION,
                                   CREATE_FILE, READ_TEST, DELETE_FILE,
                                   COMMUNICATION_TEST_WITH_EMR_APPLICATION,
                                   COMMUNICATION_TEST_WITH_WORK_LIST_APPLICATION]
    testSet[SHARED_PRINTER_TYPE] = [CONNECTION_TEST_STATUS, ESTABLISH_CONNECTION]
    
    # Collect the test results for each selected instrument.
    
    msgs = [time.asctime()]
    
    if rct.hasMode(COPY_MODE):
        instruments = rct.COPY_DEST_INSTRUMENTS
    else:
        instruments = rct.SELECTED_INSTRUMENTS
    
    for instrument in instruments:

        # Retrieve the instrument's device data.

        device = getDevice(instrument)

        # Append the list of test results for each enabled feature.

        msgs.append('\n\n[ %s %s   %s ]\n\n' % tuple(instrument))
        for feature in device[CAPABILITIES_FIELD]:
            msgs.append('%s\n' % fieldToString(feature))
            details = device[CAPABILITIES_FIELD][feature].get('TEST_DATA', {CONNECTION_TEST_STATUS: NO_RESULTS})
            setupType = device[CAPABILITIES_FIELD][feature].get(SETUP_TYPE_FIELD, SHARED_FOLDER_TYPE)
            for item in testSet[setupType]:
                if item in details:
                    msgs.append('   %s: %s\n' % formatTestDetail(item, details[item]))
    
    # Save the test results for each device.
    
    try:
        testResultsFile.writelines(msgs)
    except:
        if VERBOSE:
            alert('Could not write to the test results file.')
    
    return   # saveTestResults


def formatTestDetail(key, value):
    
    return fieldToString(key), value == 'OK' and fieldToString(PASSED) or fieldToString(value)


def displayNetworkSettings(selectedInstrument):
    
    device = getDevice(selectedInstrument)
    
    rct.showNetworkSettings(fieldToString(device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_IP_CONFIG_PREFERENCE]),
                            fieldToString(device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_IP_CONFIG]),
                            device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_IP_ADDRESS],
                            device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_SUBNET_MASK], 
                            device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_DEFAULT_GATEWAY],
                            device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_WORKGROUP_DOMAIN])
    
    return   # displayNetworkSettings


def emptySelection(selection):
    
    if not selection:
        return True
    
    for item in selection:
        if item and len(item) > 0:
            return False
    
    return True   # emptySelection


def anySelectedCapabilities(device):
    
    for feature in device[CAPABILITIES_FIELD]:
        if device[CAPABILITIES_FIELD][feature][SELECTED_FIELD]:
            return True
    
    return False   # anySelectedCapabilities


def nothingToConfigure():

    alert(_('There is nothing to configure. Please select other instruments and try again.'))
    
    return   # nothingToConfigure


def noFeaturesToConfigure():

    alert(_('There are no features to configure. Please select one or more and try again.'))
    
    return   # noFeaturesToConfigure

def nothingToDisplay():

    alert(_('There is nothing to display. Please select other instruments and try again.'))
    
    return   # nothingToDisplay


def createFolderConfirmed(folder):

    return (wx.ID_YES == askQuestion((_('The folder, %s, does not exist. Do you want to create it?')) % folder, Rct.WARNING))
    

def getInstrumentTypes():
    
    """
    getInstrumentTypes --

    Returns a list of selected instrument types. This version simply returns a list of the one
    type selected via the device type radio box.
    """
    
    return [rct.getDeviceType()]   # getInstrumentTypes


def makeInstruments(types):
    
    """
    makeInstruments --
    
    Creates a list of Device objects from the responding instruments. Those instruments will be filtered
    for the specified types.
    """
    
    # Query for available services.
    
    rct.AVAILABLE_SERVICES = serviceUser.getServices()
    
    if not rct.AVAILABLE_SERVICES:
        if DEBUG:
            print 'No \'%s\' services discovered.' % serviceType()
        return NO_AVAILABLE_SERVICES
    elif DEBUG:
        print 'AVAILABLE_SERVICES =\n', pprint.pformat(rct.AVAILABLE_SERVICES)
    
    # Query for device information from each responding service.
    
    rct.RESPONDING_DEVICES = []
    
    serviceNames = [service['SERVICE_NAME'] for service in rct.AVAILABLE_SERVICES
                                                if not FRIENDS or service['SERVICE_NAME'] in FRIENDS]
    if DEBUG:
        print 'Service names =', serviceNames
    
    deviceInfo = serviceUser.getDeviceInfo(serviceNames)
    
    if not deviceInfo:
        if DEBUG:
            print 'No device information available for \'%s\'.' % serviceNames
        return NO_DEVICE_INFORMATION

    if DEBUG:
        print 'deviceInfo =\n', pprint.pformat(deviceInfo)
    
    # Collect the device information for each responding service.
    
    for serviceName in serviceNames:
        serviceInfo = {}
        for info in deviceInfo:
            if defaultParam(info, 'SERVICE_NAME', '') == serviceName:
                serviceInfo.update(info)
        if serviceInfo:
            rct.RESPONDING_DEVICES.extend([serviceInfo])
        else:
            if DEBUG:
                print 'No SERVICE_NAME in device information for \'%s\'.' % serviceName
    if DEBUG:
        print 'RESPONDING_DEVICES =\n', pprint.pformat(rct.RESPONDING_DEVICES)
    
    # Create the device objects.
    
    if rct.RESPONDING_DEVICES:
        devices = createDevices(rct.RESPONDING_DEVICES)
    
    if not rct.RESPONDING_DEVICES or not devices:
        if DEBUG:
            print 'Services were discovered but no device information was received.'
        return NO_DEVICE_INFORMATION
    
    return devices   # makeInstruments


def featureString(feature):
    
    msgs = {HFA_ARCHIVE_SETUP: 'ARCHIVE_RETRIEVE_CONF',
            HFA_DATA_EXPORT_SETUP: 'DATA_EXPORT_CONF',
            HFA_EMR_PMS_SETUP: 'EMR_EXPORT_CONF',
            HFA_WORK_LIST_SETUP: 'WORK_LIST_CONF',
            HFA_REMOTE_PRINTER_SETUP: 'SHARED_PRINTER_CONF'}
    
    return msgs[feature]    # featureString


def createDevices(deviceInfoList):
    
    """
    createDevices --
    
    Create Device objects based on the device information list obtained from the responding instruments.
    """
    
    if not deviceInfoList:
        return []
    
    for deviceInfo in deviceInfoList:
        
        device = Device()
        if not device:
            if DEBUG:
                print 'Cannot create Device object.'
            break
        
        try:
            device[DEVICE_TYPE_FIELD] = deviceInfo['TYPE']
            device[SERIAL_NUMBER_FIELD] = deviceInfo['SERIAL_NUMBER']
            device[DEVICE_NAME_FIELD] = deviceInfo['USER_DEFINED_NAME']
            device[TEST_RESULTS_FIELD] = READY_TO_TEST
        except KeyError:
            device.delete()
            continue
        
        device[DEVICE_NETWORK_SETTINGS_FIELD] = {}
        device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_IP_CONFIG_PREFERENCE] = defaultParam(deviceInfo, 'IP_CONFIGURATION_PREFERENCE', UNKNOWN_MODE)
        device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_IP_CONFIG] = defaultParam(deviceInfo, 'CURRENT_IP_CONFIGURATION', UNKNOWN_MODE)
        device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_REMOTE_CONFIGURATION] = defaultParam(deviceInfo, 'REMOTE_CONFIGURATION', 'ENABLED') == 'ENABLED'
        device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_IP_ADDRESS] = defaultParam(deviceInfo, 'IP_ADDRESS', NOT_AVAILABLE)
        device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_SUBNET_MASK] = defaultParam(deviceInfo, 'SUBNET_MASK', NOT_AVAILABLE)
        device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_DEFAULT_GATEWAY] = defaultParam(deviceInfo, 'GATEWAY_IP_ADDRESS', NOT_AVAILABLE)
        device[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_WORKGROUP_DOMAIN] = defaultParam(deviceInfo, 'WORKGROUP_DOMAIN', '')
        
        device[RAW_DEVICE_FIELD] = copy.deepcopy(deviceInfo)
        
        device[CAPABILITIES_FIELD] = {}
        
        if device[RAW_DEVICE_FIELD].get('ARCHIVE_RETRIEVE_RCS') == 'ENABLED':
            device[CAPABILITIES_FIELD][HFA_ARCHIVE_SETUP] = copy.copy(NULL_HFA_SETUP)
        
        if device[RAW_DEVICE_FIELD].get('DATA_EXPORT_RCS') == 'ENABLED':
            device[CAPABILITIES_FIELD][HFA_DATA_EXPORT_SETUP] = copy.copy(NULL_HFA_SETUP)
        
        if device[RAW_DEVICE_FIELD].get('EMR_EXPORT_RCS') == 'ENABLED':
            device[CAPABILITIES_FIELD][HFA_EMR_PMS_SETUP] = copy.copy(NULL_HFA_SETUP)
        
        if device[RAW_DEVICE_FIELD].get('WORK_LIST_RCS') == 'ENABLED':
            device[CAPABILITIES_FIELD][HFA_WORK_LIST_SETUP] = copy.copy(NULL_HFA_SETUP)
        
        if device[RAW_DEVICE_FIELD].get('SHARED_PRINTER_RCS') == 'ENABLED':
            device[CAPABILITIES_FIELD][HFA_REMOTE_PRINTER_SETUP] = copy.copy(NULL_SHARED_PRINTER_SETUP)
    
    return Device.devices()   # createDevices


def getInstrumentList(types):
    
    """
    getInstrumentList --
    
    Retrieves the list of instruments of the specified types.
    """

    instruments = [device for device in Device.devices() if device[DEVICE_TYPE_FIELD] in types]
    
    if not instruments:
        instruments = [copy.copy(NULL_INSTRUMENT)]
    
    return instruments   # getInstrumentList


def getIPAddress(hostname):
    
    """
    getIPAddress --
    
    Obtains the host computer's IP address.
    """
    
    try:
        addressInfo = socket.getaddrinfo(hostname, None)
        if DEBUG:
            print 'Host address info:\n', pprint.pformat(addressInfo)
        return addressInfo[-1][-1][0]
    except:
        return ''   # getIPAddress


def performConnectionTests(device):

    """
    performConnectionTests --
    
    Ask the specified device to conduct a connection test for each of its enabled features.
    """
    
    results = {}
    status = True
    
    for feature in device[CAPABILITIES_FIELD]:
    
        if feature in FEATURES_NOT_TO_TEST:
           continue
        
        # Filter for the selected configuration type.
        
        if not rct.isSelectedSetupType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
            continue
        
        # Filter for selected capabilities in Custom Mode.
        
        if (rct.hasMode(CUSTOM_MODE) and
            rct.isSharedFolderType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]) and
            not device[CAPABILITIES_FIELD][feature][SELECTED_FIELD]):
            continue
        
        # Filter for non-empty source capabilities. An empty feature could have resulted from an erroneous
        # copy, but in any case we would not want to test it.
        
        if not rct.hasMode(TEST_MODE) and device.emptyCapability(feature):
            if DEBUG:
                print 'Attempting to test an empty \'%s\'' % feature, 'on [%s, %s, %s].' % device.id()
            continue
        
        serviceName = device[RAW_DEVICE_FIELD]['SERVICE_NAME']
        
        if DEBUG:
            print 'Testing', feature, 'for [%s, %s, %s].' % device.id()
        
        # Test the feature
        
        results = serviceUser.testDeviceFeature(serviceName, feature)
        results = results and results[0] or None
        
        # Record the results of testing the feature.
        
        if results and results.get(CONNECTION_TEST_STATUS, 'FAILED') == 'OK':
            device[CAPABILITIES_FIELD][feature][TEST_RESULTS_FIELD] = PASSED
        else:
            device[CAPABILITIES_FIELD][feature][TEST_RESULTS_FIELD] = FAILED
            status = False
        
        if results:
            device[CAPABILITIES_FIELD][feature]['TEST_DATA'] = results
        else:
            device[CAPABILITIES_FIELD][feature]['TEST_DATA'] = {CONNECTION_TEST_STATUS: NO_RESULTS}
    
        if DEBUG:
            print 'Test results for \'%s\':\n' % feature, pprint.pformat(results)
    
    return status   # performConnectionTests


def getTestResults(instruments):

    """
    getTestResults --

    Returns True if none of the instruments has failed a test and returns False otherwise.
    """
    
    for device in (getDevice(instrument) for instrument in instruments):
        if device[TEST_RESULTS_FIELD] == FAILED or device[TEST_RESULTS_FIELD] == NO_RESULTS:
            return False
    
    return True   # getTestResults


def setTestResults(expr, testResult):

    """
    setTestResults --

    Sets the test results for each of a list of devices or instruments to the specified value.
    An instrument is an element of the form [type, serial number, name]. Also works if passed
    a single device instead of a list of devices.
    """
    
    try:
        expr[:]   # force a TypeError exception when expr is a single device dictionary.
    except TypeError:
        expr = [expr]
    
    try:
        devices = [getDevice(instrument) for instrument in expr]
    except:
        devices = expr
    
    for device in devices:
        device[TEST_RESULTS_FIELD] = testResult
    
    return   # setTestResults


def getDevice(instrument):
    
    """
    getDevice --
    
    Returns the device corresponding to an instrument description of the form [type, serial number, name].
    """
    
    for device in Device.devices():
        if device[SERIAL_NUMBER_FIELD].lower() == instrument[1].lower():
            return device
    
    return {}   # getDevice


def anyCapabilities(instruments):
    
    """
    anyCapabilities --
    
    Returns True if any of the listed instruments report a configurable capability of the
    currently selected type and returns False otherwise. Instruments are represented by
    three element lists of the form [type, serial number, name] and are compared on serial
    number.
    """
    
    for instrument in instruments:
        device = getDevice(instrument)
        if device:
            for feature in device[CAPABILITIES_FIELD]:
                if rct.isSelectedSetupType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
                    return True
    
    return False   # anyCapabilities


def requestCapabilities(selectedInstruments):
    
    """
    requestCapabilities --
    
    Obtain the configurable settings from the specified instruments, where each instrument is a list
    of the form [type, serial number, name].
    """
    
    capabilities = []
    serviceFeatures = {}
    
    # Build a dictionary of configurable instrument features, keyed by service name.
    
    for device in [getDevice(instrument) for instrument in selectedInstruments]:
        selectedFeatures = [feature for feature in list(device[CAPABILITIES_FIELD])
                                if rct.isSelectedSetupType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD])]
        if not selectedFeatures:
            if DEBUG:
                print 'No features obtained for [%s, %s, %s].' % device.id()
            continue
        serviceName = device[RAW_DEVICE_FIELD]['SERVICE_NAME']
        serviceFeatures[serviceName] = copy.deepcopy(selectedFeatures)
    
    if DEBUG:
        print 'serviceFeatures =\n', pprint.pformat(serviceFeatures)
    
    # Request the instrument features.
     
    deviceFeatures = serviceUser.getDeviceFeatureInfo(serviceFeatures)
            
    if not deviceFeatures:
        if DEBUG:
            print 'No features obtained for \'%s\'.' % selectedInstruments
        return capabilities
    
    # Record the capabilities for each device.
    
    for device in [getDevice(instrument) for instrument in selectedInstruments]:
        featureInfo = {}
        serviceName = device[RAW_DEVICE_FIELD]['SERVICE_NAME']
        for feature in device[CAPABILITIES_FIELD]:
            for featureInfo in deviceFeatures:
                if serviceName == featureInfo['SERVICE_NAME'] and feature == featureInfo['FEATURE']:
                    break
            else:
                continue
            if featureInfo:
                device[CAPABILITIES_FIELD][feature][NETWORK_PROTOCOL_FIELD] = defaultParam(featureInfo, 'NETWORK_PROTOCOL', '')
                device[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD] = extractPathPrefix(defaultParam(featureInfo, 'SHARED_RESOURCE', ''))

                device[CAPABILITIES_FIELD][feature][USER_NAME_FIELD] = defaultParam(featureInfo, 'USERNAME', '')
                device[CAPABILITIES_FIELD][feature][PASSWORD_FIELD] = defaultParam(featureInfo, 'PASSWORD', '')

                device[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD] = defaultParam(featureInfo, 'IP_DOMAIN', '')
                device[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] = defaultParam(featureInfo, 'HOSTNAME', '')

                if (not device[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] and
                    rct.hasMode([CUSTOM_MODE, REPORT_MODE]) and
                    device[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD]):
                    device[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] = getHostName(device[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD])

                if rct.isSharedFolderType(device[CAPABILITIES_FIELD][feature][SETUP_TYPE_FIELD]):
                    sharedResource = getSharedFolder(device[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD],
                                                     device[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD])
                else:
                    sharedResource = getSharedPrinter(device[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD],
                                                      device[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD])
                device[CAPABILITIES_FIELD][feature][SHARED_RESOURCE_FIELD] = sharedResource

                device[RAW_DEVICE_FIELD][feature] = copy.copy(featureInfo)
            
        capabilities.extend([device[CAPABILITIES_FIELD]])
    
    return capabilities   # requestCapabilities


def getHostName(ipAddress):
    
    try:
        hostname = socket.gethostbyaddr(ipAddress)[0]
    except:
        hostname = ''
    
    return hostname   # getHostName


def defaultParam(data, key, defaultValue):
    
    value = data.get(key, defaultValue)
    if value == '' and defaultValue:
        value = defaultValue
    
    return value   # defaultParam


def extractPathPrefix(srcPath):
    
    """
    extractPathPrefix --
    
    Returns the first folder in a relative path or the containing folder in an absolute path.
    
    Examples:
    
    Relative path: CZM-HFA2i/Data_export --> CZM-HFA2i
    
    Absolute path: c:/CZM/HFA2i/Data_export --> c:/CZM/HFA2i
    """
    
    if not srcPath:
        return srcPath
    
    path = os.path.split(srcPath)
    while path[0] and path[1]:
        path = os.path.split(path[0])
    
    if path[0]:
        return os.path.split(srcPath)[0]
    
    return path[1]   # extractPathPrefix


def getSharedFolder(shareName, server):
    
    """
    getSharedFolder --
    
    Return the path to the named share on this computer.
    """
    
    if not CONFIG_ENABLED:
        server = rct.HOST_NAME
    
    if os.path.normcase(rct.HOST_NAME.split('.')[0]) != os.path.normcase(server.split('.')[0]):
        if DEBUG:
            print 'Shared folder server \'%s\' is not the current host \'%s\'.' % (server, rct.HOST_NAME)
        return ''
    
    try:
        shareInfo = win32net.NetShareGetInfo(rct.HOST_NAME, shareName, 2)
        return os.path.normcase(shareInfo['path'])
    except:
        return os.path.normcase(shareName)   # assume shareName is a path prefix


def getSharedPrinter(shareName, server):
    
    """
    getSharedPrinter --
    
    Return the name of a shared printer on this computer.
    """
    
    if not isWindows():
        return ''
    
    if os.path.normcase(rct.HOST_NAME.split('.')[0]) != os.path.normcase(server.split('.')[0]):
        if DEBUG:
            print 'Shared printer server \'%s\' is not the current host \'%s\'.' % (server, rct.HOST_NAME)
        return ''
    
    try:
        shareInfo = win32net.NetShareGetInfo(rct.HOST_NAME, shareName, 2)
        printerName = shareInfo['path']
    except:
        printerName = shareName
    
    try:
        return printerName[0: printerName.index(',')]
    except:
        return printerName


def copyConfig(destInstruments, srcInstrument):
    
    """
    copyConfig --
    
    Copies the configuration from a source instrument to a list of destination instruments.
    """
    
    src = getDevice(srcInstrument)
    for instrument in destInstruments:
        dest = getDevice(instrument)
        dest[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_REMOTE_CONFIGURATION] = src[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_REMOTE_CONFIGURATION]
        if src.validDomain():
            dest[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_WORKGROUP_DOMAIN] = src[DEVICE_NETWORK_SETTINGS_FIELD][DEVICE_WORKGROUP_DOMAIN]
        for feature in dest.capabilities():
            if src.emptyCapability(feature):
                if DEBUG:
                    print 'Attempting to copy an empty \'%s\'' % feature, 'from [%s, %s, %s]' % src.id(), 'to [%s, %s, %s].' % dest.id()
            else:
                if DEBUG:
                    print 'Copying \'%s\'' % feature, 'from [%s, %s, %s]' % src.id(), 'to [%s, %s, %s].' % dest.id()
                dest[CAPABILITIES_FIELD][feature][NETWORK_PROTOCOL_FIELD] = src[CAPABILITIES_FIELD][feature][NETWORK_PROTOCOL_FIELD]
                dest[CAPABILITIES_FIELD][feature][SHARED_RESOURCE_FIELD] = src[CAPABILITIES_FIELD][feature][SHARED_RESOURCE_FIELD]
                dest[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD] = src[CAPABILITIES_FIELD][feature][SHARE_NAME_FIELD]
                dest[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD] = src[CAPABILITIES_FIELD][feature][HOST_NAME_FIELD]
                dest[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD] = src[CAPABILITIES_FIELD][feature][HOST_IP_ADDRESS_FIELD]
                dest[CAPABILITIES_FIELD][feature][USER_NAME_FIELD] = src[CAPABILITIES_FIELD][feature][USER_NAME_FIELD]
                dest[CAPABILITIES_FIELD][feature][PASSWORD_FIELD] = src[CAPABILITIES_FIELD][feature][PASSWORD_FIELD]
                dest[CAPABILITIES_FIELD][feature][TEST_RESULTS_FIELD] = src[CAPABILITIES_FIELD][feature][TEST_RESULTS_FIELD]
                dest[TEST_RESULTS_FIELD] = src[TEST_RESULTS_FIELD]
    
    return   # copyConfig


def selectFolder(event):
    
    """
    selectFolder --
    
    Present a directory dialog to browse for a shared FEP message folder.
    
    TO_DO(10/08/06):
    selectFolder: Restrict directory browsing to the local computer.
    END_TO_DO
    """
    
    result = None
    dlg = wx.DirDialog(None, _('Choose or create a folder on this computer:'),
                             style = wx.DD_DEFAULT_STYLE | wx.DD_NEW_DIR_BUTTON)
    if dlg.ShowModal() == wx.ID_OK:
        result = dlg.GetPath()
    dlg.Destroy()
    if result:
        rct.sharedResourceEntry.SetValue(result)
    
    return   # selectFolder


def getRoot():

    """
    getRoot --

    Obtain the path prefix for a shared FEP message folder.
    """
    
    HOME = os.environ.get('HOMEDRIVE')
    if HOME:
        HOME += os.sep
    else:
        HOME = os.environ.get('HOME')
        if not HOME:
            if DEBUG:
                print 'Could not find environment variable for HOME, using current working directory.'
            HOME = os.getcwd()
    
    return os.path.normcase(os.path.normpath(HOME))   # getRoot


def getRctHome():
    
    """
    getRctHome --
    
    Obtain the path to the RCT working directory.
    """
    
    RCT_HOME = os.environ.get('RCT_HOME')
    if not RCT_HOME:
        if DEBUG:
            print 'Could not find environment variable for RCT_HOME, using current working directory.'
        RCT_HOME = os.getcwd()
    
    return os.path.normcase(os.path.normpath(RCT_HOME))   # getRctHome


def defaultPrinterShareName():
    
    return '-'.join([COMPANY_SHORT_NAME, getInstrumentTypes()[0], 'Printer'])


def defaultSharedFolder():
    
    return os.path.join(getRoot(), COMPANY_SHORT_NAME, getInstrumentTypes()[0])


def defaultShareName():
    
    return '-'.join([COMPANY_SHORT_NAME, getInstrumentTypes()[0]])   # defaultShareName


def defaultUserName():
    
    return 'Zeiss'


def defaultUserPassword():
    
    return 'November171846'


def getDirectoryName(path):
    
    if not os.path.exists(path):
        return ''
    
    if not os.path.isdir(path):
        path = os.path.dirname(path)
    
    return os.path.normpath(path)   # getDirectoryName


def makeShare(path = '',
              shareName = '',
              server = '',
              remark = 'HFA-NET Shared Folder'):
    
    """
    makeShare --
    
    Share a folder. The folder must exist and not already be shared.
    """
    
    if not isWindows():
        return NOT_AVAILABLE
    
    if not server:
        server = rct.HOST_NAME
    
    level = 502
    
    shareInfo = {}
    
    # Obtain the path name.
    
    if not path:
        path = defaultSharedFolder()
    
    if os.path.exists(os.path.normpath(path)):
        shareInfo['path'] = os.path.normpath(path)
    else:
        alert(_('Cannot create the share because the folder does not exist.'))
        return
    
    # Select a name for the share starting with the specified root name.
    
    if shareName:
        shareInfo['netname'] = makeShareName(shareName, shareInfo['path'], server)
    else:
        shareInfo['netname'] = makeShareName(defaultShareName(), shareInfo['path'], server)
    
    # If the share already names the specified path then return its data.
    
    try:
        getInfo = win32net.NetShareGetInfo(server, shareInfo['netname'], level)
        if os.path.normcase(getInfo['path']) == os.path.normcase(shareInfo['path']):
            return getInfo
    except:
        pass   # share name is not in use, continue creating it.
    
    shareInfo['type'] = win32netcon.STYPE_DISKTREE
        
    shareInfo['remark'] = remark
    
    shareInfo['permissions'] = 0    # permissions for systems with share-level security.
                                    # Windows NT, Windows 2000, Windows Server 2003,
                                    # and Windows XP do not support share-level security.
    
    shareInfo['max_uses'] = -1      # unlimited concurrent connections.
    
    shareInfo['passwd'] = ''        # password for share-level security. Ignored when
                                    # the server is running with user-level security.
    
    shareInfo['reserved'] = 0
    shareInfo['security_descriptor'] = None
    
    try:
        win32net.NetShareAdd(server, level, shareInfo)
        return shareInfo
    except:
        return
    
    return   # makeShare


def makeShareName(shareNameRoot, path, server = None):
    
    """
    makeShareName --
    
    Obtain a unique name for the specified shared folder. If the folder is already named
    then return that name otherwise generate a unique name by extending the shareNameRoot.
    """
    
    for n in xrange(sys.maxint):
        if n > 0:
            ext = '_' + str(n)
        else:
            ext = ''
        try:
            getInfo = win32net.NetShareGetInfo(server, shareNameRoot + ext, 2)
            if os.path.normcase(getInfo['path']) == os.path.normcase(path):
                return getInfo['netname']
        except:
            return shareNameRoot + ext
    
    return   # makeShareName



def makePrinterShare(printerName, shareName, server = None):
    
    """
    makePrinterShare --
    
    Share a printer and return its settings. The printer must exist.
    If it is already shared then its settings will be returned.
    """
    
    if not isWindows():
        return NOT_AVAILABLE

    PRINTER_DEFAULTS = {'pDataType': None,
                        'pDevMode': None,
                        'DesiredAccess': win32print.PRINTER_ALL_ACCESS}
    
    try:
        printer = win32print.OpenPrinter(printerName, PRINTER_DEFAULTS)
    except:
        return
    
    level = 2
    
    try:
        printerSettings = win32print.GetPrinter(printer, level)
    except:
        win32print.ClosePrinter(printer)
        return
    
    if (printerSettings['Attributes'] & win32print.PRINTER_ATTRIBUTE_SHARED) and printerSettings['pShareName']:
        win32print.ClosePrinter(printer)
        return printerSettings
    
    if shareName:
        printerSettings['pShareName'] = makePrinterShareName(shareName, printerName)
    else:
        printerSettings['pShareName'] = makePrinterShareName(defaultPrinterShareName(), printerName)
    
    printerSettings['Attributes'] |= win32print.PRINTER_ATTRIBUTE_SHARED
    
    try:
        win32print.SetPrinter(printer, level, printerSettings, 0)
    except:
        win32print.ClosePrinter(printer)
        return
    
    win32print.ClosePrinter(printer)
    
    return printerSettings   # makePrinterShare


def makePrinterShareName(shareNameRoot, printerName, server = None):
    
    """
    makePrinterShareName -
    
    Obtain a unique name for the specified shared printer. If the printer is already named
    then return that name otherwise generate a unique name by extending the shareNameRoot.
    """
    
    for n in xrange(sys.maxint):
        if n > 0:
            ext = '_' + str(n)
        else:
            ext = ''
        try:
            getInfo = win32net.NetShareGetInfo(server, shareNameRoot + ext, 2)
            try:
                name = getInfo['path'][0:getInfo['path'].index(',')]
            except:
                name = getInfo['path']
            
            if name.lower() == printerName.lower():
                return getInfo['netname']
        except:
            return shareNameRoot + ext
    
    return   # makePrinterShareName


def promptForAction(userInfo):
    
    """
    promptForAction --
    
    Presents a dialog warning that the current User Password may be different from the default password. If it
    is different then the operator must change the User Name or correct the User Password. If the current mode is
    Easy Mode then the option to restart is prestented so that Custom Mode can be selected.
    """
    
    if not userInfo:
        return
    
    theIssue = _('You may have previously created a User Account named \'%s\' with a password different than the default %s Password. ') % (userInfo['name'], PRODUCT_SHORT_NAME)
    
    if rct.hasMode(CUSTOM_MODE):
        response = askQuestion(theIssue +
                   _('If this is true then you will need to change the User Name or enter the correct Password. Do you want to make changes now?'),
                   Rct.WARNING)
        if response == wx.ID_YES:
            return PASSWORD_CHANGE_REQUESTED
        return userInfo
    
    response = askQuestion(theIssue +
               _('If this is true then you will need to restart %s and use Custom Mode to enter the correct Password. Do you want to restart RCT now?') % PRODUCT_SHORT_NAME,
               Rct.WARNING)
    if response == wx.ID_YES:
        return RESTART_REQUESTED
    
    return userInfo   # promptForAction


def makeUser(userName, password = None, server = None):
    
    """
    makeUser --
    
    Create a user for network share access control.
    """
    
    if not isWindows():
        return NOT_AVAILABLE
    
    level = 1
    
    userInfo = {}
    
    userInfo['name'] = userName
    
    if password:
        userInfo['password'] = password
    else:
        userInfo['password'] = rct.getUserPassword(encrypted = False)
        
    userInfo['priv'] = win32netcon.USER_PRIV_USER
    userInfo['home_dir'] = None
    userInfo['comment'] = 'CZM Networking User'
    userInfo['flags'] = (win32netcon.UF_PASSWD_CANT_CHANGE |
                         win32netcon.UF_DONT_EXPIRE_PASSWD |
                         win32netcon.UF_NORMAL_ACCOUNT |
                         win32netcon.UF_SCRIPT)
    userInfo['script_path'] = None
    
    if not server:
        server = rct.HOST_NAME
    
    try:
        getInfo = win32net.NetUserGetInfo(server, userName, level)
        if getInfo['comment'] != userInfo['comment']:
            return promptForAction(getInfo)
        return getInfo
    except:
        try:
            win32net.NetUserAdd(server, level, userInfo)
            return userInfo
        except:
            return
    
    return   # makeUser


usingNetShareSetInfo = False


def grantAccessToShare(userName, shareName, server = None):
    
    if not isWindows():
        return False
    
    # Retrieve the share's description.
    
    shareInfo = win32net.NetShareGetInfo(server, shareName, 502)
    
    # Get or create the share's security descriptor.
        
    securityDescriptor = shareInfo['security_descriptor']
    
    if securityDescriptor:
        accessControlList = securityDescriptor.GetSecurityDescriptorDacl()
    else:
        securityDescriptor = win32security.SECURITY_DESCRIPTOR()
        securityDescriptor.Initialize()
        accessControlList = win32security.ACL()
        accessControlList.Initialize()
    
    # Determine if the access control list already includes the specified user.
        
    aces = accessControlList.GetExplicitEntriesFromAcl()
    
    for ace in aces:
        try:
            account = win32security.LookupAccountSid(None, ace['Trustee']['Identifier'])
            if account[0].lower() == userName.lower():   # assume that the existing permissions are correct.
                return True
        except:
            continue
    
    # Add the user to the access control list.
    
    try:
        sid = win32security.LookupAccountName('', userName)[0]
        accessControlList.AddAccessAllowedAce(win32security.ACL_REVISION, ntsecuritycon.GENERIC_ALL, sid)
        if usingNetShareSetInfo:
            securityDescriptor.SetSecurityDescriptorDacl(True, accessControlList, False)
    except:
        return False
    
    if usingNetShareSetInfo:
        if DEBUG: print 'grantAccessToShare: Using win32net.NetShareSetInfo'
        shareInfo['security_descriptor'] = securityDescriptor
        try:
            win32net.NetShareSetInfo(server, shareName, 502, shareInfo)
        except:
            return False
    else:
        if DEBUG: print 'grantAccessToShare: Using win32security.SetNamedSecurityInfo'
        win32security.SetNamedSecurityInfo(shareName, win32security.SE_LMSHARE, win32security.DACL_SECURITY_INFORMATION,
                                           sid, None, accessControlList, None)
    
    return True   # grantAccessToShare


def grantAccessToPrinterShare(userName, shareName, printerName, server = None):
    
    """
    grantAccessToPrinterShare --
    
    Adds the specified user to a printer's access control list, returning True if successful
    and False otherwise.
    """
    
    if not isWindows():
        return False
    
    printerDefaults = {'pDataType': None,
                       'pDevMode': None,
                       'DesiredAccess': win32print.PRINTER_ALL_ACCESS}
    
    try:
        printer = win32print.OpenPrinter(printerName, printerDefaults)
    except:
        alert(_('Could not open printer \'%s\'.') % printerName)
        return False
    
    # Retrieve the printer's settings.
        
    level = 2

    try:
        printerSettings = win32print.GetPrinter(printer, level)
    except:
        alert(_('Could not obtain printer settings for \'%s\'.') % printerName)
        win32print.ClosePrinter(printer)
        return False
    
    if shareName != printerSettings['pShareName']:
        alert(_('Printer \'%s\' is already shared as \'%s\'.') % (printerName, shareName))
        return False
    
    # Get or create the printer's security descriptor.
    
    securityDescriptor = printerSettings['pSecurityDescriptor']
    
    if securityDescriptor:
        accessControlList = securityDescriptor.GetSecurityDescriptorDacl()
    else:
        securityDescriptor = win32security.SECURITY_DESCRIPTOR()
        securityDescriptor.Initialize()
        accessControlList = win32security.ACL()
        accessControlList.Initialize()
    
    # Determine if the access control list already includes the specified user.
        
    aces = accessControlList.GetExplicitEntriesFromAcl()
    
    for ace in aces:
        try:
            account = win32security.LookupAccountSid(None, ace['Trustee']['Identifier'])
            if account[0].lower() == userName.lower():   # assume that the existing permissions are correct.
                win32print.ClosePrinter(printer)
                return True
        except:
            continue
    
    # Add the user to the access control list.
    
    try:
        sid = win32security.LookupAccountName('', userName)[0]
        accessControlList.AddAccessAllowedAce(win32security.ACL_REVISION, ntsecuritycon.GENERIC_ALL, sid)
        securityDescriptor.SetSecurityDescriptorDacl(True, accessControlList, False)
    except:
        win32print.ClosePrinter(printer)
        return False
    
    printerSettings['pSecurityDescriptor'] = securityDescriptor
    try:
        win32print.SetPrinter(printer, level, printerSettings, 0)
    except:
        win32print.ClosePrinter(printer)
        return False
    
    win32print.ClosePrinter(printer)
    
    return True   # grantAccessToPrinterShare


def makeHFAfolders(sharedFolder):
    
    """
    makeHFAfolders --
    
    Creates the HFA subfolders if they do not already exist in the specified root shared
    folder.
    """
    try:
        if not os.path.exists(os.path.normpath(sharedFolder + ARCHIVE_FOLDER)):
            os.makedirs(os.path.normpath(sharedFolder + ARCHIVE_FOLDER), mode = SHARED_DIRECTORY_MODE)
        if not os.path.exists(os.path.normpath(sharedFolder + DATA_EXPORT_FOLDER)):
            os.makedirs(os.path.normpath(sharedFolder + DATA_EXPORT_FOLDER), mode = SHARED_DIRECTORY_MODE)
        if not os.path.exists(os.path.normpath(sharedFolder + EMR_EXPORT_FOLDER)):
            os.makedirs(os.path.normpath(sharedFolder + EMR_EXPORT_FOLDER), mode = SHARED_DIRECTORY_MODE)
        if rct.separateSubfolders():
            if not os.path.exists(os.path.normpath(sharedFolder + WORK_LIST_FOLDER)):
                os.makedirs(os.path.normpath(sharedFolder + WORK_LIST_FOLDER), mode = SHARED_DIRECTORY_MODE)
    except:
        return False
    
    return True   # makeHFAfolders


def userName():
    
    """
    userName --
    
    Returns the current user's name.
    """
    
    try:
        return win32api.GetUserName()
    except:
        return os.getlogin()
    
    return  # userName


def validAdmin(userName, server = None):
    
    """
    validAdmin --
    
    Returns True if the specified userName has administrative privileges on the specified server.
    If the server is not specified then it is assumed to be the local computer.
    """
    
    if not isWindows():
        print 'RCT cannot yet test for an administrator on %s systems.' % os.name
        return True 
    else:
        try:
            if not server:
                server = rct.HOST_NAME
            getInfo = win32net.NetUserGetInfo(server, userName, 1)
            return getInfo['priv'] == win32netcon.USER_PRIV_ADMIN
        except:
            return False
    
    return False   # validAdmin


def validRCTPassword(password):
    
    """
    validRCTPassword --
    
    Verifies that the Custom Mode password is equal to the default user password.
    """
    
    if password == defaultUserPassword():
        return True
    else:
        return False


def notYetSupported(event):
    
    """
    notYetSupported --
    
    Present a not-yet-supported debug message.
    """
    
    dlg = wx.MessageDialog(None, 'This feature or function is not yet supported.', 'Note',
                           wx.OK | wx.ICON_INFORMATION)
    dlg.ShowModal()
    dlg.Destroy()
    
    return   # notYetSupported


def askQuestion(msg, title = _('Question')):
    
    """
    askQuestion --
    
    Present a question and return a yes or no response.
    """
    
    dlg = wx.MessageDialog(None, msg, title, wx.YES_NO | wx.YES_DEFAULT | wx.ICON_QUESTION)
    result = dlg.ShowModal()
    dlg.Destroy()

    return result   # askQuestion


def alert(message):
    
    """
    alert --
    
    Present an alert message with an OK button.
    """
    
    dlg = wx.MessageDialog(None, message, Rct.WARNING, wx.OK | wx.ICON_EXCLAMATION)
    dlg.ShowModal()
    dlg.Destroy()
    
    return   # alert


def testReport(text):
    
    """
    testReport --
    
    Display a test result dialog with an OK button.
    """
    
    dlg = wx.MessageDialog(None, text, _('Test Results'), wx.OK | wx.ICON_INFORMATION)
    dlg.ShowModal()
    dlg.Destroy()
    
    return   # testReport


def getButtonName(buttonLabel):
    
    return buttonLabel.replace('&', '').strip(' <>')


class messageWindow(wx.Panel):
    
    """
    messageWindow --
    
    A simple boxed message window used to display messages while communicating with connected
    instruments.
    """
    
    def __init__(self, parent, ID = wx.ID_ANY, label = '',
                 pos = wx.DefaultPosition, size = (LIST_WIDTH, 25)):
        
        wx.Panel.__init__(self, parent, id = ID, pos = pos, size = size)
        
        self.label = label
        self.SetBackgroundColour(RCT_MSG_BG)
        self.SetMinSize(size)
        self.Bind(wx.EVT_PAINT, self.onPaint)
    
    
    def setMessage(self, msg):
        
        self.label = msg
        self.Refresh(True)
        
    
    def onPaint(self, evt):
        
        windowSize = self.GetClientSize()
        context = wx.PaintDC(self)
        textWidth, textHeight = context.GetTextExtent(self.label)
        context.SetFont(wx.Font(((isWindows()) and 10 or 12), wx.SWISS, wx.NORMAL, wx.BOLD))
        context.DrawText(self.label, (windowSize.width - textWidth) / 2, (windowSize.height - textHeight) / 2)


class titleWindow(wx.Panel):

    """
    titleWindow --

    A simple window used to display title image and text.
    """

    def __init__(self, parent, ID = wx.ID_ANY, imageFile = '', label = '',
                 pos = wx.DefaultPosition, size = (100, 50), padding = 10, fontSize = 18):

        wx.Panel.__init__(self, parent, id = ID, pos = pos, size = size)

        self.padding = padding
        if imageFile:
            logoImage = wx.Image(imageFile, wx.BITMAP_TYPE_ANY)
            width = logoImage.GetWidth()
            height = logoImage.GetHeight()
            self.image = logoImage.Scale(width / 2, height / 2).ConvertToBitmap()
        else:
            self.image = None
        self.label = label
        self.fontSize = fontSize
        self.SetBackgroundColour(RCT_TITLE_BG)
        self.SetMinSize(size)
        self.Bind(wx.EVT_PAINT, self.onPaint)


    def setLabel(self, msg):

        self.label = msg
        self.Refresh(True)


    def onPaint(self, evt):

        windowSize = self.GetClientSize()
        context = wx.PaintDC(self)
        if self.image:
            context.DrawBitmap(self.image, self.image.GetWidth() / 2, (windowSize.height - self.image.GetHeight()) / 2, False)
        textWidth, textHeight = context.GetTextExtent(self.label)
        context.SetFont(wx.Font(self.fontSize, wx.SWISS, wx.ITALIC, wx.BOLD))
        if self.image:
            x = 2 * self.image.GetWidth()
        else:
            x = self.padding
        if isWindows():
            y = ((windowSize.height - textHeight) / 2) - (textHeight / 4)
        else:
            y = (windowSize.height - textHeight) / 2
        context.DrawText(self.label, x, y)


class HtmlFileDialog(wx.Dialog):
    
    def __init__(self, parent, title = '', intro = '', filename = '', summary = '', okLabel = 'Ok', cancelLabel = 'Cancel'):
        
        banner = introText = summaryText = None
        
        wx.Dialog.__init__(self, parent, -1, title = '', size = (610, 455))
        
        if title:  
            banner = titleWindow(self, wx.ID_ANY, None, title, size = (500, 50), fontSize = 12)
        
        if intro:
            introText = wx.StaticText(self, wx.ID_ANY, intro, size = (450, 35))
        
        html = wx.html.HtmlWindow(self, style = wx.SUNKEN_BORDER)
        
        if os.path.exists(filename):
            html.LoadFile(filename)
        else:
            html.SetPage('Cannot find source file \'%s\'.' % filename)
        
        if summary:
            summaryText = wx.StaticText(self, wx.ID_ANY, summary)
        
        okButton = wx.Button(self, wx.ID_OK, okLabel)
        cancelButton = wx.Button(self, wx.ID_CANCEL, cancelLabel)
        cancelButton.SetDefault()
        
        buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonSizer.Add(okButton, 0, wx.ALIGN_LEFT | wx.ALL, 0)
        buttonSizer.Add((20, 0))
        buttonSizer.Add(cancelButton, 0, wx.ALIGN_LEFT | wx.ALL, 0)
        buttonSizer.Add((10, 0))
        
        sizer = wx.BoxSizer(wx.VERTICAL)
        if banner:
            sizer.Add(banner, 0, wx.EXPAND | wx.ALL, 0)
        if introText:
            sizer.Add(introText, 0, wx.EXPAND | wx.ALL, 10)
        sizer.Add(html, 1, wx.EXPAND | wx.ALL, 5)
        if summaryText:
             sizer.Add(summaryText, 0, wx.EXPAND | wx.ALL, 10)
             sizer.Add((0, 5))
        sizer.Add(buttonSizer, 0, wx.ALIGN_BOTTOM | wx.ALIGN_RIGHT | wx.ALL, 10)
        sizer.Add((0, 10))
        
        self.SetSizer(sizer)
        self.Layout()


class Worker(threading.Thread):
    
    """
    Worker --
    
    Background thread to communicate with the Service User client. Worker expects to
    be supplied a specific task defining the work.
    """
    
    def __init__(self, task):
        
        self.task = task
        threading.Thread.__init__(self)
    
    
    def stop(self):
        
        pass
    
    
    def run(self):
        
        try:
            result = self.task.run(self)
        except:
            result = UNEXPECTED_RESULT
        
        wx.CallAfter(self.task.complete, self, result)


def delay(delayTime = 3):
    
    """
    delay --
    
    A simple delay (default 3 seconds) to simulate communicating with devices.
    """
    
    time.sleep(delayTime)


class browseServicesTask(object):
    
    """
    browseServicesTask --
    
    A Worker task to browse for device services and obtain device information from connected devices.
    """
    
    def __init__(self, wizardMgr):
        
        self.wizardMgr = wizardMgr
    
    
    def run(self, worker):
        
        if worker:
            self.wizardMgr.addWorker(worker)
        
        result = makeInstruments(self.wizardMgr.getDeviceType())
        
        if not CONFIG_ENABLED:
            delay()
        
        return result
    
    
    def complete(self, worker, result):
        
        self.browsingComplete(worker, result)
    
    
    def browsingComplete(self, browser, result):
        
        self.wizardMgr.removeWorker(browser)
        
        if result == UNEXPECTED_RESULT:
            self.wizardMgr.cancelButton.Enable()
            self.wizardMgr.displayMessage(self.wizardMgr.UNEXPECTED_RESULT_MSG, WAIT_TO_RETURN)
        elif result == NO_AVAILABLE_SERVICES:
            self.wizardMgr.enableNavigationButtons(GOING_BACKWARD)
            self.wizardMgr.displayMessage(_('No instruments are responding. Please try again later.'),
                                          WAIT_TO_RETURN)
        elif result == NO_DEVICE_INFORMATION:
            self.wizardMgr.enableNavigationButtons(GOING_BACKWARD)
            self.wizardMgr.displayMessage(_('Could not retrieve instrument configuration data. Please try again later.'),
                                          WAIT_TO_RETURN)
        else:
            self.wizardMgr.enableNavigationButtons(GOING_FORWARD)
            self.wizardMgr.displayMessage(_('Browsing is complete, please select \'%s\' to continue.') % self.wizardMgr.nextButtonLabel,
                                          DONT_WAIT_TO_RETURN)


class requestCapabilitiesTask(object):
    
    """
    requestCapabilitiesTask --
    
    A Worker task to request configurable capabilities from selected instruments.
    """
    
    def __init__(self, wizardMgr, instruments):
        
        self.wizardMgr = wizardMgr
        self.instruments = instruments
    
    
    def run(self, worker):
        
        if worker:
            self.wizardMgr.addWorker(worker)
        
        capabilities = requestCapabilities(self.instruments)
        
        if not CONFIG_ENABLED:
            delay()
        
        return capabilities
    
    
    def complete(self, worker, result):
        
        self.requestingCapabilitiesComplete(worker, result)
    
    
    def requestingCapabilitiesComplete(self, browser, result):
        
        self.wizardMgr.removeWorker(browser)
        
        if result == UNEXPECTED_RESULT:
            self.wizardMgr.cancelButton.Enable()
            self.wizardMgr.displayMessage(self.wizardMgr.UNEXPECTED_RESULT_MSG, WAIT_TO_RETURN)
            return
        
        if self.wizardMgr.hasMode(COPY_MODE):
            if not result:
                self.wizardMgr.enableNavigationButtons(GOING_BACKWARD)
                self.wizardMgr.displayMessage(_('You must select an instrument with capabilities to copy.'),
                                              WAIT_TO_RETURN)
                return
            capabilities = result[0]
            src = getDevice(self.wizardMgr.COPY_SRC_INSTRUMENT)
            for capability in capabilities:
                if not src.emptyCapability(capability):
                    self.wizardMgr.enableNavigationButtons(GOING_FORWARD)
                    self.wizardMgr.displayMessage(_('Browsing is complete, please select \'%s\' to continue.') % self.wizardMgr.nextButtonLabel,
                                                  DONT_WAIT_TO_RETURN)
                    return
            self.wizardMgr.enableNavigationButtons(GOING_BACKWARD)
            self.wizardMgr.displayMessage(_('You must select an instrument with capabilities to copy.'),
                                          WAIT_TO_RETURN)
            return
        
        self.wizardMgr.enableNavigationButtons(GOING_FORWARD)
        self.wizardMgr.displayMessage(_('Browsing is complete, please select \'%s\' to continue.') % self.wizardMgr.nextButtonLabel,
                                      DONT_WAIT_TO_RETURN)


class configureDevicesTask(object):
    
    """
    configureDevicesTask --
    
    A Worker task to configure devices.
    """
    
    def __init__(self, wizardMgr, devices):
            
        self.wizardMgr = wizardMgr
        self.devices = devices
    
    
    def run(self, worker):
        
        if worker:
            self.wizardMgr.addWorker(worker)
        
        result = True
        for device in self.devices:
            if sendConfig(device):
                result = False
        
        if not CONFIG_ENABLED:
            delay()
        
        return result
    
    
    def complete(self, worker, result):
        
        self.configComplete(worker, result)
    
    
    def configComplete(self, builder, result):
        
        self.wizardMgr.removeWorker(builder)
        
        if result == UNEXPECTED_RESULT:
            self.wizardMgr.cancelButton.Enable()
            self.wizardMgr.displayMessage(self.wizardMgr.UNEXPECTED_RESULT_MSG, WAIT_TO_RETURN)
            return
        
        self.wizardMgr.recordConfigResults(result)
        
        if result:
            self.wizardMgr.enableNavigationButtons(GOING_FORWARD)
            self.wizardMgr.displayMessage(_('Configuration is complete, please select \'%s\' to continue.') % self.wizardMgr.nextButtonLabel,
                                          DONT_WAIT_TO_RETURN)
        else:
            self.wizardMgr.enableNavigationButtons(GOING_FORWARD)
            self.wizardMgr.displayMessage(_('One or more instruments were not fully configured.'),
                                          WAIT_TO_RETURN)


class testDevicesTask(object):
    
    """
    testDevicesTask --
    
    A Worker task to perform connection tests with a specified device.
    """
    
    def __init__(self, wizardMgr, device):
        
        self.wizardMgr = wizardMgr
        self.device = device
    
    
    def run(self, worker):
        
        if worker:
            self.wizardMgr.addWorker(worker)
        
        self.wizardMgr.displayTestProgressMessage()
            
        result = performConnectionTests(self.device)
        
        if result:
            self.device[TEST_RESULTS_FIELD] = PASSED
        else:
            self.device[TEST_RESULTS_FIELD] = FAILED
        
        if not CONFIG_ENABLED:
            delay()
        
        return result
    
    
    def complete(self, worker, result):
        
        self.testComplete(worker, result)
    
    
    def testComplete(self, tester, result):
        
        self.wizardMgr.removeWorker(tester)
        
        if result == UNEXPECTED_RESULT:
            alert(self.wizardMgr.UNEXPECTED_RESULT_MSG)
            setTestResults(self.device, NO_RESULTS)
        
        self.wizardMgr.showTestInstruments()


class Rct(object):
    
    languageCodes = {u'English': 'en', u'Franais': 'fr', u'Italiano': 'it', u'Deutsch': 'de', u'Espaol': 'es'}
    languages = {'en': u'English', 'fr': u'Franais', 'it': u'Italiano', 'de': u'Deutsch', 'es': u'Espaol'}
    
    instruments = [HFA, OCT, AC_OCT, GDX, FRODO, ATLAS, PHOENIX]
    disabledInstruments = [OCT, AC_OCT, GDX, FRODO, ATLAS, PHOENIX]
    
    modes = [EASY_MODE, COPY_MODE, CUSTOM_MODE, REPORT_MODE, TEST_MODE]
    
    INSTRUMENT_DISPLAY_NOTES = ''
    
    SELECT_DEFAULT_PRINTER = ''
    PERFORM_CONNECTION_TESTS = ''
    SEPARATE_EMR_WORK_LIST_FOLDERS = ''
    VIEW_TEST_RESULTS_LABEL = ''
    SAVE_TEST_RESULTS_LABEL = ''
    CHANGE_LANGUAGE_LABEL = ''
    OPEN_PORTS_LABEL = ''
    VIEW_DOC_LABEL = ''
    
    UNKNOWN = ''
    WARNING = ''
    QUESTION = ''
    
    # Network settings display strings.
    
    NETWORK = ''
    IP_CONFIG_PREF_MSG = ''
    NETWORK_SETTINGS_MSG = ''
    
    # Product name strings.
    
    PRODUCT_LONG_NAME = ''
    
    # Other display strings.
    
    SUCCESSFUL_CONFIG_MSG = ''
    CONFIG_FAILED_MSG = ''
    CONFIG_COMPLETED_MSG = ''
    TEST_PREP_COMPLETE_MSG = ''
    SELECT_BEGIN_TESTING_MSG = ''
    NO_INSTRUMENT_MSG = ''
    SHARE_USAGE_MSG = ''
    SHARED_RESOURCE_PROMPT = ''
    BROWSING_IN_PROGRESS_MSG = ''
    FEATURE_REQUEST_IN_PROGRESS_MSG = ''
    CONFIGURATION_PROGRESS_MSG = ''
    NEXT_CONNECTION_TEST_MSG = ''
    CONNECTION_TEST_PROGRESS_MSG = ''
    UNEXPECTED_RESULT_MSG = ''
    NO_TESTS_PERFORMED_MSG = ''
    HOW_TO_TEST_MSG = ''
    TESTING_COMPLETE_MSG = ''
    
    
    def __init__(self):
        
        self.wizard = RctWizard(PRODUCT_SHORT_NAME)
        
        self.instrumentTypeSet = [type for type in Rct.instruments if type not in Rct.disabledInstruments]
        
        self.cancelButton =  self.wizard.FindWindowById(wx.ID_CANCEL)
        self.cancelButtonLabel = getButtonName(self.cancelButton.GetLabel())
        
        self.nextButton = self.wizard.FindWindowById(wx.ID_FORWARD)
        self.nextButtonLabel = getButtonName(self.nextButton.GetLabel())
        
        self.backButton = self.wizard.FindWindowById(wx.ID_BACKWARD)
        self.backButtonLabel = getButtonName(self.backButton.GetLabel())

        self.COMMIT_PAGE_PROMPT = '\n'.join([_('%s is now ready to configure your instruments.') % PRODUCT_SHORT_NAME, '\n',
                                             _('Select \'%s\' to begin the configuration.') % self.nextButtonLabel ,
                                             _('Select \'%s\' to review the setup and make any changes.') % self.backButtonLabel,
                                             _('Select \'%s\' to exit without updating the instruments.') % self.cancelButtonLabel, '\n',
                                             _('If you wish to test your instruments after they have been configured please select %s\'%s\':') % ('\n', Rct.PERFORM_CONNECTION_TESTS)
                                            ])
        
        self.deviceTypeRadioBox = None              # select from different device types if there are more than one.
        self.languageRadioBox = None                # select the local language.
        self.modesRadioBox = None                   # select the configuration mode.
        self.customText = None                      # grant access to custom mode.
        self.setupTypePrompt = None                 # prompt for a configurable setup, Shared Folder or Shared Printer.
        self.setupTypeRadioBox = None               # select a configurable setup.
        self.instrumentSelectionPrompt = None       # prompt for instrument selection.
        self.instrumentListCtrl = None              # selection list of configurable instruments.
        self.instrumentDescriptor = None            # identifies the instrument being configured by type, serial number, and name.
        self.capabilityListCtrl = None              # selection list of configurable features (CUSTOM MODE).
        self.sharedResourceLabel = None             # labels the shared resource text entry field.
        self.sharedResourceEntry = None             # shared resource text entry field.
        self.shareNameEntry = None                  # share name field.
        self.userNamePrompt = None                  # prompt for user name entry (CUSTOM MODE).
        self.userNameText = None                    # user name entry field (CUSTOM MODE).
        self.pwdText = None                         # user password entry field (CUSTOM MODE).
        self.copySrcListCtrl = None                 # selection list for source instruments (COPY MODE).
        self.copyDstListCtrl = None                 # selection list for destination instruments (COPY MODE).
        self.reportInstrListCtrl = None             # selection list of instruments to display (REPORT MODE).
        self.networkSettings = None                 # labels for network settings display (REPORT MODE).
        self.networkSettingsInstrument = None       # identifies an instrument (type, serial number, name) for network settings display (REPORT MODE).
        self.reportingInstrument = None             # identifies an instrument (type, serial number, name) for detailed configuration display (REPORT MODE).
        self.reportDetailListCtrl = None            # displays the instrument feature details (REPORT MODE).
        self.subfoldersCheckBox = None              # enables creation of separate subfolders for selected features.
        self.performTestsCheckBox = None            # enables post configuration connection tests (EASY MODE, COPY MODE, CUSTOM MODE).
        self.defaultPrinterCheckBox = None          # causes the RCT to request that the selected shared printer be made the instrument's default printer.
        self.printerListCtrl = None                 # selection list for locally shared printers.
        self.testStartMessage = None                # prompt for start of connection testing.
        self.testProgressMessage = None             # connection test progress message.
        self.testStartListCtrl = None               # list of instruments to be tested.
        self.testProgressListCtrl = None            # list of instruments being tested.
        self.completionMessage = None               # finish page prompt.
        self.testResultsListCtrl = None             # list of complete test results.
        self.testResultsBox = None                  # text box containing test results buttons.
        self.viewResultsButton = None               # displays a dialog listing an instrument's connection test results.
        self.saveResultsButton = None               # saves the test results messages to a text file.
        self.msgBox = None                          # message text box.
        
        self.SELECTED_SETUP_TYPE = SHARED_FOLDER_TYPE
        
        self.userName = defaultUserName()           # name and
        self.userPassword = defaultUserPassword()   # password to control access to shared resources.
        
        self.SHARED_FOLDER = ''
        self.SHARE_NAME = ''
        self.SHARED_PRINTER = ''
        
        self.DEVICE_TYPE = ''
        self.MODE = EASY_MODE
        
        self.LANGUAGE = ''
        
        self.SELECTED_INSTRUMENTS = []
        self.COPY_SRC_INSTRUMENT = []
        self.COPY_DEST_INSTRUMENTS = []
        self.AVAILABLE_SERVICES = []
        self.RESPONDING_DEVICES = []
        
        self.savedOriginalPage = None
        self.savedPrevPage = NO_PAGE
        self.savedNextPage = None
        
        self.workers = []
        self.configResults = True
        
        # Get the host name and IP address.
        
        self.HOST_NAME = socket.gethostname()
        
        self.HOST_IP_ADDRESS = getIPAddress(self.HOST_NAME)
        if DEBUG:
            print 'The host computer is %s at %s.' % (self.HOST_NAME, self.HOST_IP_ADDRESS)
        
        # Rct.__init__
    
    
    def licenseAgreement(self):
        
        """
        licenseAgreement --
        
        Presents the Apple Bonjour End User License Agreement and prompts the user to agree or decline.
        The response is recorded in the RCT properties file. If the user declines then RCT exits and
        the agreement will be presented again the next time this user runs the RCT. If the user agrees
        then RCT proceeds and the agreement will not be presented the next time this user runs the RCT.
        Note that the license agreement is presented separately for each user.
        """
        
        try:
            license = rct_properties['RCT_BONJOUR_LICENSE']
            if license.lower() == 'agree':
                return True
        except:
            rct_properties['RCT_BONJOUR_LICENSE'] = 'decline'
        
        try:
            language = rct_properties['RCT_LANG']
            if not language:
                language = 'en'
            licenseFile = 'bonjour_%s.htm' % language
        except:
            licenseFile = 'bonjour_en.htm'
        
        dlg = HtmlFileDialog(None, title = _('License Agreement'), 
                                   intro = _('EasyConnect RCT 1.0 makes use of the Bonjour zero configuration networking technology provided by Apple, Inc. ') +
                                           _('Please read the following license agreement carefully.'),
                                   filename = licenseFile,
                                   summary = _('Do you agree with the terms and conditions of this license?'),
                                   okLabel = _('Agree'), 
                                   cancelLabel = _('Decline'))
        response = dlg.ShowModal()
        
        if response == wx.ID_OK:
            rct_properties['RCT_BONJOUR_LICENSE'] = 'agree'
            return True
        
        rct_properties['RCT_BONJOUR_LICENSE'] = 'decline'
        
        return False   # Rct.licenseAgreement
    
    
    def setLocalStrings():
        
        """
        setLocalStrings --
        
        Static method to initialize localized strings.
        """
        
        Rct.INSTRUMENT_DISPLAY_NOTES =   _('If an instrument does not appear, select \'%s\' and try again.')

        Rct.SELECT_DEFAULT_PRINTER =     _('Make this the default printer for the selected instruments')
        Rct.PERFORM_CONNECTION_TESTS =   _('Perform connection tests')                                      # perform connection tests check box label.
        Rct.CREATE_SEPARATE_SUBFOLDERS = _('Create separate subfolders for Work List and EMR/PMS Export')   # separate subfolder check box label.
        Rct.VIEW_TEST_RESULTS_LABEL =    _('View Test Results')                                             # view connection test results button label.
        Rct.SAVE_TEST_RESULTS_LABEL =    _('Save Test Results')                                             # save connection test results button label.
        Rct.CHANGE_LANGUAGE_LABEL =      _('Change Language')                                               # change language button label.
        Rct.OPEN_PORTS_LABEL =           _('Open Ports')                                                    # command to open ports in Windows firewall.
        Rct.VIEW_DOC_LABEL =             _('User Guide')                                                    # view the RCT User Guide.

        Rct.UNKNOWN = _('Unknown')
        Rct.WARNING = _('Warning')
        Rct.QUESTION = _('Question')

        # Network settings display strings.

        Rct.NETWORK = _('Network')
        
        Rct.IP_CONFIG_PREF_MSG = _('Network IP Configuration Preference: %s')
        Rct.NETWORK_SETTINGS_MSG = '\n'.join([_('%s Settings:'),
                                              _('   HFA IP Address:  %s'),
                                              _('   HFA Subnet Mask: %s'),
                                              _('   Default Gateway: %s'), '\n',
                                              _('Workgroup/Domain Name: %s')])
                                              
        # Product name strings.

        Rct.PRODUCT_LONG_NAME = 'EasyConnect%s RCT' % u'\u2122'

        # Other display strings.

        Rct.SUCCESSFUL_CONFIG_MSG =           _('Congratulations. You have completed configuring your instruments.')
        Rct.CONFIG_FAILED_MSG =               _('We were not able to configure some instruments.')
        Rct.CONFIG_COMPLETED_MSG =            _('You have completed configuring your instruments.')
        Rct.TEST_PREP_COMPLETE_MSG =          _('Preparation is complete.')
        Rct.SELECT_BEGIN_TESTING_MSG =        _('Select \'%s\' to begin testing.')
        Rct.SHARE_USAGE_MSG =                 _('Instruments will use a shared folder to store their data on this computer.')
        Rct.SHARED_RESOURCE_PROMPT =          _('Please select \'%s\' to use this folder:')
        Rct.BROWSING_IN_PROGRESS_MSG =        _('Please wait while we look for available instruments . . .')
        Rct.FEATURE_REQUEST_IN_PROGRESS_MSG = _('Please wait while we retrieve configuration details . . .')
        Rct.CONFIGURATION_PROGRESS_MSG =      _('Please wait while we configure your instruments . . .')
        Rct.NEXT_CONNECTION_TEST_MSG =        _('Select \'%s\' to test another instrument.')
        Rct.CONNECTION_TEST_PROGRESS_MSG =    _('Connection test in progress . . .')
        Rct.UNEXPECTED_RESULT_MSG =           _('Unexpected result from worker. Please report.')
        Rct.NO_TESTS_PERFORMED_MSG =          _('No tests were performed.')
        Rct.HOW_TO_TEST_MSG =                 _('You can run the connection tests any time by starting %s and selecting Test mode.') % PRODUCT_SHORT_NAME
        Rct.TESTING_COMPLETE_MSG =            _('Testing is complete, select \'%s\' to continue.')
        
        # Rct.setLocalStrings
    
    setLocalStrings = staticmethod(setLocalStrings)


    def setRctLocale(initialLanguage):

        """
        setRctLocale --

        Initialize the RCT language settings.
        """

        if initialLanguage.upper() != 'NONE':
            try:
                wxLocale = wx.Locale(locales[initialLanguage][0], wx.LOCALE_LOAD_DEFAULT)
            except:
                wxLocale = wx.Locale(wx.LANGUAGE_ENGLISH, wx.LOCALE_LOAD_DEFAULT)

            langdir = os.path.join(getRctHome(), 'locale')

            try:
                Lang = gettext.translation('rct', langdir, languages = [initialLanguage])
                gettextFunction = Lang.ugettext
                Lang.install(unicode = 1)
            except:
                try:
                    initialLanguage = 'en'
                    Lang = gettext.translation('rct', langdir, languages = [initialLanguage])
                    gettextFunction = Lang.ugettext
                    Lang.install(unicode = 1)
                except:
                    if DEBUG:
                        print 'Cannot load the default language dictionary,', initialLanguage

            if DEBUG:
                print 'The host computer\'s locale was set to \'%s\'.' % initialLanguage
            
            return (wxLocale, gettextFunction)   # Rct.setRctLocale
    
    setRctLocale = staticmethod(setRctLocale)
    
    
    def requestingFeatureDetails(self):
        
        return self.savedOriginalPage in [DEVICE_SELECT_PAGE, COPY_SOURCE_PAGE, REPORT_DEVICE_SELECT_PAGE]
    
    
    def configuringDevices(self):
        
        return self.savedOriginalPage is COMMIT_PAGE
    
    
    def displayMessage(self, text, waitForNext):
        
        """
        displayMessage --
        
        Displays a message on the wizard's message page, MESSAGE_PAGE. If waitForNext is True, it
        waits for the operator to select the 'Next' button. If waitForNext is False it automatically
        selects the 'Next' button to transition to the next page.
        """
        
        self.msgBox.setMessage(text)
        
        if not waitForNext:
            self.selectButton(self.nextButton)


    def messageDone(self, success):

        if success:
            if DEBUG_PAGING: print 'Restoring saved next page =', self.savedNextPage
            page = self.savedNextPage
        else:
            if DEBUG_PAGING: print 'Restoring saved previous page =', self.savedPrevPage
            page = self.savedPrevPage
    
        return page   # Rct.messageDone
    
    
    def configurationMessage(self):
        
        if self.configResults:
            return Rct.SUCCESSFUL_CONFIG_MSG
        else:
            return Rct.CONFIG_FAILED_MSG
    
    
    def recordConfigResults(self, result):
        
        self.configResults = result
    
    
    def saveNextPrevPages(self, nextPage, prevPage, originalPage = None):
        
        self.savedNextPage = nextPage
        self.savedPrevPage = prevPage
        
        if originalPage:
            self.savedOriginalPage = originalPage
        else:
            self.savedOriginalPage = prevPage
        
        if DEBUG_PAGING: print 'Saving next page =', self.savedNextPage, 'and previous page =', self.savedPrevPage
    
    
    def makePage(self, prevPage = prevPage, nextPage = nextPage):
        
        """
        makePage --
        
        Convenience method to create an RCT page and add it to the wizard.
        """
        
        try:
            page = TitledPage(self.wizard, '%s %s' % (Rct.PRODUCT_LONG_NAME, PRODUCT_VERSION),
                              prevPage, nextPage)
            self.wizard.addPage(page)
            if CREATE_CMD_POPUP:
                page.addPopupMenuItems([{'text': Rct.VIEW_DOC_LABEL, 'action': viewDoc}])
            return page
        except:
            if DEBUG:
                print 'makePage could not create page.'
            return   # makePage
    
    
    def layoutPages(self, initialLanguage):
        
        # Layout the welcome page (Page 0).
        
        if self.makePage():
            
            welcome = wx.StaticText(self.wizard.pages[START_PAGE], wx.ID_ANY,
                                    _('Welcome to the %s %s, a networking configuration tool for the %s.') % (COMPANY_SHORT_NAME, Rct.PRODUCT_LONG_NAME, rct.getDeviceType()))
            
            self.wizard.pages[START_PAGE].addItem(welcome)
            
            # If there is more than one device type add the device type radio buttons.
            
            if len(self.instrumentTypeSet) > 1:
                self.deviceTypeRadioBox = wx.RadioBox(self.wizard.pages[START_PAGE], wx.ID_ANY,
                                                      _('Please select the type of instrument you want to configure'),
                                                      pos = (10, 10), size = wx.DefaultSize,
                                                      choices = self.instrumentTypeSet,
                                                      majorDimension = 7, style = wx.RA_SPECIFY_COLS,
                                                      validator = wx.DefaultValidator)
    
                self.wizard.pages[START_PAGE].addItem(self.deviceTypeRadioBox)
            
            # Add the language selection radio buttons.
            
            self.languageRadioBox = wx.RadioBox(self.wizard.pages[START_PAGE], wx.ID_ANY, _('Select a language'),
                                                pos = (10, 10), size = wx.DefaultSize,
                                                choices = [language for language in Rct.languageCodes],
                                                majorDimension = 5, style = wx.RA_SPECIFY_COLS,
                                                validator = wx.DefaultValidator)
            try:
                self.languageRadioBox.SetStringSelection(Rct.languages[initialLanguage])
            except:
                self.languageRadioBox.SetStringSelection('English')
            
            self.wizard.pages[START_PAGE].Bind(wx.EVT_RADIOBOX, onLanguageChanged, self.languageRadioBox)
            
            self.wizard.pages[START_PAGE].addItem(self.languageRadioBox)
            
            # Add the rct mode radio buttons.
            
            self.modesRadioBox = wx.RadioBox(self.wizard.pages[START_PAGE], wx.ID_ANY, _('Select a configuration mode'),
                                             pos = (10, 10), size = wx.DefaultSize,
                                             choices = [fieldToString(mode) for mode in Rct.modes],
                                             majorDimension = (not isWindows() and initialLanguage != 'en') and 3 or 5,
                                             style = wx.RA_SPECIFY_COLS, validator = wx.DefaultValidator)
            self.wizard.pages[START_PAGE].addItem(self.modesRadioBox)
            
            # Add the Open Ports command to the start page popup menu.
            
            if CREATE_CMD_POPUP and isWindows():
                self.wizard.pages[START_PAGE].addPopupMenuItems([{'text': Rct.OPEN_PORTS_LABEL, 'action': openPorts}])
            
            # Add the utility commands.
            
            utilityButtons = []
            
            if isWindows():
                self.openPortsButton = wx.Button(self.wizard.pages[START_PAGE], label = Rct.OPEN_PORTS_LABEL, size = self.defaultButtonSize())
                self.wizard.pages[START_PAGE].Bind(wx.EVT_BUTTON, openPorts, self.openPortsButton)
                self.openPortsText = wx.StaticText(self.wizard.pages[START_PAGE], wx.ID_ANY, _('Open ports in Windows XP Firewall for remote file sharing'))
                utilityButtons.append({'button': self.openPortsButton, 'text': self.openPortsText})
            
            self.viewDocButton = wx.Button(self.wizard.pages[START_PAGE], label = Rct.VIEW_DOC_LABEL, size = self.defaultButtonSize())
            self.wizard.pages[START_PAGE].Bind(wx.EVT_BUTTON, viewDoc, self.viewDocButton)
            self.viewDocText = wx.StaticText(self.wizard.pages[START_PAGE], wx.ID_ANY, _('View the EasyConnect RCT 1.0 User Guide'))
            utilityButtons.append({'button': self.viewDocButton, 'text': self.viewDocText})
            helpPrompt = _('Press the F1 key to display the RCT User Guide.')
            
            utilitiesSizer = wx.BoxSizer(wx.VERTICAL)
            for btn in utilityButtons:
                buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
                buttonSizer.Add(btn['button'], 0, wx.ALIGN_CENTRE | wx.ALL, 0)
                buttonSizer.Add((20, 0))
                buttonSizer.Add(btn['text'], 0, wx.ALIGN_CENTRE | wx.ALL, 0)
                utilitiesSizer.Add(buttonSizer)
                utilitiesSizer.Add((0, 10))
            
            if not isWindows():
                self.wizard.pages[START_PAGE].addItem(utilitiesSizer)
            else:
                menuBox = wx.StaticBox(self.wizard.pages[START_PAGE], wx.ID_ANY, ((LABEL_UTILITIES) and _('Other Utilities') or ''))
                menuSizer = wx.StaticBoxSizer(menuBox, wx.VERTICAL)
                menuSizer.Add(utilitiesSizer, 0, wx.ALL, 2)
                self.wizard.pages[START_PAGE].addItem(menuSizer)
            
            # Add prompt for Next button.
            
            self.wizard.pages[START_PAGE].addItem((0, 10), 1)
            prompt = wx.StaticText(self.wizard.pages[START_PAGE], wx.ID_ANY,
                                   _('Select \'%s\' to begin configuring your instruments.') % rct.nextButtonLabel)
                        
            self.wizard.pages[START_PAGE].addItem((0, 10))
            self.wizard.pages[START_PAGE].addItem(prompt)
    
        # Layout the starting page for Custom Mode (Page 1).
        
        if self.makePage():
            customPrompt = wx.StaticText(self.wizard.pages[CUSTOM_PASSWORD_PAGE], wx.ID_ANY, _('Please enter the password for Custom mode:'))
            self.wizard.pages[CUSTOM_PASSWORD_PAGE].addItem(customPrompt)
            
            self.customText = wx.TextCtrl(self.wizard.pages[CUSTOM_PASSWORD_PAGE], wx.ID_ANY, '', size = (175, -1), style = wx.TE_PASSWORD)
            self.wizard.pages[CUSTOM_PASSWORD_PAGE].addItem(self.customText)
        
        # Layout the setup type selection page (Page 2).
        
        if self.makePage():
            self.setupTypePrompt = wx.StaticText(self.wizard.pages[SETUP_TYPE_PAGE], wx.ID_ANY,
                                                 _('What type of setup would you like to perform?'))
            self.wizard.pages[SETUP_TYPE_PAGE].addItem(self.setupTypePrompt)
            
            setupTypes = [fieldToString(SHARED_FOLDER_TYPE), fieldToString(SHARED_PRINTER_TYPE)]
            self.setupTypeRadioBox = wx.RadioBox(self.wizard.pages[SETUP_TYPE_PAGE], wx.ID_ANY, '',
                                                 pos = (10, 10), size = wx.DefaultSize,
                                                 choices = setupTypes,
                                                 majorDimension = 1, style = wx.RA_SPECIFY_COLS,
                                                 validator = wx.DefaultValidator)
            self.wizard.pages[SETUP_TYPE_PAGE].addItem(self.setupTypeRadioBox)
        
        # Layout the instrument selection page (Page 3).
        
        if self.makePage():
            self.instrumentSelectionPrompt = wx.StaticText(self.wizard.pages[DEVICE_SELECT_PAGE], wx.ID_ANY, 
                                                           _('Please select one or more instruments:'))
            self.wizard.pages[DEVICE_SELECT_PAGE].addItem(self.instrumentSelectionPrompt)
            
            self.instrumentListCtrl = InstrumentList(self.wizard.pages[DEVICE_SELECT_PAGE],
                                                     heading = [DEVICE_TYPE_FIELD,
                                                                SERIAL_NUMBER_FIELD,
                                                                DEVICE_NAME_FIELD])
            self.wizard.pages[DEVICE_SELECT_PAGE].addItem(self.instrumentListCtrl)
            
            self.wizard.pages[DEVICE_SELECT_PAGE].addItem(wx.StaticText(self.wizard.pages[DEVICE_SELECT_PAGE], wx.ID_ANY,
                                                                        Rct.INSTRUMENT_DISPLAY_NOTES % self.backButtonLabel))
        
        # Layout the detailed settings page (Page 4 CUSTOM MODE).
        
        if self.makePage():
            self.instrumentDescriptor = wx.StaticText(self.wizard.pages[FEATURE_SELECT_PAGE], wx.ID_ANY,
                                                      _('There is no instrument to configure.'))
            self.wizard.pages[FEATURE_SELECT_PAGE].addItem(self.instrumentDescriptor)
            
            self.wizard.pages[FEATURE_SELECT_PAGE].addItem(wx.StaticText(self.wizard.pages[FEATURE_SELECT_PAGE], wx.ID_ANY,
                                                       _('Please select one or more features to configure:')))
            
            self.capabilityListCtrl = InstrumentList(self.wizard.pages[FEATURE_SELECT_PAGE],
                                                     heading = [FEATURE_FIELD,
                                                                HOST_NAME_FIELD,
                                                                SHARED_RESOURCE_FIELD,
                                                                USER_NAME_FIELD],
                                                     size = (LIST_WIDTH, DETAIL_LIST_HEIGHT))
            
            self.wizard.pages[FEATURE_SELECT_PAGE].addItem(self.capabilityListCtrl)
        
        # Layout the shared folder page (Page 5).
        
        if self.makePage():
            self.sharePrompt = wx.StaticText(self.wizard.pages[FOLDER_SELECT_PAGE], wx.ID_ANY, 
                                             '\n'.join([Rct.SHARE_USAGE_MSG, Rct.SHARED_RESOURCE_PROMPT % self.nextButtonLabel]))
            self.wizard.pages[FOLDER_SELECT_PAGE].addItem(self.sharePrompt)
            self.wizard.pages[FOLDER_SELECT_PAGE].addItem((0, 10))
            
            # Create the shared folder entry.
            
            self.sharedResourceLabel = wx.StaticText(self.wizard.pages[FOLDER_SELECT_PAGE], wx.ID_ANY, '%s:' % fieldToString(SHARED_FOLDER_TYPE))
            self.sharedResourceEntry = wx.TextCtrl(self.wizard.pages[FOLDER_SELECT_PAGE], wx.ID_ANY, defaultSharedFolder(), size = (420, -1))
            
            browseButton = wx.Button(self.wizard.pages[FOLDER_SELECT_PAGE], wx.ID_ANY, _('Browse'))
            self.wizard.pages[FOLDER_SELECT_PAGE].Bind(wx.EVT_BUTTON, selectFolder, browseButton)
            
            # Create the share name display.
            
            shareNameLabelText = _('Share Name')
            shareNameLabel = wx.StaticText(self.wizard.pages[FOLDER_SELECT_PAGE], wx.ID_ANY, '%s:' % shareNameLabelText)
            self.shareNameEntry = wx.StaticText(self.wizard.pages[FOLDER_SELECT_PAGE], wx.ID_ANY, 'Share Name goes here.')
            
            # Add controls to the layout.
            
            shareSizer = wx.FlexGridSizer(cols = 3, hgap = 6, vgap = 6)
            shareSizer.AddMany([self.sharedResourceLabel, self.sharedResourceEntry,
                                browseButton, shareNameLabel, self.shareNameEntry])
            
            self.wizard.pages[FOLDER_SELECT_PAGE].addItem(shareSizer)
            
            # Create a check box for the subfolder option: separate folders for EMR/PMS Export and Work List communication.
            
            self.subfoldersCheckBox = wx.CheckBox(self.wizard.pages[FOLDER_SELECT_PAGE], wx.ID_ANY, Rct.CREATE_SEPARATE_SUBFOLDERS)
            self.subfoldersCheckBox.SetValue(False)
            
            self.wizard.pages[FOLDER_SELECT_PAGE].addItem((0, 20))
            self.wizard.pages[FOLDER_SELECT_PAGE].addItem(self.subfoldersCheckBox)
        
        # Layout the page prompting for the share's user name and password (Page 6).
        
        if self.makePage():
            self.userNamePrompt = wx.StaticText(self.wizard.pages[USER_NAME_PAGE], wx.ID_ANY,
                                                _('Please enter the user name and password for the share:'))
            
            self.wizard.pages[USER_NAME_PAGE].addItem(self.userNamePrompt)
            
            userNameLabel = wx.StaticText(self.wizard.pages[USER_NAME_PAGE], wx.ID_ANY, '%s:' % fieldToString(USER_NAME_FIELD))
            self.userNameText = wx.TextCtrl(self.wizard.pages[USER_NAME_PAGE], wx.ID_ANY, defaultUserName(), size = (175, -1))
            self.userNameText.SetInsertionPoint(0)
            
            pwdLabel = wx.StaticText(self.wizard.pages[USER_NAME_PAGE], wx.ID_ANY, '%s:' % fieldToString(PASSWORD_FIELD))
            self.pwdText = wx.TextCtrl(self.wizard.pages[USER_NAME_PAGE], wx.ID_ANY, defaultUserPassword(),
                                       size = (175, -1), style = wx.TE_PASSWORD)
            
            userNameSizer = wx.FlexGridSizer(cols = 2, hgap = 6, vgap = 6)
            userNameSizer.AddMany([userNameLabel, self.userNameText, pwdLabel, self.pwdText])
            
            self.wizard.pages[USER_NAME_PAGE].addItem(userNameSizer)
        
        # Layout the Copy mode source instrument selection page (Page 7 COPY MODE).
        
        if self.makePage():
            self.wizard.pages[COPY_SOURCE_PAGE].addItem(wx.StaticText(self.wizard.pages[COPY_SOURCE_PAGE], wx.ID_ANY,
                                                       _('Please select the source instrument:')))
            
            self.copySrcListCtrl = InstrumentList(self.wizard.pages[COPY_SOURCE_PAGE],
                                                  heading = [DEVICE_TYPE_FIELD,
                                                             SERIAL_NUMBER_FIELD,
                                                             DEVICE_NAME_FIELD])
            self.wizard.pages[COPY_SOURCE_PAGE].addItem(self.copySrcListCtrl)
            
            self.wizard.pages[COPY_SOURCE_PAGE].addItem(wx.StaticText(self.wizard.pages[COPY_SOURCE_PAGE], wx.ID_ANY,
                                                                      Rct.INSTRUMENT_DISPLAY_NOTES % self.backButtonLabel))
        
        # Layout the Copy mode destination instruments selection page (Page 8 COPY MODE).
        
        if self.makePage():
            self.wizard.pages[COPY_DEST_PAGE].addItem(wx.StaticText(self.wizard.pages[COPY_DEST_PAGE], wx.ID_ANY, 
                                                       _('Please select the destination instruments:')))

            self.copyDstListCtrl = InstrumentList(self.wizard.pages[COPY_DEST_PAGE],
                                                  heading = [DEVICE_TYPE_FIELD,
                                                             SERIAL_NUMBER_FIELD,
                                                             DEVICE_NAME_FIELD])
            self.wizard.pages[COPY_DEST_PAGE].addItem(self.copyDstListCtrl)

            self.wizard.pages[COPY_DEST_PAGE].addItem(wx.StaticText(self.wizard.pages[COPY_DEST_PAGE], wx.ID_ANY,
                                                                    Rct.INSTRUMENT_DISPLAY_NOTES % self.backButtonLabel))
        
        # Layout the Report mode instrument selection page (Page 9 REPORT MODE).

        if self.makePage():
            self.wizard.pages[REPORT_DEVICE_SELECT_PAGE].addItem(wx.StaticText(self.wizard.pages[REPORT_DEVICE_SELECT_PAGE], wx.ID_ANY, 
                                                       _('Please select the instruments whose configuration you wish to display:')))

            self.reportInstrListCtrl = InstrumentList(self.wizard.pages[REPORT_DEVICE_SELECT_PAGE],
                                                      heading = [DEVICE_TYPE_FIELD,
                                                                 SERIAL_NUMBER_FIELD,
                                                                 DEVICE_NAME_FIELD])
            self.wizard.pages[REPORT_DEVICE_SELECT_PAGE].addItem(self.reportInstrListCtrl)
            
            self.wizard.pages[REPORT_DEVICE_SELECT_PAGE].addItem(wx.StaticText(self.wizard.pages[REPORT_DEVICE_SELECT_PAGE], wx.ID_ANY,
                                                                               Rct.INSTRUMENT_DISPLAY_NOTES % self.backButtonLabel))
        
        # Layout the report network settings page (Page 10 REPORT MODE).

        if self.makePage():
            self.networkSettingsInstrument = wx.StaticText(self.wizard.pages[REPORT_NETWORKING_PAGE], wx.ID_ANY, Rct.NO_INSTRUMENT_MSG)
            self.wizard.pages[REPORT_NETWORKING_PAGE].addItem(self.networkSettingsInstrument)

            self.networkSettings = wx.StaticText(self.wizard.pages[REPORT_NETWORKING_PAGE], wx.ID_ANY,
                                                 Rct.NETWORK_SETTINGS_MSG % (Rct.NETWORK,
                                                                             Rct.UNKNOWN, Rct.UNKNOWN, Rct.UNKNOWN, Rct.UNKNOWN))

            self.wizard.pages[REPORT_NETWORKING_PAGE].addItem(self.networkSettings)
        
        # Layout the report feature detail page (Page 11 REPORT MODE).

        if self.makePage():
            self.reportingInstrument = wx.StaticText(self.wizard.pages[REPORT_FEATURE_PAGE], wx.ID_ANY, Rct.NO_INSTRUMENT_MSG)

            self.wizard.pages[REPORT_FEATURE_PAGE].addItem(self.reportingInstrument)

            self.reportDetailListCtrl = InstrumentList(self.wizard.pages[REPORT_FEATURE_PAGE], 
                                                       heading = [FEATURE_FIELD,
                                                                  HOST_NAME_FIELD,
                                                                  SHARED_RESOURCE_FIELD,
                                                                  USER_NAME_FIELD],
                                                       size = (LIST_WIDTH, DETAIL_LIST_HEIGHT))

            self.wizard.pages[REPORT_FEATURE_PAGE].addItem(self.reportDetailListCtrl)
        
        # Layout the commit page (Page 12).

        if self.makePage():
            self.wizard.pages[COMMIT_PAGE].addItem(wx.StaticText(self.wizard.pages[COMMIT_PAGE], wx.ID_ANY, self.COMMIT_PAGE_PROMPT))

            self.performTestsCheckBox = wx.CheckBox(self.wizard.pages[COMMIT_PAGE], wx.ID_ANY, Rct.PERFORM_CONNECTION_TESTS)
            self.performTestsCheckBox.SetValue(False)

            self.wizard.pages[COMMIT_PAGE].addItem(self.performTestsCheckBox)
        
        # Layout the shared printer selection page (Page 13).

        if self.makePage():
            sharedPrinterPrompt = wx.StaticText(self.wizard.pages[PRINTER_SELECT_PAGE], wx.ID_ANY, 
                                                _('Please select a printer. ') +
                                                _('Each of your instruments will be able to use this printer.'))
            self.wizard.pages[PRINTER_SELECT_PAGE].addItem(sharedPrinterPrompt)

            # Create the list of shared printers.

            self.printerListCtrl = PrinterList(self.wizard.pages[PRINTER_SELECT_PAGE],
                                               heading = [fieldToString(PRINTER_NAME_FIELD),
                                                          fieldToString(SHARE_NAME_FIELD)])

            self.wizard.pages[PRINTER_SELECT_PAGE].addItem(self.printerListCtrl)

            if DEFAULT_PRINTER_ENABLED:
                self.defaultPrinterCheckBox = wx.CheckBox(self.wizard.pages[PRINTER_SELECT_PAGE], wx.ID_ANY, Rct.SELECT_DEFAULT_PRINTER)
                self.defaultPrinterCheckBox.SetValue(False)

                self.wizard.pages[PRINTER_SELECT_PAGE].addItem(self.defaultPrinterCheckBox)
        
        # Layout the test start page (Page 14).
        
        if self.makePage():
            self.testStartMessage = wx.StaticText(self.wizard.pages[START_TEST_PAGE], wx.ID_ANY,
                                                  ' '.join([Rct.CONFIG_COMPLETED_MSG,
                                                            Rct.SELECT_BEGIN_TESTING_MSG % self.nextButtonLabel]))
            self.wizard.pages[START_TEST_PAGE].addItem(self.testStartMessage)
            self.wizard.pages[START_TEST_PAGE].addItem((0, 1))
            
            # Display instrument connection test results.
    
            self.testStartListCtrl = InstrumentList(self.wizard.pages[START_TEST_PAGE],
                                                    heading = [DEVICE_TYPE_FIELD,
                                                               SERIAL_NUMBER_FIELD,
                                                               DEVICE_NAME_FIELD,
                                                               TEST_RESULTS_FIELD])
            self.wizard.pages[START_TEST_PAGE].addItem(self.testStartListCtrl)
        
        # Layout the test continuation page (Page 15).
        
        if self.makePage():
            self.testProgressMessage = wx.StaticText(self.wizard.pages[TESTING_PAGE], wx.ID_ANY, Rct.CONNECTION_TEST_PROGRESS_MSG)
            self.wizard.pages[TESTING_PAGE].addItem(self.testProgressMessage)
            self.wizard.pages[TESTING_PAGE].addItem((0, 1))
            
            # Display instrument connection test results.
    
            self.testProgressListCtrl = InstrumentList(self.wizard.pages[TESTING_PAGE],
                                                       heading = [DEVICE_TYPE_FIELD,
                                                                  SERIAL_NUMBER_FIELD,
                                                                  DEVICE_NAME_FIELD,
                                                                  TEST_RESULTS_FIELD])
            self.wizard.pages[TESTING_PAGE].addItem(self.testProgressListCtrl)
        
        # Layout the message page (Page 16).
        
        if self.makePage():
            messageBox = wx.StaticBox(self.wizard.pages[MESSAGE_PAGE], wx.ID_ANY, _('Notice:'))
            messageSizer = wx.StaticBoxSizer(messageBox, wx.VERTICAL)
            self.msgBox = messageWindow(self.wizard.pages[MESSAGE_PAGE], label = Rct.BROWSING_IN_PROGRESS_MSG)
            messageSizer.Add(self.msgBox, 0, wx.ALL | wx.EXPAND, 2)
            
            self.wizard.pages[MESSAGE_PAGE].addItem((0, 0), 1)
            self.wizard.pages[MESSAGE_PAGE].addItem(messageSizer)
        
        # Layout the completion page (Page 17).
        
        if self.makePage():
            self.completionMessage = wx.StaticText(self.wizard.pages[FINISHED_PAGE], wx.ID_ANY,
                                                   Rct.HOW_TO_TEST_MSG)
            self.wizard.pages[FINISHED_PAGE].addItem(self.completionMessage)
            self.wizard.pages[FINISHED_PAGE].addItem((0, 1))
            
            # Create the instrument test results list.
            
            self.testResultsListCtrl = InstrumentList(self.wizard.pages[FINISHED_PAGE],
                                                      heading = [DEVICE_TYPE_FIELD,
                                                                 SERIAL_NUMBER_FIELD,
                                                                 DEVICE_NAME_FIELD,
                                                                 TEST_RESULTS_FIELD],
                                                      size = (LIST_WIDTH, DETAIL_LIST_HEIGHT))
            self.wizard.pages[FINISHED_PAGE].addItem(self.testResultsListCtrl)
            self.wizard.pages[FINISHED_PAGE].addItem((0, 1))
            
            # Create the test results box.
            
            self.testResultsBox = wx.StaticBox(self.wizard.pages[FINISHED_PAGE], wx.ID_ANY, _('Test Results:'))
            
            # Create the view test results button.
            
            self.viewResultsButton = wx.Button(self.wizard.pages[FINISHED_PAGE], label = Rct.VIEW_TEST_RESULTS_LABEL, style = wx.BU_EXACTFIT)
            self.wizard.pages[FINISHED_PAGE].Bind(wx.EVT_BUTTON, viewTestResults, self.viewResultsButton)
            
            # Create the save test results button.
            
            self.saveResultsButton = wx.Button(self.wizard.pages[FINISHED_PAGE], label = Rct.SAVE_TEST_RESULTS_LABEL, style = wx.BU_EXACTFIT)
            self.wizard.pages[FINISHED_PAGE].Bind(wx.EVT_BUTTON, saveResults, self.saveResultsButton)
            
            # Create test results sizer.
            
            testResultsSizer = wx.StaticBoxSizer(self.testResultsBox, wx.HORIZONTAL)
            
            # Add the view and save results buttons.
            
            testResultsSizer.Add(self.viewResultsButton, 0, wx.ALL, 2)
            testResultsSizer.Add((10, 0))
            testResultsSizer.Add(self.saveResultsButton, 0, wx.ALL, 2)
            
            self.wizard.pages[FINISHED_PAGE].addItem((0, 0), 1)
            self.wizard.pages[FINISHED_PAGE].addItem(testResultsSizer)
            
            testResultsSizer.Fit(self.wizard.pages[FINISHED_PAGE])
            
        return   # Rct.layoutPages
    
    
    def getLanguageSelection(self):
        
        self.LANGUAGE = self.languageRadioBox.GetStringSelection()
        
        return Rct.languageCodes[self.LANGUAGE]   # Rct.getLanguageSelection
    
    
    def getLanguageCode(self):
        
        if self.LANGUAGE:
            return Rct.languageCodes[self.LANGUAGE]
        
        return   # Rct.getLanguageCode
    
    
    def setLanguage(self, languageCode):
        
        if languageCode:
            self.LANGUAGE = Rct.languages[languageCode]
        else:
            self.LANGUAGE = None
    
    
    def setMode(self):
        
        self.MODE = stringToField(self.modesRadioBox.GetStringSelection())
        
        return self.MODE   # Rct.setMode
    
    
    def getMode(self):
        
        return self.MODE   # Rct.getMode
    
    
    def hasMode(self, modes):
        
        return self.MODE == modes or self.MODE in modes   # Rct.hasMode
    
    
    def setDeviceType(self):
        
        try:
            if len(self.instrumentTypeSet) > 1:
                self.DEVICE_TYPE = self.deviceTypeRadioBox.GetStringSelection()
            else:
                self.DEVICE_TYPE = self.instrumentTypeSet[0]
        except:
            self.DEVICE_TYPE = HFA
        
        return self.DEVICE_TYPE   # Rct.setDeviceType
    
    
    def getDeviceType(self):
        
        if self.DEVICE_TYPE:
            return self.DEVICE_TYPE
        else:
            return self.instrumentTypeSet[0]   # Rct.getDeviceType
    
    
    def setSelectedSetupType(self):
        
        try:
            self.SELECTED_SETUP_TYPE = stringToField(self.setupTypeRadioBox.GetStringSelection())
        except:
            self.SELECTED_SETUP_TYPE = ALL_SETUP_TYPES
        
        return self.SELECTED_SETUP_TYPE   # Rct.setSelectedSetupType
    
    
    def setAllSetupTypes(self):
        
        self.SELECTED_SETUP_TYPE = ALL_SETUP_TYPES
        
        return self.SELECTED_SETUP_TYPE
    
    
    def folderTypeSelected(self):
        
        return self.SELECTED_SETUP_TYPE == SHARED_FOLDER_TYPE
    
    
    def printerTypeSelected(self):
        
        return self.SELECTED_SETUP_TYPE == SHARED_PRINTER_TYPE
    
    
    def allTypesSelected(self):
        
        return self.SELECTED_SETUP_TYPE == ALL_SETUP_TYPES
    
    
    def isSharedFolderType(self, type):
        
        return type == SHARED_FOLDER_TYPE
    
    
    def isSharedPrinterType(self, type):
        
        return type == SHARED_PRINTER_TYPE
    
    
    def isSelectedSetupType(self, type):
        
        return (self.SELECTED_SETUP_TYPE == ALL_SETUP_TYPES or
                type == self.SELECTED_SETUP_TYPE)
    
    
    def showUserName(self):
        
        self.userNameText.SetValue(self.userName)
    
    
    def setUserName(self, userName = ''):
        
        if userName:
            self.userName = userName
        else:
            try:
                self.userName = self.userNameText.GetValue()
            except:
                self.userName = defaultUserName()
        
        return self.userName    # Rct.setUserName
    
    
    def getUserName(self):
        
        return self.userName
    
    
    def validUserName(self):
        
        """
        validUserName --
        
        Determines if the entered user name is valid. Currently any non-empty
        user name is acceptable.
        """
        
        if self.userNameText.GetValue():
            return True
        
        return False   # Rct.validUserName
    
    
    def showUserPassword(self):
        
        self.pwdText.SetValue(self.userPassword)
    
    
    def setUserPassword(self, password = ''):
        
        if password:
            self.userPassword = password
        else:
            try:
                self.userPassword = self.pwdText.GetValue()
            except:
                self.userPassword = defaultUserPassword()
        
        return self.userPassword   # Rct.setUserPassword
    
    
    def getUserPassword(self, encrypted = False):
        
        if encrypted:
            return ces.encrypt(NETWORK_ENCRYPT_ID, str(self.userPassword))
        else:
            return self.userPassword
    
    
    def validUserPassword(self):
        
        """
        validUserPassword --
        
        Determines if the entered user password is valid. Currently any non-empty
        password is acceptable.
        """
        
        if self.pwdText.GetValue():
            return True
        
        return False   # Rct.validUserPassword
    
    
    def clearNetworkSettings(self):
        
        """
        clearNetworkSettings --
        
        Used by RCT Report Mode to clear the current network settings display.
        """
        
        self.networkSettings.SetLabel(Rct.NETWORK_SETTINGS_MSG % (Rct.NETWORK, 
                                                                  Rct.UNKNOWN, Rct.UNKNOWN, Rct.UNKNOWN, Rct.UNKNOWN))
        
        return   # Rct.clearNetworkSettings
    
    
    def showNetworkSettings(self, ipConfigPref, ipConfig, ipAddress, subnetMask, defaultGateway, domain):
        
        """
        showNetworkSettings --
        
        Used by RCT Report Mode to display the specified network settings.
        """
        
        if ipConfig == ipConfigPref:
            rct.networkSettings.SetLabel(Rct.NETWORK_SETTINGS_MSG % (ipConfig, ipAddress,
                                                                     subnetMask, defaultGateway, domain))
        else:
            rct.networkSettings.SetLabel('\n'.join([Rct.IP_CONFIG_PREF_MSG % ipConfigPref,
                                                    Rct.NETWORK_SETTINGS_MSG % (ipConfig, ipAddress,
                                                                                subnetMask, defaultGateway, domain)]))
        return   # Rct.showNetworkSettings
    
    
    def performTests(self):
        
        """
        performTests --
        
        Returns the state of the perform tests check box. Checked means perform connection tests after
        configuring instruments. Unchecked is the default state.
        """
        
        return self.performTestsCheckBox.IsChecked()
    
    
    def separateSubfolders(self):
        
        """
        separateSubfolders --
        
        Returns the state of the subfolders check box. Checked means create separate EMR/PMS Export and
        Work List subfolders in the shared folder. Unchecked is the default state.
        """
        
        return self.subfoldersCheckBox.IsChecked()
    
    
    def defaultPrinterRequested(self):
        
        """
        defaultPrinterRequested --
        
        Returns the state of the configure default printer check box. Checked means request that the
        configured instrument select the shared printer as its default. Unchecked is the default state.
        This control is currently not used.
        """
        
        if DEFAULT_PRINTER_ENABLED:
            return self.defaultPrinterCheckBox.IsChecked()
        else:
            return False
    
    
    def addWorker(self, worker):
        
        self.workers.append(worker)
    
    
    def removeWorker(self, worker):
        
        try:
            self.workers.remove(worker)
        except:
            pass
    
    
    def enableNavigationButtons(self, forward):
        
        if forward:
            self.nextButton.Enable()
        else:
            self.backButton.Enable()
        self.cancelButton.Enable()
    
    
    def disableNavigationButtons(self):
        
        self.backButton.Disable()
        self.nextButton.Disable()
        self.cancelButton.Disable()
    
    
    def showTestInstruments(self):
        
        if devicesToTest:
            if devicesToTest.hasMore():
                self.testProgressMessage.SetLabel(Rct.NEXT_CONNECTION_TEST_MSG % self.nextButtonLabel)
            else:
                self.testProgressMessage.SetLabel(Rct.TESTING_COMPLETE_MSG % self.nextButtonLabel)
        
        if self.hasMode(COPY_MODE):
            self.testProgressListCtrl.displaySelectedInstruments(self.COPY_DEST_INSTRUMENTS)
        else:
            self.testProgressListCtrl.displaySelectedInstruments(self.SELECTED_INSTRUMENTS)
        
        self.nextButton.Enable()
        self.backButton.Enable()
        self.cancelButton.Enable()
    
    
    def displayTestProgressMessage(self):
        
        """
        displayTestProgressMessage --
        
        Replaces the wizard's test page prompt with the connection test progress message.
        """
        
        self.testProgressMessage.SetLabel(Rct.CONNECTION_TEST_PROGRESS_MSG)
    
    
    def defaultButtonSize(self, scale = 1.45):
        
        """
        defaultButtonSize --
        
        Returns a button size tuple based on the wizard's next button size, but 45% wider if not otherwise
        specified by its scale argument.
        """
        
        size = self.nextButton.GetSize()
        
        return (size[0] * scale, size[1])   # defaultButtonSize
    
    
    def setMaxButtonSize(self, buttons):
        
        w = 0
        h = 0
        
        for b in buttons:
            size = b['button'].GetSize()
            w = max(w, size[0])
            h = max(h, size[1])
        
        for b in buttons:
            b['button'].SetSize((w, h))
    
    
    def selectButton(self, button):
        
        """
        selectButton --
        
        Activate a button as if selected via the user interface.
        """

        buttonEvent = wx.CommandEvent(wx.wxEVT_COMMAND_BUTTON_CLICKED)
        buttonEvent.SetEventObject(button)
        buttonEvent.SetId(button.GetId())
        button.GetEventHandler().ProcessEvent(buttonEvent)
    
    
    def start(self):
        
        """
        start --
        
        Start running the RCT wizard after binding its event handlers.
        """
        
        if self.wizard.pages[START_PAGE]:

            self.wizard.FitToPage(self.wizard.pages[START_PAGE])

            self.wizard.Bind(wx.wizard.EVT_WIZARD_PAGE_CHANGING, onPageChanging)
            self.wizard.Bind(wx.wizard.EVT_WIZARD_PAGE_CHANGED, onPageChanged)
            self.wizard.Bind(wx.wizard.EVT_WIZARD_CANCEL, onCancel)
            self.wizard.Bind(wx.wizard.EVT_WIZARD_FINISHED, onFinished)

            self.wizard.run()
    
    
    def stop(self):
        
        """
        stop --
        
        Terminate the RCT wizard by selecting its Cancel button.
        """
        
        self.selectButton(self.cancelButton)
    
    
    def restart(self):
        
        """
        restart --
        
        Terminate the RCT wizard and signal it to restart.
        """
        
        global forceRestart
        
        self.getLanguageSelection()
        forceRestart = True
        self.stop()


def serviceType():
    
    if INCLUDE_SERVICE_TYPE_QUALIFIER:
        return SERVICE_TYPE + ',' + rct.getDeviceType().lower() + '-RCS'
    
    return SERVICE_TYPE   # serviceType   


def startBonjour(option = 'install'):
    
    """
    startBonjour --
    
    Installs the Bonjour service on Windows systems. Can also run Bonjour directly, recording the
    process name for later termination.
    """
    
    global bonjourInstalled
    result = 0
    
    if not isWindows():
        return result
    
    if option.lower() == 'server':
        try:
            process = subprocess.Popen(['mDNSResponder.exe', '-q', '-server'])
            rctProcs.append((process, 'mDNSResponder'))
        except:
            if DEBUG:
                print 'Exception while attempting to run Bonjour mDNSResponder directly: %s.' % result
    elif option.lower() == 'install' and not bonjourInstalled:
        try:
            result = 0
            cmd = 'installBonjourService.bat'
            result = os.system(cmd)
            if result == 0:
                bonjourInstalled = True
            else:
                process = subprocess.Popen(['mDNSResponder.exe', '-q', '-server'])
                rctProcs.append((process, 'mDNSResponder'))
        except:
            if DEBUG:
                print 'Exception while installing Bonjour service: %s.' % result
    
    return result   # startBonjour 


def shutdown():
    
    """
    shutdown --
    
    Clean up before exiting so that rct can restart if necessary.
    """
    
    global _
    global wxlocale
    global rct
    global serviceUser
    global instrumentSelector
    global devicesToTest
    global rct_properties
    global forceRestart
    
    saveRctProperties(rct_properties, 'rct_props.txt')
    
    # Remove the temporary Service User directory.
    
    try:
        serviceUser.removeServiceDirectory()
    except:
        if DEBUG: print 'Error removing temporary directory.'
    serviceUser = None
    
    # Delete the device instances.
    
    for device in tuple(Device.devices()):
         device.delete()
    
    # Re-initialize global variables.
    
    rct = None
    
    instrumentSelector = None
    devicesToTest = None
    
    _ = None
    wxlocale = None
    
    forceRestart = False
    
    # Terminate any subprocesses that are still running.
    
    try:
        if rctProcs:
            for process, name in rctProcs:
                handle = win32api.OpenProcess(win32con.PROCESS_TERMINATE, False, process.pid)
                win32api.TerminateProcess(handle, -1)
                win32api.CloseHandle(handle)
    except:
        pass


def validLocale(locale):
    
    return locale in [langCode for langCode in locales]


def parseProperty(param):

    name = ''
    value = ''
        
    paramList = param.replace('::', '=').split('=')

    try:
        name = paramList[0].strip()
        value = paramList[1].strip()
        if name and not value:
            return (name, None)
        else:
            return (name, value)
    except:
        return (name, None)   # parseProperty


def getRctProperties(propertyFileName):
    
    """
    getRctProperties --
    
    Reads RCT program properties from the specified property file in the application's data
    directory.
    """
    
    rct_properties = {}
    
    try:     # Windows application directory.
        appDir = shell.SHGetSpecialFolderPath(0, shellcon.CSIDL_APPDATA)
    except:
        try: # Posix home directory
            appDir = os.environ.get('HOME')
        except:
            appDir = '.'
    
    rctPropertyDir = os.path.join(appDir, '%s/%s' % (COMPANY_NAME, PRODUCT_SHORT_NAME))
    
    if not os.path.isdir(rctPropertyDir):
        print 'Making directories', rctPropertyDir
        os.makedirs(rctPropertyDir, mode = SHARED_DIRECTORY_MODE)
    rctPropertyPath = os.path.join(rctPropertyDir, propertyFileName)
    
    try:
        propFile = open(rctPropertyPath)
    except IOError:
        if DEBUG:
            print 'Could not open property file \'%s\'.' % rctPropertyPath
        return rct_properties
    
    prop = propFile.readline().strip()
    while prop and prop[0] == '#':
        prop = propFile.readline().strip()
    
    while prop:
        name, value = parseProperty(prop)
        if name:
            rct_properties[name] = value
        prop = propFile.readline().strip()
    
    return rct_properties   # getRctProperties  


def saveRctProperties(rct_properties, propertyFileName):
    
    """
    saveRctProperties --
    
    Writes RCT program properties to the specified property file in the application's data
    directory.
    """
    
    try:     # Windows application directory.
        appDir = shell.SHGetSpecialFolderPath(0, shellcon.CSIDL_APPDATA)
    except:
        try: # Posix home directory
            appDir = os.environ.get('HOME')
        except:
            appDir = '.'
    
    rctPropertyDir = os.path.join(appDir, '%s/%s' % (COMPANY_NAME, PRODUCT_SHORT_NAME))
    
    if not os.path.isdir(rctPropertyDir):
        print 'Making directories', rctPropertyDir
        os.makedirs(rctPropertyDir, mode = SHARED_DIRECTORY_MODE)
    rctPropertyPath = os.path.join(rctPropertyDir, propertyFileName)
    
    try:
        propFile = open(rctPropertyPath, 'w')
    except IOError:
        print 'Could not open property file \'%s\'.' % rctPropertyPath
        return False
    
    for prop in rct_properties:
        propFile.write('%s = %s\n' % (prop, rct_properties[prop]))
    
    propFile.close()
    
    return True   # saveRctProperties
    

def main(initialLocale):
    
    global _
    global wxlocale
    global rct
    global serviceUser
    global rct_properties
    
    # Initialize the application.
    
    app = wx.PySimpleApp()
    
    # Change to the working directory.
    
    os.chdir(getRctHome())
    
    # Initialize the user interface language.
    
    (wxlocale, _) = Rct.setRctLocale(initialLocale)
    Rct.setLocalStrings()
    
    # Create and layout the RCT wizard.
    
    rct = Rct()
    
    if not rct:
        alert(_('%s could not create the wizard. Please report this problem.') % PRODUCT_SHORT_NAME)
        return
        
    rct.layoutPages(initialLocale)
    
    # Create the single Service User instance.
    
    serviceUser = ServiceUser(serviceType(), CONFIG_ENABLED, VERBOSE, DELETE_FILES)
    
    # Start the wizard.
    
    rct.start()
    
    # If a new language was selected, then prepare to return the new language code.
    
    languageCode = rct.getLanguageCode()
    if not forceRestart and languageCode == initialLocale:
        languageCode = None
    
    # Save language code for next time.
    
    if languageCode:
        rct_properties['RCT_LANG'] = languageCode
    
    # Clean up for exit.
    
    shutdown()
    
    if DEBUG: print 'Exiting RCT.'
    
    return languageCode   # main


if __name__ == '__main__':
    
    # Get the RCT program properties.
    
    rct_properties = getRctProperties('rct_props.txt')
    
    # Determine the initial display language.
    
    try:
        initialLocale = rct_properties['RCT_LANG']
    except:
        initialLocale = DEFAULT_LOCALE
    
    # Get the program arguments.
    
    p = optparse.OptionParser()
    
    p.add_option('--saveFiles',      '-s', action = 'store_true',
                                           default = False,   dest = 'saveFiles',
                                           help = 'Do not delete scripts, saving the last instance of each file.')
    p.add_option('--debug',          '-d', action = 'store_true',
                                           default = False, dest = 'debug',
                                           help = 'Send debug information to stdout.')
    p.add_option('--verbose',        '-v', action = 'store_true',
                                           default = False, dest = 'verbose',
                                           help = 'Print additional messages and debug information.')
    p.add_option('--debugPaging',    '-p', action = 'store_true',
                                           default = False, dest = 'debugPaging',
                                           help = 'Send paging debug information to stdout.')
    p.add_option('--configDisabled', '-c', action = 'store_false', dest = 'configEnabled',
                                           default = True, 
                                           help = 'Disable configuration, do not invoke the Service User client.')
    p.add_option('--friends',        '-f', action = 'store', dest = 'friends',
                                           default = '',
                                           help = 'Specify allowed services, e.g. -f "RCS75054321 RCS7506166".')
    p.add_option('--locale',         '-l', action = 'store', dest = 'locale',
                                           default = initialLocale,
                                           help = 'Specify the local, e.g. -l de.')
    
    options, args = p.parse_args()
    
    VERBOSE = options.verbose
    DELETE_FILES = not options.saveFiles
    DEBUG = options.debug
    DEBUG_PAGING = options.debugPaging
    CONFIG_ENABLED = options.configEnabled
    FRIENDS = options.friends.split()
    theLocale = options.locale
    
    if DEBUG and rct_properties:
        print '\n%s, Version %s' % (rct_properties['RCT_PROGRAM_NAME'], rct_properties['RCT_VERSION'])
        print 'Copyright %s\n' % rct_properties['RCT_COPYRIGHT']
    
    if DEBUG and not isWindows():
        print 'Non-Windows platform:\n', pprint.pformat(wx.PlatformInfo)
    
    # Launch the main application. Restart if main() returns a locale, a new language was selected.
    
    while validLocale(theLocale):
        theLocale = main(theLocale)
        if theLocale and DEBUG: print 'The new language is \'%s\'.' % theLocale
